;;;============================================================================
;;;
;;; Ttulo: Sintaxis de la lgica proposicional
;;; Parte del proyecto de tesis:
;;;    "Verificacin de sistemas de razonamiento automtico"
;;; FJMM 21-09-2000
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../3-logica/sintaxis")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 1) Smbolos y conectivas
;;;
;;;============================================================================

(defun es-simbolo-proposicional (x)
  (or (and (symbolp x) (not (equal x nil)))
      (and (integerp x) (>= x 0))))

(defun es-conectiva-monaria (c)
  (member-equal c '(-)))

(defun es-conectiva-binaria (c)
  (member-equal c '(& \/ -> <->)))

;;;============================================================================
;;;
;;; 2) Representacin de las frmulas
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Funciones de acceso
;;;
;;;----------------------------------------------------------------------------

(defun op (exp)
  (cond ((and (integerp exp) (< exp 0)) '-)
	((> (len exp) 0) (car exp))
	(t nil)))

(defun arg1 (exp)
  (cond ((and (integerp exp) (< exp 0)) (- exp))
	((> (len exp) 1) (cadr exp))
	(t nil)))

(defun arg2 (exp)
  (if (> (len exp) 2)
      (caddr exp)
      nil))

;;;----------------------------------------------------------------------------
;;;
;;; 2.2) Funciones reconocedoras
;;;
;;;----------------------------------------------------------------------------

(defun es-monaria (exp)
  (or (and (integerp exp)
	   (< exp 0))
      (and (es-conectiva-monaria (op exp))
	   (= (len exp) 2)
	   (not (and (integerp (cadr exp))
		     (> (cadr exp) 0)))
	   (null (cddr exp)))))

(defun es-binaria (exp)
  (and (es-conectiva-binaria (op exp))
       (= (len exp) 3)
       (null (cdddr exp))))

(defthm monaria-y-binaria->contradiccion
  (implies (and (es-monaria exp)
                (es-binaria exp))
           nil)
  :rule-classes :forward-chaining)

(defthm es-simbolo-proposicional-not-monaria-binaria
  (implies (es-simbolo-proposicional x)
           (and (not (es-monaria x))
                (not (es-binaria x)))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.3) Funciones constructoras
;;;
;;;----------------------------------------------------------------------------

(defun negacion (F)
  (if (and (integerp F) (> F 0))
      (- F)
      (list '- F)))

(defun conjuncion (F G)
  (list '& F G))

(defun disyuncion (F G)
  (list '\/ F G))

(defun implicacion (F G)
  (list '-> F G))

(defun equivalencia (F G)
  (list '<-> F G))

(defthm negacion-es-monaria
  (es-monaria (negacion F)))

(defthm con-dis-imp-sii-es-binaria
  (and (es-binaria (conjuncion F G))
       (es-binaria (disyuncion F G))
       (es-binaria (implicacion F G))
       (es-binaria (equivalencia F G))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.4) Relacin entre funciones de acceso y constructoras
;;;
;;;----------------------------------------------------------------------------

(defthm acceso-negacion
  (and (equal (op (negacion F)) '-)
       (equal (arg1 (negacion F)) F)))

(defthm acceso-conjuncion
  (and (equal (op (conjuncion F G)) '&)
       (equal (arg1 (conjuncion F G)) F)
       (equal (arg2 (conjuncion F G)) G)))

(defthm acceso-disyuncion
  (and (equal (op (disyuncion F G)) '\/)
       (equal (arg1 (disyuncion F G)) F)
       (equal (arg2 (disyuncion F G)) G)))

(defthm acceso-implicacion
  (and (equal (op (implicacion F G)) '->)
       (equal (arg1 (implicacion F G)) F)
       (equal (arg2 (implicacion F G)) G)))

(defthm acceso-equivalencia
  (and (equal (op (equivalencia F G)) '<->)
       (equal (arg1 (equivalencia F G)) F)
       (equal (arg2 (equivalencia F G)) G)))

;;;============================================================================
;;;
;;; 3) Frmulas proposicionales
;;;
;;;============================================================================

(defun es-atomica (F)
  (es-simbolo-proposicional F))

(defthm atomica->sin-operador
  (implies (es-atomica F)
           (not (op F))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.1) Nmero de conectivas de una frmula
;;;
;;;----------------------------------------------------------------------------

(defun list-integer-count (x)
  (if (consp x)
      (+ 1 (list-integer-count (car x))
	 (list-integer-count (cdr x)))
      (if (and (integerp x)
	       (< x 0))
	  1
	  0)))

(defthm list-integer-count-arg1-decrece
  (implies (>= (len exp) 2)
           (< (list-integer-count (arg1 exp))
              (list-integer-count exp))))

(defthm list-integer-count-arg2-decrece
  (implies (>= (len exp) 3)
           (< (list-integer-count (arg2 exp))
              (list-integer-count exp))))

(in-theory (disable negacion (negacion)
                    conjuncion (conjuncion)
                    disyuncion (disyuncion)
                    implicacion (implicacion)
                    equivalencia (equivalencia)
                    op (op)
                    arg1 (arg1)
                    arg2 (arg2)))

(defun numero-conectivas (F)
  (declare (xargs :measure (list-integer-count F)
		  :hints (("Goal" :in-theory (enable arg1 arg2)))))

  (cond ((es-monaria F)
         (+ (numero-conectivas (arg1 F)) 1))
        ((es-binaria F)
         (+ (numero-conectivas (arg1 F))
            (numero-conectivas (arg2 F)) 1))
        (t 0)))

(in-theory (disable list-integer-count-arg1-decrece
                    list-integer-count-arg2-decrece))

;;;   El resultado para frmulas monarias es el siguiente:

(defthm numero-conectivas-monaria-decrece
  (implies (es-monaria F)
           (< (numero-conectivas (arg1 F))
              (numero-conectivas F)))
  :rule-classes :linear)

(defthm numero-conectivas-binaria-decrece
  (implies (es-binaria F)
           (and (< (numero-conectivas (arg1 F))
                   (numero-conectivas F))
                (< (numero-conectivas (arg2 F))
                   (numero-conectivas F))))
  :rule-classes :linear)

;;;----------------------------------------------------------------------------
;;;
;;; 3.2) Definicin de es-proposicional
;;;
;;;----------------------------------------------------------------------------

(defun es-proposicional (F)
  (declare (xargs :measure (numero-conectivas F)
                  :hints (("Goal" 
                           :in-theory (disable numero-conectivas
                                               es-monaria es-binaria)))))
  (cond ((es-monaria F) 
         (es-proposicional (arg1 F)))
        ((es-binaria F) 
         (and (es-proposicional (arg1 F))
              (es-proposicional (arg2 F))))
        (t (es-atomica F))))

;;;============================================================================
;;;
;;; 4) Principio de induccin estructural
;;;
;;;============================================================================

(encapsulate
 (((q0 *) => *))

 (local
  (defun q0 (F)
    (declare (ignore F))
    t))

 (defthm q0-es-booleana
   (booleanp (q0 F))
   :rule-classes :type-prescription)

 (defthm caso-base
   (implies (es-atomica F)
            (q0 F)))

 (defthm paso-de-induccion-1
   (implies (and (es-monaria F)
                 (q0 (arg1 F)))
            (q0 F)))

 (defthm paso-de-induccion-2
   (implies (and (es-binaria F)
                 (q0 (arg1 F))
                 (q0 (arg2 F)))
            (q0 F)))

 )

(defthm induccion-estructural
  (implies (es-proposicional F)
           (q0 F)))

;;;============================================================================
;;;
;;; 5) Clasificacin de frmulas
;;;
;;;============================================================================

(defun es-negacion (F)
  (and (equal (op F) '-)
       (es-proposicional F)))

(defun es-conjuncion (F)
  (and (equal (op F) '&)
       (es-proposicional F)))

(defun es-disyuncion (F)
  (and (equal (op F) '\/)
       (es-proposicional F)))

(defun es-implicacion (F)
  (and (equal (op F) '->)
       (es-proposicional F)))

(defun es-equivalencia (F)
  (and (equal (op F) '<->)
       (es-proposicional F)))

(defthm negacion-es-negacion
  (implies (es-proposicional F)
           (es-negacion (negacion F))))

(defthm conjuncion-es-conjuncion
  (implies (and (es-proposicional F)
                (es-proposicional G))
           (es-conjuncion (conjuncion F G))))

(defthm disyuncion-es-disyuncion
  (implies (and (es-proposicional F)
                (es-proposicional G))
           (es-disyuncion (disyuncion F G))))

(defthm implicacion-es-implicacion
  (implies (and (es-proposicional F)
                (es-proposicional G))
           (es-implicacion (implicacion F G))))

(defthm equivalencia-es-equivalencia
  (implies (and (es-proposicional F)
                (es-proposicional G))
           (es-equivalencia (equivalencia F G))))

(deftheory definiciones-conectivas
  '(es-monaria es-negacion
    es-binaria es-conjuncion es-disyuncion es-implicacion es-equivalencia))

;;; 1)
;;;   La negacin de una frmula no es igual al original:

(defthm negacion-no-igual-al-original
  (not (equal (negacion F) F))
  :hints (("Goal" :in-theory (enable negacion))))

;;; 2)
;;;   La negacin del argumento de una frmula proposicional negativa es igual
;;; a la original. Este resultado precisa de las funciones de acceso op y
;;; arg1 y de la funcin constructora negacion:

(defthm construccion-negacion
  (implies (es-negacion F)
           (equal (negacion (arg1 F)) F))
  :rule-classes :elim
  :hints (("Goal" :in-theory (enable op arg1 negacion))))

;;; 3)
;;;   Toda frmula negativa es monaria:

(defthm es-negacion-es-monaria
  (implies (es-negacion F)
           (es-monaria F))
  :rule-classes :type-prescription)

;;; 4)
;;;   Toda frmula conjuntiva, disyuntiva, de implicacin o equivalencia es
;;; binaria: 

(defthm es-con-dis-imp-sii-es-binaria
  (implies (or (es-conjuncion F)
               (es-disyuncion F)
               (es-implicacion F)
               (es-equivalencia F))
           (es-binaria F))
  :rule-classes :type-prescription
  :hints (("Goal" :in-theory (enable op))))

;;; 5)
;;;   El argumento de toda frmula negativa es una frmula proposicional:

(defthm argumentos-monarias-proposicionales
  (implies (es-negacion F)
           (es-proposicional (arg1 F))))

;;; 6)
;;;   Los argumentos de toda frmula conjuntiva, disyuntiva, implicacin o
;;; equivalencia, son frmulas proposicionales:

(defthm argumentos-binarias-proposicionales
  (implies (or (es-implicacion F)
               (es-disyuncion F)
               (es-conjuncion F)
               (es-equivalencia F))
           (and (es-proposicional (arg1 F))
                (es-proposicional (arg2 F))))
  :hints (("Goal" :in-theory (enable op))))

;;;----------------------------------------------------------------------------
;;;
;;; 5.2) Otra definicin de es-proposicional
;;;
;;;----------------------------------------------------------------------------

(defthm otra-definicion-de-es-proposicional
  (iff (es-proposicional F)
       (or (es-atomica F)
           (es-negacion F)
           (es-conjuncion F)
           (es-disyuncion F)
           (es-implicacion F)
           (es-equivalencia F)))
  :rule-classes :definition
  :hints (("Goal" :in-theory (enable op))))

(defthm es-proposicional-no-compuesta-es-atomica
  (implies (and (es-proposicional F)
                (not (es-negacion F))
                (not (es-conjuncion F))
                (not (es-disyuncion F))
                (not (es-implicacion F))
                (not (es-equivalencia F)))
           (es-atomica F))
  :hints (("Goal" :in-theory (disable definiciones-conectivas))))

;;;============================================================================
;;;
;;; 6) Teorema de descomposicin nica
;;;
;;;============================================================================

(defthm unicidad-es-atomica
  (implies (es-atomica F)
           (and (not (es-negacion F))
                (not (es-conjuncion F))
                (not (es-disyuncion F))
                (not (es-implicacion F))
                (not (es-equivalencia F)))))

(defthm unicidad-es-negacion
  (implies (es-negacion F)
           (and (not (es-atomica F))
                (not (es-conjuncion F))
                (not (es-disyuncion F))
                (not (es-implicacion F))
                (not (es-equivalencia F)))))

(defthm unicidad-es-conjuncion
  (implies (es-conjuncion F)
           (and (not (es-atomica F))
                (not (es-negacion F))
                (not (es-disyuncion F))
                (not (es-implicacion F))
                (not (es-equivalencia F)))))

(defthm unicidad-es-disyuncion
  (implies (es-disyuncion F)
           (and (not (es-atomica F))
                (not (es-negacion F))
                (not (es-conjuncion F))
                (not (es-implicacion F))
                (not (es-equivalencia F)))))

(defthm unicidad-es-implicacion
  (implies (es-implicacion F)
           (and (not (es-atomica F))
                (not (es-negacion F))
                (not (es-conjuncion F))
                (not (es-disyuncion F))
                (not (es-equivalencia F)))))

(defthm unicidad-es-equivalencia
  (implies (es-equivalencia F)
           (and (not (es-atomica F))
                (not (es-negacion F))
                (not (es-conjuncion F))
                (not (es-disyuncion F))
                (not (es-implicacion F)))))

;;;============================================================================
;;;
;;; 7) Smbolos proposicionales de una frmula
;;;
;;;============================================================================

(defun simbolos-proposicionales (F)
  (declare (xargs :measure (numero-conectivas F)
                  :hints (("Goal" 
                           :hands-off (es-monaria es-binaria)
                           :in-theory (disable numero-conectivas)))
                  :verify-guards nil))
  (cond ((es-monaria F) 
         (simbolos-proposicionales (arg1 F)))
        ((es-binaria F) 
         (union-equal (simbolos-proposicionales (arg1 F))
                      (simbolos-proposicionales (arg2 F))))
        ((es-simbolo-proposicional F) (list F))
        (t nil)))

(defthm union-equal-de-listas-propias-es-lista-propia
  (implies (and (true-listp lista1)
                (true-listp lista2))
           (true-listp (union-equal lista1 lista2))))

(defthm simbolos-proposicionales-es-lista-propia
  (true-listp (simbolos-proposicionales F))
  :hints (("Goal" :in-theory (disable definiciones-conectivas))))

(verify-guards simbolos-proposicionales)

(defthm pertenencia-a-una-union
  (iff (member-equal elemento (union-equal lista-1 lista-2))
       (or (member-equal elemento lista-1)
           (member-equal elemento lista-2))))

(defthm union-no-duplica 
  (implies (and (no-duplicatesp-equal lista-1)
                (no-duplicatesp-equal lista-2))
           (no-duplicatesp-equal (union-equal lista-1 lista-2))))

(defthm simbolos-proposicionales-es-conjunto
  (no-duplicatesp-equal (simbolos-proposicionales F))
  :hints (("Goal" :in-theory (disable definiciones-conectivas))))

(defthm subsetp-equal-cons
  (implies (subsetp-equal conj-1 conj-2)
	   (subsetp-equal conj-1 (cons elemento conj-2))))

(defthm subsetp-equal-reflexiva
  (subsetp-equal conjunto conjunto))

(defthm subsetp-equal-simbolos-proposicionales-monaria-lst
  (implies (and (es-monaria F)
                (subsetp-equal (simbolos-proposicionales F) lst))
           (subsetp-equal (simbolos-proposicionales (arg1 F)) lst))
  :hints (("Goal" :in-theory (disable definiciones-conectivas))))

(defthm subsetp-equal-simbolos-proposicionales-monaria
  (implies (es-monaria F)
           (subsetp-equal (simbolos-proposicionales (arg1 F))
			  (simbolos-proposicionales F)))
  :hints (("Goal" :in-theory (disable definiciones-conectivas))))
 
(defthm subsetp-equal-union
  (implies (subsetp-equal (union-equal conj-1 conj-2) conj-3)
           (and (subsetp-equal conj-1 conj-3)
                (subsetp-equal conj-2 conj-3)))
  :hints (("Goal"
           :induct (union-equal conj-1 conj-2))))

(defthm subsetp-equal-simbolos-proposicionales-binaria-lst
  (implies (and (es-binaria F)
                (subsetp-equal (simbolos-proposicionales F) lst))
           (and (subsetp-equal (simbolos-proposicionales (arg1 F)) lst)
                (subsetp-equal (simbolos-proposicionales (arg2 F)) lst)))
  :hints (("Goal" :in-theory (disable definiciones-conectivas))))

(defthm subsetp-equal-simbolos-proposicionales-binaria
  (implies (es-binaria F)
           (and (subsetp-equal (simbolos-proposicionales (arg1 F))
			       (simbolos-proposicionales F))
                (subsetp-equal (simbolos-proposicionales (arg2 F))
			       (simbolos-proposicionales F))))
  :hints (("Goal" :in-theory (disable definiciones-conectivas))))

(in-theory (disable otra-definicion-de-es-proposicional))

;;;============================================================================