;;;============================================================================
;;;
;;; Ttulo: Semntica de la lgica proposicional (Kleene)
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../3-logica/semantica-K")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

;;;   En este libro (semantica-K.lisp) se desarrolla la semntica de la
;;; lgica proposicional basada en la lgica trivalorada fuerte de Kleene.

(include-book "../3-logica/sintaxis")

(in-theory (disable definiciones-conectivas))

(in-theory (enable otra-definicion-de-es-proposicional))

;;;============================================================================
;;;
;;; 1) Asignaciones
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Valores de verdad
;;;
;;;----------------------------------------------------------------------------

(defconst *V* '(1))

(defconst *F* '(0))

;;;   El tercer valor de verdad es nil

(defun es-valor-de-verdad (v)
  (or (equal v *V*)
      (equal v *F*)))

(defthm valor-de-verdad-es-V-o-F
  (implies (es-valor-de-verdad v)
           (or (equal v *V*)
               (equal v *F*)))
  :rule-classes :forward-chaining)

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Valor de un smbolo
;;;
;;;----------------------------------------------------------------------------

(defun valor-de-simbolo-K (x sigma)
  (declare (xargs :guard (alistp sigma)))
  (cond ((es-simbolo-proposicional x)
	 (let ((valor (cdr (assoc-equal x sigma))))
	   (if (es-valor-de-verdad valor)
	       valor
	       nil)))
	(t nil)))

(defthm valor-de-simbolo-K-es-F-o-V
  (implies (valor-de-simbolo-K x sigma)
	   (or (equal (valor-de-simbolo-K x sigma) *F*)
	       (equal (valor-de-simbolo-K x sigma) *V*)))
  :rule-classes :forward-chaining)

(defun asume-valor (x valor sigma)
  (declare (xargs :guard (alistp sigma)))
  (acons x valor sigma))

;;;============================================================================
;;;
;;; 2) Funciones de verdad
;;;
;;;============================================================================

(defun funcion-de-verdad-de-negacion-K (i)
  (cond ((equal i *F*) *V*)
	((equal i *V*) *F*)
	(t nil)))

(defun funcion-de-verdad-de-conjuncion-K (i j)
  (cond ((equal i *F*) *F*)
	((equal i *V*) j)
	(t (cond ((equal j *F*) *F*)
		 (t nil)))))

(defun funcion-de-verdad-de-disyuncion-K (i j)
  (cond ((equal i *F*) j)
	((equal i *V*) *V*)
	(t (cond ((equal j *V*) *V*)
		 (t nil)))))

(defun funcion-de-verdad-de-implicacion-K (i j)
  (cond ((equal i *F*) *V*)
	((equal i *V*) j)
	(t (cond ((equal j *V*) *V*)
		 (t nil)))))

(defun funcion-de-verdad-de-equivalencia-K (i j)
  (cond ((equal i *F*) (cond ((equal j *F*) *V*)
			     ((equal j *V*) *F*)
			     (t nil)))
	((equal i *V*) (cond ((equal j *F*) *F*)
			     ((equal j *V*) *V*)
			     (t nil)))
	(t nil)))

(deftheory funciones-de-verdad
  '(funcion-de-verdad-de-negacion-K
    funcion-de-verdad-de-conjuncion-K
    funcion-de-verdad-de-disyuncion-K
    funcion-de-verdad-de-implicacion-K
    funcion-de-verdad-de-equivalencia-K))

;;;============================================================================
;;;
;;; 3) Valor de una frmula en una asignacin
;;;
;;;============================================================================

(defun valor-K (F sigma)
  (declare (xargs :measure (numero-conectivas F)
		  :guard (alistp sigma)))
  (cond ((es-negacion F)
	 (funcion-de-verdad-de-negacion-K
	  (valor-K (arg1 F) sigma)))
	((es-conjuncion F) 
	 (funcion-de-verdad-de-conjuncion-K
	  (valor-K (arg1 F) sigma) 
	  (valor-K (arg2 F) sigma)))
	((es-disyuncion F) 
	 (funcion-de-verdad-de-disyuncion-K
	  (valor-K (arg1 F) sigma) 
	  (valor-K (arg2 F) sigma)))
	((es-implicacion F) 
	 (funcion-de-verdad-de-implicacion-K
	  (valor-K (arg1 F) sigma) 
	  (valor-K (arg2 F) sigma)))
	((es-equivalencia F) 
	 (funcion-de-verdad-de-equivalencia-K
	  (valor-K (arg1 F) sigma) 
	  (valor-K (arg2 F) sigma)))
	(t (valor-de-simbolo-K F sigma))))

(defun modelo-K (sigma F)
  (declare (xargs :guard (alistp sigma)))
  (equal (valor-K F sigma) *V*))

(defun no-modelo-K (sigma F)
  (declare (xargs :guard (alistp sigma)))
  (equal (valor-K F sigma) *F*))
  
(defthm valor-K-es-proposicional
  (implies (valor-K F sigma)
	   (es-proposicional F)))

(defthm valor-K-es-F-o-V
  (implies (valor-K F sigma)
	   (or (equal (valor-K F sigma) *F*)
	       (equal (valor-K F sigma) *V*)))
  :rule-classes :forward-chaining)

;;;----------------------------------------------------------------------------
;;;
;;; 3.1) Complementario de una frmula
;;;
;;;----------------------------------------------------------------------------

(defun complementario (F)
  (if (es-negacion F)
      (arg1 F)
      (negacion F)))

(defthm valor-K-complementario
  (implies (es-proposicional F)
	   (equal (valor-K (complementario F) sigma)
		  (funcion-de-verdad-de-negacion-K (valor-K F sigma)))))

; (defthm modelo-K-complementario
;   (implies (and (es-proposicional F)
; 		(valor-K F sigma))
; 	   (iff (modelo-K sigma (complementario F))
; 		(not (modelo-K sigma F)))))

(defthm modelo-K-complementario
  (implies (and (es-proposicional F)
		(modelo-K sigma (complementario F)))
	   (not (modelo-K sigma F))))

(defthm complementario-no-igual-al-original
  (not (equal (complementario F) F)))

(defthm complementario-es-formula
  (implies (es-proposicional F)
           (es-proposicional (complementario F)))
  :hints (("Goal" :in-theory (disable otra-definicion-de-es-proposicional))))

;;;============================================================================