;;;============================================================================
;;;
;;; Ttulo: Semntica de la lgica proposicional
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../3-logica/semantica")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

;;;   En este libro (semantica.lisp) se desarrolla la semntica clsica de la
;;; lgica proposicional.

(include-book "../3-logica/sintaxis")

(in-theory (disable definiciones-conectivas))

(in-theory (enable otra-definicion-de-es-proposicional))

;;;============================================================================
;;;
;;; 1) Asignaciones
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Valores de verdad
;;;
;;;----------------------------------------------------------------------------

(defconst *V* '(1))

(defconst *F* '(0))

(defun es-valor-de-verdad (v)
  (or (equal v *V*)
      (equal v *F*)))

(defthm valor-de-verdad-es-1-o-0
  (implies (es-valor-de-verdad v)
           (or (equal v *V*)
               (equal v *F*)))
  :rule-classes :forward-chaining)

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Valor de un smbolo
;;;
;;;----------------------------------------------------------------------------

(defun valor-de-simbolo (x sigma)
  (declare (xargs :guard (alistp sigma)))
  (cond ((es-simbolo-proposicional x)
         (let ((valor (cdr (assoc-equal x sigma))))
           (if (es-valor-de-verdad valor)
               valor
               *F*)))
        (t *F*)))

(defun asume-valor (x valor sigma)
  (declare (xargs :guard (alistp sigma)))
  (acons x valor sigma))

;;;============================================================================
;;;
;;; 2) Funciones de verdad
;;;
;;;============================================================================

(defun funcion-de-verdad-de-negacion (i)
  (if (equal i *V*) *F* *V*))

(defun funcion-de-verdad-de-conjuncion (i j)
  (if (equal i *V*)
      (if (equal j *V*) *V* *F*) 
      *F*))

(defun funcion-de-verdad-de-disyuncion (i j)
  (if (equal i *V*)
      *V*
      (if (equal j *V*) *V* *F*)))

(defun funcion-de-verdad-de-implicacion (i j)
  (if (equal i *V*)
      (if (equal j *V*) j *F*)
      *V*))

(defun funcion-de-verdad-de-equivalencia (i j)
  (if (equal i *V*)
      (if (equal j *V*) *V* *F*)
      (if (equal j *V*) *F* *V*)))

(deftheory funciones-de-verdad
  '(funcion-de-verdad-de-negacion
    funcion-de-verdad-de-conjuncion
    funcion-de-verdad-de-disyuncion
    funcion-de-verdad-de-implicacion
    funcion-de-verdad-de-equivalencia))

;;;============================================================================
;;;
;;; 3) Valor de una frmula en una asignacin
;;;
;;;============================================================================

(defun valor (F sigma)
  (declare (xargs :measure (numero-conectivas F)
                  :guard (alistp sigma)))
  (cond ((es-negacion F)
         (funcion-de-verdad-de-negacion
          (valor (arg1 F) sigma)))
        ((es-conjuncion F) 
         (funcion-de-verdad-de-conjuncion
          (valor (arg1 F) sigma) 
          (valor (arg2 F) sigma)))
        ((es-disyuncion F) 
         (funcion-de-verdad-de-disyuncion
          (valor (arg1 F) sigma) 
          (valor (arg2 F) sigma)))
        ((es-implicacion F) 
         (funcion-de-verdad-de-implicacion
          (valor (arg1 F) sigma) 
          (valor (arg2 F) sigma)))
        ((es-equivalencia F) 
         (funcion-de-verdad-de-equivalencia
          (valor (arg1 F) sigma) 
          (valor (arg2 F) sigma)))
        (t (valor-de-simbolo F sigma))))

(defun modelo (sigma F)
  (declare (xargs :guard (alistp sigma)))
  (equal (valor F sigma) *V*))

(defthm modelo-es-proposicional
  (implies (and (modelo sigma F)
                (not (es-valor-de-verdad F)))
           (es-proposicional F)))

(defthm no-proposicional-no-modelo
  (implies (and (not (es-proposicional F))
                (not (es-valor-de-verdad F)))
           (not (modelo sigma F))))

(defthm valor-es-proposicional
  (implies (equal (valor F sigma) *V*)
	   (es-proposicional F)))

(defthm valor-es-0-o-1
  (or (equal (valor F sigma) *F*)
      (equal (valor F sigma) *V*))
  :rule-classes ((:forward-chaining
                  :trigger-terms ((valor F sigma)))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.1) Complementario de una frmula
;;;
;;;----------------------------------------------------------------------------

(defun complementario (F)
  (if (es-negacion F)
      (arg1 F)
      (negacion F)))

(defthm valor-complementario
  (implies (es-proposicional F)
	   (equal (valor (complementario F) sigma)
		  (funcion-de-verdad-de-negacion (valor F sigma)))))

(defthm modelo-complementario
  (implies (es-proposicional F)
	   (iff (modelo sigma (complementario F))
		(not (modelo sigma F)))))

(defthm complementario-no-igual-al-original
  (not (equal (complementario F) F)))

(defthm complementario-es-formula
  (implies (es-proposicional F)
           (es-proposicional (complementario F)))
  :hints (("Goal" :in-theory (disable otra-definicion-de-es-proposicional))))

;;;============================================================================
