;;;============================================================================
;;;
;;; Ttulo: Tablas de verdad (Kleene)
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../3-logica/tablas-de-verdad-K")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

(include-book "../3-logica/semantica-K")

(in-theory (disable definiciones-conectivas
		    es-simbolo-proposicional
		    es-valor-de-verdad
		    otra-definicion-de-es-proposicional))

;;;============================================================================
;;;
;;; 1) Asignaciones relevantes de una frmula 
;;;
;;;============================================================================

(defun es-lista-propia-de-alistas (alist)
  (if (atom alist)
      (equal alist nil)
      (and (alistp (car alist))
	   (es-lista-propia-de-alistas (cdr alist)))))

(defun asume-valor-lista (x v sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (if (endp sigma-lst)
      nil
      (cons (asume-valor x v (car sigma-lst))
	    (asume-valor-lista x v (cdr sigma-lst)))))

(defthm asume-valor-lista-conserva-lista-propia-de-alistas
  (implies (es-lista-propia-de-alistas sigma-lst)
	   (es-lista-propia-de-alistas (asume-valor-lista x v sigma-lst))))

(defun asignaciones (Xs)
  (declare (xargs :verify-guards nil
		  :guard (true-listp Xs)))
  (if (endp Xs)
      (list nil)
      (let ((sigma-lst (asignaciones (cdr Xs))))
	(append (asume-valor-lista (car Xs) *F* sigma-lst)
		(asume-valor-lista (car Xs) *V* sigma-lst)))))

(defthm append-conserva-lista-propia-de-alistas
  (implies (and (es-lista-propia-de-alistas alist1)
		(es-lista-propia-de-alistas alist2))
	   (es-lista-propia-de-alistas (append alist1 alist2))))

(defthm asignaciones-es-lista-propia-de-alistas
  (es-lista-propia-de-alistas (asignaciones Xs)))

(verify-guards asignaciones)

(defun asignaciones-formula (F)
  (asignaciones (simbolos-proposicionales F)))

(encapsulate
 ()

 (local
  (defun asignacion-lista (sigma Xs)
    (declare (xargs :guard (and (alistp sigma)
				(true-listp Xs))))
    (cond ((endp Xs) (endp sigma))
	  (t (and (equal (car (car sigma)) (car Xs))
		  (es-valor-de-verdad (cdr (car sigma)))
		  (asignacion-lista (cdr sigma) (cdr Xs)))))))

 (local
  (defthm asignacion-lista-asigna-valores-de-verdad
    (implies (and (member-equal x Xs)
		  (asignacion-lista sigma Xs)
		  (es-simbolo-proposicional x))
	     (valor-de-simbolo-K x sigma))))
  
 (local
  (defthm member-equal-asume-valor-lista
    (implies (member-equal sigma (asume-valor-lista x v Xs))
	     (and (equal (caar sigma) x)
		  (equal (cdar sigma) v)
		  (member-equal (cdr sigma) Xs)))))
 (local
  (defthm member-equal-append
    (iff (member-equal elemento (append lista1 lista2))
	 (or (member-equal elemento lista1)
	     (member-equal elemento lista2)))))

 (local
  (defthm member-equal-asignaciones-car
    (implies (and (consp Xs)
		  (member-equal sigma (asignaciones Xs)))
	     (and (equal (caar sigma) (car Xs))
		  (es-valor-de-verdad (cdar sigma))
		  (member-equal (cdr sigma) (asignaciones (cdr Xs)))))
    :rule-classes :forward-chaining))

 (local
  (defthm asignaciones-genera-asignacion-lista
    (implies (member-equal sigma (asignaciones Xs))
	     (asignacion-lista sigma Xs))))
 
 (local
  (defthm valor-simbolo-en-asignaciones-simbolos
    (implies (and (member-equal x Xs)
		  (member-equal sigma (asignaciones Xs))
		  (es-simbolo-proposicional x))
	     (valor-de-simbolo-K x sigma))
    :hints (("Goal" :in-theory (disable valor-de-simbolo-K)))))

 (local
  (defthm asignaciones-asigna-valor
    (implies (and (es-proposicional F)
		  (subsetp-equal (simbolos-proposicionales F) Xs)
		  (member-equal sigma (asignaciones Xs)))
	     (valor-K F sigma))
    :hints (("Goal" 
	     :induct (valor-K F sigma)
	     :in-theory (disable valor-de-simbolo-K))
	    ("Subgoal *1/6" 
	     :in-theory (enable otra-definicion-de-es-proposicional)))))

 (defthm asignaciones-asigna-valor-instance
   (implies (and (es-proposicional F)
		 (member-equal sigma (asignaciones-formula F)))
	    (valor-K F sigma))
   :hints (("Goal" :use (:instance asignaciones-asigna-valor
				   (Xs (simbolos-proposicionales F))))))

 )

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Construccin de una asignacin relevante a partir de una asignacin
;;;
;;;----------------------------------------------------------------------------

(defun asignacion-inducida-K (Xs sigma)
  (declare (xargs :guard (and (true-listp Xs)
			      (alistp sigma))))
  (if (endp Xs)
      nil
      (if (valor-de-simbolo-K (car Xs) sigma)
	  (asume-valor (car Xs)
		       (valor-de-simbolo-K (car Xs) sigma)
		       (asignacion-inducida-K (cdr Xs) sigma))
	  (asume-valor (car Xs) *F*
		       (asignacion-inducida-K (cdr Xs) sigma)))))

(encapsulate
 ()

 (local
  (defthm asignacion-inducida-K-equivale-a-la-original-simbolos
    (implies (and (not (es-negacion F))
		  (not (es-conjuncion F))
		  (not (es-disyuncion F))
		  (not (es-implicacion F))
		  (not (es-equivalencia F))
		  (valor-K F sigma)
		  (subsetp-equal (simbolos-proposicionales F) Xs))
	     (equal (valor-K F (asignacion-inducida-K Xs sigma))
		    (valor-K F sigma)))
    :hints (("Goal" :induct (asignacion-inducida-K Xs sigma)))))

 (defthm asignacion-inducida-K-equivale-a-la-original-aux
   (implies (and (valor-K F sigma)
		 (subsetp-equal (simbolos-proposicionales F) Xs))
	    (equal (valor-K F (asignacion-inducida-K Xs sigma))
		   (valor-K F sigma)))
   :hints (("Goal" :induct (valor-K F sigma))))

 )

(encapsulate
 ()

 (local
  (defthm member-equal-append
    (iff (member-equal elt (append lst1 lst2))
	 (or (member-equal elt lst1)
	     (member-equal elt lst2)))))

 (defthm asignacion-inducida-K-es-asignacion
   (member-equal (asignacion-inducida-K Xs sigma)
		 (asignaciones Xs)))

 )

;;;============================================================================
;;;
;;; 2) Decisin de validez basada en tablas de verdad
;;;
;;;============================================================================

(defun valida-en-asignaciones-K (F sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (if (endp sigma-lst)
      t
      (and (modelo-K (car sigma-lst) F)
	   (valida-en-asignaciones-K F (cdr sigma-lst)))))

(defun TAUT-tablas-de-verdad-K (F)
  (valida-en-asignaciones-K F (asignaciones-formula F)))

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Correccin de TAUT-tablas-de-verdad-K
;;;
;;;----------------------------------------------------------------------------

(defthm correccion-valida-en-asignaciones-K
  (implies (and (member-equal sigma lista)
		(valida-en-asignaciones-K F lista))
	   (modelo-K sigma F))
  :rule-classes nil
  :hints (("Goal" :hands-off (valor-K))))

(defthm correccion-TAUT-tablas-de-verdad-K
  (implies (and (es-proposicional F)
		(TAUT-tablas-de-verdad-K F)
		(valor-K F sigma))
	   (modelo-K sigma F))
  :hints (("Goal"
	   :use (:instance correccion-valida-en-asignaciones-K
			   (sigma (asignacion-inducida-K
				   (simbolos-proposicionales F)
				   sigma))
			   (lista (asignaciones-formula F))))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Correccin de TAUT-tablas-de-verdad-K
;;;
;;;----------------------------------------------------------------------------

(defun primer-no-modelo-K (F sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (if (endp sigma-lst)
      nil
      (if (no-modelo-K (car sigma-lst) F)
	  (car sigma-lst)
	  (primer-no-modelo-K F (cdr sigma-lst)))))

(defun valor-K-en-asignaciones (F sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (cond ((endp sigma-lst) t)
	(t (and (valor-K F (car sigma-lst))
		(valor-K-en-asignaciones F (cdr sigma-lst))))))

(defthm completitud-valida-en-asignaciones-K
  (implies (and (not (valida-en-asignaciones-K F sigma-lst))
		(valor-K-en-asignaciones F sigma-lst))
	   (no-modelo-K (primer-no-modelo-K F sigma-lst) F)))

(defun NO-MOD-tablas-de-verdad-K (F)
  (primer-no-modelo-K F (asignaciones-formula F)))

(include-book "../6-extensiones/listas")

(defthm valor-K-en-asignaciones-formula
  (implies (es-proposicional F)
	   (valor-K-en-asignaciones F (asignaciones-formula F)))
  :hints ((extension-conjuntiva-a-listas
	   (asignaciones-formula F)
	   (lambda (elt)
	     (valor-K F elt))
	   (lambda (lst)
	     (valor-K-en-asignaciones F lst)))))

(defthm completitud-TAUT-tablas-de-verdad-K
  (implies (and (es-proposicional F)
		(not (TAUT-tablas-de-verdad-K F)))
	   (no-modelo-K (NO-MOD-tablas-de-verdad-K F) F))
  :hints (("Goal" :in-theory (disable no-modelo-K
				      asignaciones-formula))))

;;;============================================================================
;;;
;;; 3) Decisin de satisfacibilidad basada en tablas de verdad
;;;
;;;============================================================================

(defun satisfacible-en-asignaciones-K (F sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (if (endp sigma-lst)
      nil
      (or (modelo-K (car sigma-lst) F)
	  (satisfacible-en-asignaciones-K F (cdr sigma-lst)))))

(defun SAT-tablas-de-verdad-K (F)
  (satisfacible-en-asignaciones-K F (asignaciones-formula F)))

;;;----------------------------------------------------------------------------
;;;
;;; 3.1) Correccin de SAT-tablas-de-verdad-K
;;;
;;;----------------------------------------------------------------------------

(defun primer-modelo-K (F sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (if (endp sigma-lst)
      nil
      (if (modelo-K (car sigma-lst) F)
	  (car sigma-lst)
	  (primer-modelo-K F (cdr sigma-lst)))))

(defun MOD-tablas-de-verdad-K (F)
  (primer-modelo-K F (asignaciones-formula F)))

(defthm primer-modelo-K-es-modelo-K
  (implies (primer-modelo-K F sigma-lst)
	   (modelo-K (primer-modelo-K F sigma-lst) F)))

(defthm satisfacible-en-asignacines-primer-modelo-K
  (implies (satisfacible-en-asignaciones-K F sigma-lst)
	   (modelo-K (primer-modelo-K F sigma-lst) F)))

(defthm correccion-SAT-tablas-de-verdad-K
  (implies (and (es-proposicional F)
		(SAT-tablas-de-verdad-K F))
	   (modelo-K (MOD-tablas-de-verdad-K F) F))
  :hints (("Goal" :in-theory (disable modelo-K))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.2) Completitud de SAT-tablas-de-verdad-K
;;;
;;;----------------------------------------------------------------------------

(defthm existe-modelo-K-satisfacible-en-asignaciones-K
  (implies (and (member-equal sigma sigma-lst)
		(modelo-K sigma F))
	   (satisfacible-en-asignaciones-K F sigma-lst)))

(defthm completitud-SAT-tablas-de-verdad-K
  (implies (and (es-proposicional F)
		(modelo-K sigma F))
	   (SAT-tablas-de-verdad-K F))
  :hints (("Goal" 
	   :use (:instance existe-modelo-K-satisfacible-en-asignaciones-K
			   (sigma-lst (asignaciones 
				       (simbolos-proposicionales F)))
			   (sigma (asignacion-inducida-K 
				   (simbolos-proposicionales F)
				   sigma))))))

;;;============================================================================
;;;
;;; 4) Ejemplos de uso
;;;
;;;============================================================================

#|

;;;   Un ejemplo de utilizacin de este ATP es el siguiente:

(include-book "../3-logica/tablas-de-verdad-K")

;;;   Cargamos el archivo con las frmulas de Pelletier y evaluamos los
;;; ejemplos: 

(include-book "../0-ejemplos/pelletier")

(value :q)

(load "../0-ejemplos/ejemplos.lisp")

(evalua-ejemplos 'TAUT-tablas-de-verdad-K *pelletier*)

;;;   Los resultados obtenidos fueron los siguientes:

;;; +---------------------------------------------------------------+--------+
;;; | Frmula                                                       | Tiempo |
;;; +---------------------------------------------------------------+--------+
;;; |  1) (p -> q) <-> (-q -> -p)                                   |  0.000 |
;;; |  2) (-(-p) <-> p)                                             |  0.000 |
;;; |  3) -(p -> q) -> (q -> p)                                     |  0.000 |
;;; |  4) (-p -> q) <-> (-q -> p)                                   |  0.000 |
;;; |  5) ((p | q) -> (p | r)) -> (p | (q -> r))                    |  0.000 |
;;; |  6) (p | -p)                                                  |  0.000 |
;;; |  7) (p | -(-(-p)))                                            |  0.000 |
;;; |  8) ((p -> q) -> p) -> p                                      |  0.000 |
;;; |  9) [(p | q) & (-p | q) & (p | -q)] -> -(-p | -q)             |  0.000 |
;;; | 10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q) |  0.000 |
;;; | 11) (p <-> p)                                                 |  0.000 |
;;; | 12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]                   |  0.000 |
;;; | 13) [p | (q & r)] <-> [(p | q) & (p | r)]                     |  0.010 |
;;; | 14) (p <-> q) <-> ((q | -p) & (-q | p))                       |  0.000 |
;;; | 15) (p -> q) <-> (-p | q)                                     |  0.000 |
;;; | 16) (p -> q) | (q -> p)                                       |  0.000 |
;;; | 17) ((p & (q -> r)) -> s) <-> ((-p | q | s) & (-p | -r | s))  |  0.000 |
;;; +---------------------------------------------------------------+--------+

|#

;;;============================================================================
