;;;============================================================================
;;;
;;; Ttulo: Tablas de verdad
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../3-logica/tablas-de-verdad")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

(include-book "../3-logica/semantica")

(in-theory (disable definiciones-conectivas
		    es-simbolo-proposicional
		    es-valor-de-verdad
		    otra-definicion-de-es-proposicional))

;;;============================================================================
;;;
;;; 1) Asignaciones relevantes de una frmula 
;;;
;;;============================================================================

(defun es-lista-propia-de-alistas (alist)
  (if (atom alist)
      (equal alist nil)
      (and (alistp (car alist))
	   (es-lista-propia-de-alistas (cdr alist)))))

(defun asume-valor-lista (x v sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (if (endp sigma-lst)
      nil
      (cons (asume-valor x v (car sigma-lst))
	    (asume-valor-lista x v (cdr sigma-lst)))))

(defthm asume-valor-lista-conserva-lista-propia-de-alistas
  (implies (es-lista-propia-de-alistas sigma-lst)
	   (es-lista-propia-de-alistas (asume-valor-lista x v sigma-lst))))

(defun asignaciones (Xs)
  (declare (xargs :verify-guards nil
		  :guard (true-listp Xs)))
  (if (endp Xs)
      (list nil)
      (let ((sigma-lst (asignaciones (cdr Xs))))
	(append (asume-valor-lista (car Xs) *F* sigma-lst)
		(asume-valor-lista (car Xs) *V* sigma-lst)))))

(defthm append-conserva-lista-propia-de-alistas
  (implies (and (es-lista-propia-de-alistas alist1)
		(es-lista-propia-de-alistas alist2))
	   (es-lista-propia-de-alistas (append alist1 alist2))))

(defthm asignaciones-es-lista-propia-de-alistas
  (es-lista-propia-de-alistas (asignaciones Xs)))

(verify-guards asignaciones)

(defun asignaciones-formula (F)
  (asignaciones (simbolos-proposicionales F)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Construccin de una asignacin relevante a partir de una asignacin
;;;
;;;----------------------------------------------------------------------------

(defun asignacion-inducida (Xs sigma)
  (declare (xargs :guard (and (true-listp Xs)
			      (alistp sigma))))
  (if (endp Xs)
      nil
      (asume-valor (car Xs)
		   (valor-de-simbolo (car Xs) sigma)
		   (asignacion-inducida (cdr Xs) sigma))))
 
(encapsulate
 ()

 (local
  (defthm asignacion-inducida-equivale-a-la-original-simbolos
    (implies (and (not (es-negacion F))
		  (not (es-conjuncion F))
		  (not (es-disyuncion F))
		  (not (es-implicacion F))
		  (not (es-equivalencia F))
		  (subsetp-equal (simbolos-proposicionales F)
				 Xs))
	     (equal (valor F (asignacion-inducida Xs sigma))
		    (valor F sigma)))
    :hints (("Goal" :induct (asignacion-inducida Xs sigma)))))

 (defthm asignacion-inducida-equivale-a-la-original
   (implies (subsetp-equal (simbolos-proposicionales F) Xs)
	    (equal (valor F (asignacion-inducida Xs sigma))
		   (valor F sigma)))
   :hints (("Goal" :induct (valor F sigma))))

 )

(encapsulate
 ()

 (local
  (defthm member-equal-append
    (iff (member-equal elt (append lst1 lst2))
	 (or (member-equal elt lst1)
	     (member-equal elt lst2)))))

 (defthm asignacion-inducida-es-asignacion
   (member-equal (asignacion-inducida Xs sigma)
		 (asignaciones Xs)))

 )

;;;============================================================================
;;;
;;; 2) Decisin de validez basada en tablas de verdad
;;;
;;;============================================================================

(defun valida-en-asignaciones (F sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (if (endp sigma-lst)
      t
      (and (modelo (car sigma-lst) F)
	   (valida-en-asignaciones F (cdr sigma-lst)))))

(defun TAUT-tablas-de-verdad (F)
  (valida-en-asignaciones F (asignaciones-formula F)))

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Correccin de TAUT-tablas-de-verdad
;;;
;;;----------------------------------------------------------------------------

(defthm correccion-valida-en-asignaciones
  (implies (and (member-equal sigma sigma-lst)
		(valida-en-asignaciones F sigma-lst))
	   (modelo sigma F))
  :rule-classes nil
  :hints (("Goal" :hands-off (valor))))

(defthm correccion-TAUT-tablas-de-verdad 
  (implies (and (es-proposicional F)
		(TAUT-tablas-de-verdad F))
	   (modelo sigma F))
  :hints (("Goal"
	   :use (:instance correccion-valida-en-asignaciones
			   (sigma (asignacion-inducida
					(simbolos-proposicionales F)
					sigma))
			   (sigma-lst (asignaciones-formula F))))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Correccin de TAUT-tablas-de-verdad
;;;
;;;----------------------------------------------------------------------------

(defun primer-contramodelo (F sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (if (endp sigma-lst)
      nil
      (if (not (modelo (car sigma-lst) F))
	  (car sigma-lst)
	  (primer-contramodelo F (cdr sigma-lst)))))

(defthm completitud-valida-en-asignaciones
  (implies (not (valida-en-asignaciones F sigma-lst))
	   (not (modelo (primer-contramodelo F sigma-lst) F))))

(defun NO-MOD-tablas-de-verdad (F)
  (primer-contramodelo F (asignaciones-formula F)))

(defthm completitud-TAUT-tablas-de-verdad
  (implies (and (es-proposicional F)
		(not (TAUT-tablas-de-verdad F)))
	   (not (modelo (NO-MOD-tablas-de-verdad F) F)))
  :hints (("Goal" :in-theory (disable modelo))))

;;;============================================================================
;;;
;;; 3) Decisin de satisfacibilidad basada en tablas de verdad
;;;
;;;============================================================================

(defun es-satisfacible-en-asignaciones (F sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (if (endp sigma-lst)
      nil
      (or (modelo (car sigma-lst) F)
	  (es-satisfacible-en-asignaciones F (cdr sigma-lst)))))

(defun SAT-tablas-de-verdad (F)
  (es-satisfacible-en-asignaciones F (asignaciones-formula F)))

;;;----------------------------------------------------------------------------
;;;
;;; 3.1) Correccin de SAT-tablas-de-verdad
;;;
;;;----------------------------------------------------------------------------

(defun primer-modelo (F sigma-lst)
  (declare (xargs :guard (es-lista-propia-de-alistas sigma-lst)))
  (if (endp sigma-lst)
      nil
      (if (modelo (car sigma-lst) F)
	  (car sigma-lst)
	  (primer-modelo F (cdr sigma-lst)))))

(defun MOD-tablas-de-verdad (F)
  (primer-modelo F (asignaciones-formula F)))

(defthm primer-modelo-es-modelo
  (implies (primer-modelo F sigma-lst)
	   (modelo (primer-modelo F sigma-lst) F)))

(defthm es-satisfacible-en-asignacines-primer-modelo
  (implies (es-satisfacible-en-asignaciones F sigma-lst)
	   (modelo (primer-modelo F sigma-lst) F)))

(defthm correccion-SAT-tablas-de-verdad
  (implies (and (es-proposicional F)
		(SAT-tablas-de-verdad F))
	   (modelo (MOD-tablas-de-verdad F) F))
  :hints (("Goal" :in-theory (disable modelo))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.2) Completitud de SAT-tablas-de-verdad
;;;
;;;----------------------------------------------------------------------------

(defthm existe-modelo-es-satisfacible-en-asignaciones
  (implies (and (member-equal sigma sigma-lst)
		(modelo sigma F))
	   (es-satisfacible-en-asignaciones F sigma-lst)))

(defthm completitud-es-satisfactible
  (implies (and (es-proposicional F)
		(modelo sigma F))
	   (SAT-tablas-de-verdad F))
  :hints (("Goal" 
	   :use (:instance existe-modelo-es-satisfacible-en-asignaciones
			   (sigma-lst (asignaciones 
				       (simbolos-proposicionales F)))
			   (sigma (asignacion-inducida 
				   (simbolos-proposicionales F)
				   sigma))))))

;;;============================================================================
;;;
;;; 4) Ejemplos de uso
;;;
;;;============================================================================

#|

;;;   Un ejemplo de utilizacin de este ATP es el siguiente:

(include-book "../3-logica/tablas-de-verdad")

;;;   Cargamos el archivo con las frmulas de Pelletier y evaluamos los
;;; ejemplos: 

(include-book "../0-ejemplos/pelletier")

(value :q)

(load "../0-ejemplos/ejemplos.lisp")

(evalua-ejemplos 'TAUT-tablas-de-verdad *pelletier*)

;;;   Los resultados obtenidos fueron los siguientes:

;;; +---------------------------------------------------------------+--------+
;;; | Frmula                                                       | Tiempo |
;;; +---------------------------------------------------------------+--------+
;;; |  1) (p -> q) <-> (-q -> -p)                                   |  0.000 |
;;; |  2) (-(-p) <-> p)                                             |  0.000 |
;;; |  3) -(p -> q) -> (q -> p)                                     |  0.000 |
;;; |  4) (-p -> q) <-> (-q -> p)                                   |  0.000 |
;;; |  5) ((p | q) -> (p | r)) -> (p | (q -> r))                    |  0.000 |
;;; |  6) (p | -p)                                                  |  0.000 |
;;; |  7) (p | -(-(-p)))                                            |  0.000 |
;;; |  8) ((p -> q) -> p) -> p                                      |  0.000 |
;;; |  9) [(p | q) & (-p | q) & (p | -q)] -> -(-p | -q)             |  0.000 |
;;; | 10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q) |  0.000 |
;;; | 11) (p <-> p)                                                 |  0.000 |
;;; | 12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]                   |  0.000 |
;;; | 13) [p | (q & r)] <-> [(p | q) & (p | r)]                     |  0.000 |
;;; | 14) (p <-> q) <-> ((q | -p) & (-q | p))                       |  0.000 |
;;; | 15) (p -> q) <-> (-p | q)                                     |  0.000 |
;;; | 16) (p -> q) | (q -> p)                                       |  0.000 |
;;; | 17) ((p & (q -> r)) -> s) <-> ((-p | q | s) & (-p | -r | s))  |  0.010 |
;;; +---------------------------------------------------------------+--------+

|#

;;;============================================================================
