;;;============================================================================
;;;
;;; Ttulo: Tableros semnticos (Kleene)
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../4-tableros/tableros-K")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

(include-book "../4-tableros/uniforme-K")

(in-theory (disable definiciones-conectivas
		    definicion-uniforme-de-es-proposicional))

;;;============================================================================
;;;
;;; 1) Tableros semnticos proposicionales
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Ramas de un tablero
;;;
;;;----------------------------------------------------------------------------

(defun es-rama-tablero (rama)
  (declare (xargs :guard (true-listp rama)))
  (or (endp rama)
      (and (es-proposicional (car rama))
           (es-rama-tablero (cdr rama)))))

(defthm miembro-de-es-rama-tablero-es-formula
  (implies (and (es-rama-tablero rama)
                (member-equal F rama))
           (es-proposicional F)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Semntica asociada a las ramas
;;;
;;;----------------------------------------------------------------------------

(defun modelo-rama-K (sigma rama)
  (declare (xargs :guard (and (alistp sigma)
                              (true-listp rama))))
  (cond ((endp rama) t)
	(t (and (modelo-K sigma (car rama))
		(modelo-rama-K sigma (cdr rama))))))

(defthm modelo-rama-K-append
  (iff (modelo-rama-K sigma (append rama-1 rama-2))
       (and (modelo-rama-K sigma rama-1)
	    (modelo-rama-K sigma rama-2))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Ramas cerradas
;;;
;;;----------------------------------------------------------------------------

(defun rama-cerrada (rama)
  (declare (xargs :guard (true-listp rama)))
  (cond ((endp rama) nil)
        ((member-equal (complementario (car rama)) (cdr rama)) t)
        (t (rama-cerrada (cdr rama)))))

(encapsulate
 ()

 (local
  (defthm rama-con-formula-falsa
    (implies (and (member-equal F rama)
                  (not (modelo-K sigma F))
                  (es-proposicional F))
	     (not (modelo-rama-K sigma rama)))))

 (local
  (in-theory (disable complementario)))

 (local
  (defthm rama-con-complementario-formula-verdadera
    (implies (and (member-equal (complementario F) rama)
		  (modelo-K sigma F)
                  (es-proposicional F))
	     (not (modelo-rama-K sigma rama)))))

 (defthm rama-cerrada-no-tiene-modelo-K
   (implies (and (rama-cerrada rama)
                 (es-rama-tablero rama))
            (not (modelo-rama-K sigma rama))))

 )

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Las reglas de expansin de tableros preservan los modelos
;;;
;;;----------------------------------------------------------------------------

(defun elimina-una (elt lista)
  (declare (xargs :guard (true-listp lista)))
  (cond ((endp lista) lista)
        ((equal (car lista) elt) (cdr lista))
        (t (cons (car lista) (elimina-una elt (cdr lista))))))

(defthm true-listp-elimina-una
  (implies (true-listp lista)
           (true-listp (elimina-una F lista))))

(defthm es-rama-tablero-elimina-una
  (implies (es-rama-tablero lista)
           (es-rama-tablero (elimina-una F lista))))

(defun aade-elemento (F lista)
  (declare (xargs :guard (true-listp lista)))
  (if (member-equal F lista)
      lista
      (cons F lista)))

(defthm true-listp-aade-elemento
  (implies (true-listp lista)
           (true-listp (aade-elemento F lista))))

(defthm es-rama-tablero-aade-elemento
  (iff (es-rama-tablero (aade-elemento F lista))
       (and (es-proposicional F)
            (es-rama-tablero lista))))

(defthm modelo-rama-K-elimina-una
  (implies (modelo-rama-K sigma rama)
	   (modelo-rama-K sigma (elimina-una F rama))))

(encapsulate
 ()

 (local
  (defthm modelo-rama-K-aade-elemento
    (iff (modelo-rama-K sigma (aade-elemento F rama))
	 (modelo-rama-K sigma (cons F rama)))))

 (local
  (in-theory (disable aade-elemento alfa-formula beta-formula
		      componente-1 componente-2)))

 (defthm modelo-rama-K-doble-negacion
   (implies (and (member-equal F rama)
		 (doble-negacion F))
	    (iff (modelo-rama-K sigma (aade-elemento (componente-neg-neg F)
						      (elimina-una F rama)))
		 (modelo-rama-K sigma rama))))
 
 (defthm modelo-rama-K-alfa-formula
   (implies 
    (and (member-equal F rama)
	 (alfa-formula F))
    (iff (modelo-rama-K sigma 
		      (aade-elemento (componente-1 F)
				      (aade-elemento (componente-2 F)
						      (elimina-una F rama))))
	 (modelo-rama-K sigma rama))))
 
 (defthm modelo-rama-K-beta-formula
   (implies
    (and (beta-formula F)
	 (member-equal F rama))
    (iff (modelo-rama-K sigma rama)
	 (or (modelo-rama-K sigma (aade-elemento (componente-1 F)
						(elimina-una F rama)))
	     (modelo-rama-K sigma (aade-elemento (componente-2 F)
						(elimina-una F rama)))))))
 
 )

;;;============================================================================
;;;
;;; 2) Bsqueda de un modelo de una rama: MOD-rama
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Funcin de seleccin genrica
;;;
;;;----------------------------------------------------------------------------

(encapsulate
 (((seleccion *) => *))

 (local
  (defun seleccion (rama)
    (declare (xargs :guard (true-listp rama)))
    (cond ((endp rama) nil)
          ((es-literal (car rama)) (seleccion (cdr rama)))
          (t (car rama)))))

 (local (in-theory (disable es-atomica (:type-prescription es-atomica)
                            es-literal (:type-prescription es-negacion)
                            (:type-prescription es-literal)
                            unicidad-es-negacion)))

 (defthm si-existe-no-literal-seleccion-pertenece
   (implies (seleccion rama)
            (member-equal (seleccion rama) rama)))

 (defthm si-existe-no-literal-seleccion-no-es-literal
   (implies (seleccion rama)
            (not (es-literal (seleccion rama)))))

 (defthm si-no-seleccion-miembros-literales
   (implies (and (not (seleccion rama))
                 (es-rama-tablero rama)
                 (member-equal F rama))
            (es-literal F)))

 )

;;;----------------------------------------------------------------------------
;;;
;;; 2.2) Definicin de MOD-rama
;;;
;;;----------------------------------------------------------------------------

(defthm consp-aade-elemento
  (consp (aade-elemento F lista)))

;;;   Ahora ya podemos definir la funcin que implementa el algoritmo para
;;; comprobar la satisfacibilidad:

(defun MOD-rama (rama)
  (declare (xargs :mode :program
		  :guard (true-listp rama)))

  (cond
   ((endp rama) nil)
   ((rama-cerrada rama) nil)
   (t (let ((F (seleccion rama)))
        (cond ((doble-negacion F)
               (MOD-rama
                (aade-elemento (componente-neg-neg F)
                                (elimina-una F rama)))
               )
              ((alfa-formula F)
               (MOD-rama
                (aade-elemento (componente-1 F)
                                (aade-elemento (componente-2 F)
                                                (elimina-una F rama))))
               )
              ((beta-formula F)
               (or (MOD-rama
                    (aade-elemento (componente-1 F)
                                    (elimina-una F rama)))
                   (MOD-rama
                    (aade-elemento (componente-2 F)
                                    (elimina-una F rama))))
               )
              (t rama))))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.3) Prueba de terminacin
;;;
;;;----------------------------------------------------------------------------

(defun medida-uniforme-rama (rama)
  (declare (xargs :guard (true-listp rama)))
  (cond ((endp rama) 0)
        (t (+ (medida-uniforme (car rama))
	      (medida-uniforme-rama (cdr rama))))))

(defthm medida-uniforme-rama-cons
  (equal (medida-uniforme-rama (cons F rama))
         (+ (medida-uniforme F) (medida-uniforme-rama rama))))

(in-theory (disable alfa-formula beta-formula doble-negacion
		    componente-1 componente-2 componente-neg-neg))

(defthm medida-uniforme-rama-elimina-una-formula
  (implies (member-equal F rama)
	   (= (medida-uniforme-rama rama)
	      (+ (medida-uniforme F)
		 (medida-uniforme-rama (elimina-una F rama)))))
  :rule-classes nil)

(defthm medida-uniforme-rama-elimina-una-seleccion
  (implies (seleccion rama)
	   (= (medida-uniforme-rama rama)
	      (+ (medida-uniforme (seleccion rama))
		 (medida-uniforme-rama 
		  (elimina-una (seleccion rama) rama)))))
  :hints (("Goal"
	   :use (:instance medida-uniforme-rama-elimina-una-formula
			   (F (seleccion rama))))))

(defthm compuesta-uniforme-seleccion
  (implies (or (alfa-formula (seleccion rama))
               (beta-formula (seleccion rama))
               (doble-negacion (seleccion rama)))
           (seleccion rama))
  :rule-classes ((:forward-chaining
                  :trigger-terms ((alfa-formula (seleccion rama))
                                  (beta-formula (seleccion rama))
                                  (doble-negacion (seleccion rama))))))

(verify-termination MOD-rama
  (declare (xargs :measure (medida-uniforme-rama rama))))

;;;============================================================================
;;;
;;; 3) Correccin y completitud de MOD-rama
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 3.1) Prueba de la correccin
;;;
;;;----------------------------------------------------------------------------

(defun genera-modelo-literales-K (lista-L)
  (declare (xargs :guard (true-listp lista-L)))
  (cond ((endp lista-L) nil)
        ((es-literal-positivo (car lista-L))
         (asume-valor (car lista-L) *V*
                      (genera-modelo-literales-K (cdr lista-L))))
        ((es-literal-negativo (car lista-L))
         (asume-valor (arg1 (car lista-L)) *F*
                      (genera-modelo-literales-K (cdr lista-L))))
        (t (genera-modelo-literales-K (cdr lista-L)))))

(encapsulate
 ()

 (local
  (defthm genera-modelo-literales-K-es-correcta-literal
    (implies (and (es-literal L)
		  (not (rama-cerrada rama))
		  (member-equal L rama))
	     (modelo-K (genera-modelo-literales-K rama) L))
    :hints (("Goal" 
	     :induct (member-equal L rama)))))

 (local
  (defthm caso-base-correccion-aux
    (implies (and (es-rama-tablero rama)
		  (consp rama)
		  (not (doble-negacion (seleccion rama)))
		  (not (alfa-formula (seleccion rama)))
		  (not (beta-formula (seleccion rama)))
		  (member-equal L rama))
	     (es-literal L))
    :rule-classes :forward-chaining
    :hints (("Goal"
	     :in-theory (disable es-literal)
	     :cases ((seleccion rama)))
	    ("Subgoal 1"
	     :use (:instance definicion-uniforme-de-es-proposicional
			     (F (seleccion rama)))))))

 (local
  (include-book "../6-extensiones/listas"))
 
 (local
  (in-theory (disable es-literal rama-cerrada modelo-K)))

 (defthm caso-base-correccion-MOD-rama
   (implies
    (and (consp rama)
	 (es-rama-tablero rama)
	 (not (rama-cerrada rama))
	 (not (doble-negacion (seleccion rama)))
	 (not (alfa-formula (seleccion rama)))
	 (not (beta-formula (seleccion rama))))
    (modelo-rama-K (genera-modelo-literales-K rama) rama))
   :hints ((extension-conjuntiva-a-listas
	    rama
	    (lambda (elt)
	      (modelo-K (genera-modelo-literales-K rama) elt))
	    (lambda (lst)
	      (modelo-rama-K (genera-modelo-literales-K rama) lst)))))

 )

(defthm correccion-MOD-rama
  (implies (and (es-rama-tablero rama)
		(MOD-rama rama))
	   (modelo-rama-K (genera-modelo-literales-K
			   (MOD-rama rama)) rama))
  :rule-classes nil
  :hints (("Goal" :in-theory (disable es-rama-tablero aade-elemento
                                      rama-cerrada))))

;;;----------------------------------------------------------------------------
;;;
;;; 3.2) Prueba de la completitud
;;;
;;;----------------------------------------------------------------------------

(defthm completitud-MOD-rama
  (implies (and (consp rama)
		(es-rama-tablero rama)
		(modelo-rama-K sigma rama))
	   (MOD-rama rama))
  :rule-classes nil
  :hints (("Goal" :in-theory (disable es-rama-tablero aade-elemento
                                      rama-cerrada))))

;;;============================================================================
;;;
;;; 4) El algoritmo SAT-tableros-semanticos
;;;
;;;============================================================================

(defun SAT-tableros-semanticos (F)
  (MOD-rama (list F)))

;;;----------------------------------------------------------------------------
;;;
;;; 4.1) Prueba de la correccin
;;;
;;;----------------------------------------------------------------------------

(defthm correccion-SAT-tableros-semanticos
  (implies (and (es-proposicional F)
		(SAT-tableros-semanticos F))
	   (modelo-K (genera-modelo-literales-K
		      (SAT-tableros-semanticos F)) F))
  :hints (("Goal"
	   :use (:instance correccion-MOD-rama
			   (rama (list F))))))

;;;----------------------------------------------------------------------------
;;;
;;; 4.2) Prueba de la completitud
;;;
;;;----------------------------------------------------------------------------

(defthm completitud-SAT-tableros-semanticos
  (implies (and (es-proposicional F)
		(modelo-K sigma F))
	   (SAT-tableros-semanticos F))
  :hints (("Goal"
	   :use (:instance completitud-MOD-rama
			   (rama (list F))))))

;;;============================================================================
;;;
;;; 4) El algoritmo DAT-tableros-semanticos
;;;
;;;============================================================================

(defun DAT-tableros-semanticos (F)
  (not (MOD-rama (list (negacion F)))))

;;;----------------------------------------------------------------------------
;;;
;;; 4.1) Correccin y completitud de DAT-tableros-semanticos
;;;
;;;----------------------------------------------------------------------------

(defthm correccion-DAT-tableros-semanticos
  (implies (and (es-proposicional F)
		(DAT-tableros-semanticos F)
		(valor-K F sigma))
	   (modelo-K sigma F))
  :hints (("Goal"
	   :use (:instance completitud-MOD-rama
			   (rama (list (negacion F)))))))

(defthm completitud-DAT-tableros-semanticos
  (implies (and (es-proposicional F)
		(not (DAT-tableros-semanticos F)))
	   (no-modelo-K (genera-modelo-literales-K
			 (SAT-tableros-semanticos (negacion F))) F))
  :hints (("Goal"
	   :use (:instance correccion-MOD-rama
			   (rama (list (negacion F)))))))

;;;============================================================================
;;;
;;; 6) Consecuencia por tableros
;;;
;;;============================================================================

(defun DEDUC-tableros-semanticos (lista-H F)
  (declare (xargs :guard (true-listp lista-H)))
  (not (MOD-rama (cons (negacion F) lista-H))))

(defthm DEDUC-tableros-semanticos-sin-hipotesis
  (equal (DAT-tableros-semanticos F)
         (DEDUC-tableros-semanticos nil F)))

;;;----------------------------------------------------------------------------
;;;
;;; 6.1) Prueba de la correccin
;;;
;;;----------------------------------------------------------------------------

(defthm correccion-DEDUC-tableros-semanticos
  (implies (and (es-rama-tablero lista-H)
                (es-proposicional F)
                (DEDUC-tableros-semanticos lista-H F)
		(modelo-rama-K sigma lista-H)
		(valor-K F sigma))
           (modelo-K sigma F))
  :hints (("Goal"
	   :use (:instance completitud-MOD-rama
			   (rama (cons (negacion F) lista-H))))))

;;;----------------------------------------------------------------------------
;;;
;;; 6.2) Prueba de la completitud
;;;
;;;----------------------------------------------------------------------------

(defthm completitud-consecuencia-por-tableros
  (let ((sigma (genera-modelo-literales-K
		(MOD-rama (cons (negacion F) lista-H)))))
    (implies (and (es-rama-tablero lista-H)
		  (es-proposicional F)
		  (not (DEDUC-tableros-semanticos lista-H F)))
	     (and (modelo-rama-K sigma lista-H)
		  (no-modelo-K sigma F))))
  :hints (("Goal"
	   :use (:instance correccion-MOD-rama
			   (rama (cons (negacion F) lista-H))))))

;;;============================================================================
;;;
;;; 7) Ejemplos de uso
;;;
;;;============================================================================

#|

;;;   Un ejemplo de utilizacin de este DAT es el siguiente:

(include-book "../4-tableros/tableros-K")

;;;   Necesitamos una funcin de seleccin concreta, para ello indicamos que se
;;; va a permitir redefiniciones en el sistema.

(set-ld-redefinition-action '(:warn . :overwrite) state)

(defun seleccion (lista)
  (declare (xargs :guard (true-listp lista)))
  (cond ((endp lista) nil)
        ((es-literal (car lista)) (seleccion (cdr lista)))
        (t (car lista))))

;;;   Cargamos el archivo de evaluacin de ejemplos:

(value :q)

(compile 'seleccion)

(load "../0-ejemplos/ejemplos.lisp")

;;;   Cargamos el archivo con las frmulas de Pelletier y evaluamos los
;;; ejemplos: 

(include-book "../0-ejemplos/pelletier")

(evalua-ejemplos 'DAT-tableros-semanticos *pelletier*)

;;;   Los resultados obtenidos fueron los siguientes:

;;; +---------------------------------------------------------------+--------+
;;; | Frmula                                                       | Tiempo |
;;; +---------------------------------------------------------------+--------+
;;; |  1) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  2) ((p) <-> p)                                             |  0.000 |
;;; |  3) (p -> q) -> (q -> p)                                     |  0.000 |
;;; |  4) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  5) ((p | q) -> (p | r)) -> (p | (q -> r))                    |  0.000 |
;;; |  6) (p | p)                                                  |  0.000 |
;;; |  7) (p | ((p)))                                            |  0.000 |
;;; |  8) ((p -> q) -> p) -> p                                      |  0.010 |
;;; |  9) [(p | q) & (p | q) & (p | q)] -> (p | q)             |  0.000 |
;;; | 10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q) |  0.000 |
;;; | 11) (p <-> p)                                                 |  0.000 |
;;; | 12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]                   |  0.010 |
;;; | 13) [p | (q & r)] <-> [(p | q) & (p | r)]                     |  0.000 |
;;; | 14) (p <-> q) <-> ((q | p) & (q | p))                       |  0.000 |
;;; | 15) (p -> q) <-> (p | q)                                     |  0.000 |
;;; | 16) (p -> q) | (q -> p)                                       |  0.000 |
;;; | 17) ((p & (q -> r)) -> s) <-> ((p | q | s) & (p | r | s))  |  0.000 |
;;; +---------------------------------------------------------------+--------+

;;;   Cargamos el archivo con las frmulas de Urquhart y evaluamos los
;;; ejemplos: 

(include-book "../0-ejemplos/urquhart")

;;;   Los resultados obtenidos en tiempo con las frmulas de Urquhart son los
;;; siguientes:

(evalua-ejemplos 'DAT-tableros-semanticos *urquhart* t)

;;; +=========+=====+============+
;;; | Frmula | Sol | Tiempo     |
;;; +---------+-----+------------+
;;; | U_1     | T   |      0.000 |
;;; | U_2     | T   |      0.010 |
;;; | U_3     | T   |      0.000 |
;;; | U_4     | T   |      0.020 |
;;; | U_5     | T   |      0.070 |
;;; | U_6     | T   |      0.170 |
;;; | U_7     | T   |      0.420 |
;;; | U_8     | T   |      1.080 |
;;; | U_9     | T   |      2.820 |
;;; | U_10    | T   |      6.640 |
;;; | U_11    | T   |     15.590 |
;;; | U_12    | T   |     36.650 |
;;; | U_13    | T   |     84.910 |
;;; | U_14    | T   |    197.620 |
;;; | U_15    | T   |    669.670 |
;;; +=========+=====+============+

(evalua-ejemplos 'SAT-tableros-semanticos *urquhart*)

;;; +=========+============+
;;; | Frmula | Tiempo     |
;;; +---------+------------+
;;; | U_1     |      0.000 |
;;; | U_2     |      0.010 |
;;; | U_3     |      0.000 |
;;; | U_4     |      0.000 |
;;; | U_5     |      0.000 |
;;; | U_6     |      0.000 |
;;; | U_7     |      0.010 |
;;; | U_8     |      0.010 |
;;; | U_9     |      0.010 |
;;; | U_10    |      0.000 |
;;; | U_11    |      0.010 |
;;; | U_12    |      0.010 |
;;; | U_13    |      0.010 |
;;; | U_14    |      0.030 |
;;; | U_15    |      0.020 |
;;; +=========+============+

;;;   Cargamos el archivo con las frmulas que representan el problema de las N
;;; reinas y evaluamos los ejemplos: 

(include-book "../0-ejemplos/reinas")

(evalua-ejemplos 'SAT-tableros-semanticos *n-reinas*)

;;; +==========+============+
;;; | Frmula  | Tiempo     |
;;; +----------+------------+
;;; | Reinas-2 |      0.000 |
;;; | Reinas-3 |      0.140 |
;;; | Reinas-4 |      1.270 |
;;; | Reinas-5 |      5.800 |
;;; | Reinas-6 |    525.170 |
;;; +==========+============+

|#

;;;============================================================================