;;;============================================================================
;;;
;;; Ttulo: Secuentes
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../5-secuentes/secuentes")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

(include-book "../3-logica/semantica")

(in-theory (disable definiciones-conectivas
		    otra-definicion-de-es-proposicional))

;;;============================================================================
;;;
;;; 1) El sistema G'
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Secuentes
;;;
;;;----------------------------------------------------------------------------

(defun lista-formulas (lista-F)
  (declare (xargs :guard (true-listp lista-F)))
  (or (endp lista-F)
      (and (es-proposicional (car lista-F))
	   (lista-formulas (cdr lista-F)))))

(defthm miembro-de-lista-formulas-es-formula 
  (implies (and (lista-formulas lista-F)
		(member-equal F lista-F))
	   (es-proposicional F)))

(defun es-secuente (S)
  (declare (xargs :guard (and (true-listp S)
			      (true-listp (car S)))))
  (and (consp S)
       (lista-formulas (car S))
       (lista-formulas (cdr S))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Semntica asociada a los secuentes
;;;
;;;----------------------------------------------------------------------------

(defun modelo-conjuncion (sigma Gamma)
  (declare (xargs :guard (and (alistp sigma)
			      (true-listp Gamma))))
  (cond ((endp Gamma) t)
	(t (and (modelo sigma (car Gamma))
		(modelo-conjuncion sigma (cdr Gamma))))))

(defun modelo-disyuncion (sigma Delta)
  (declare (xargs :guard (and (alistp sigma)
			      (true-listp Delta))))
  (cond ((endp Delta) nil)
	(t (or (modelo sigma (car Delta))
	       (modelo-disyuncion sigma (cdr Delta))))))

(defun modelo-secuente (sigma S)
  (declare (xargs :guard (and (alistp sigma)
			      (true-listp S)
			      (true-listp (car S)))))
  (if (modelo-conjuncion sigma (car S))
      (modelo-disyuncion sigma (cdr S))
      t))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Secuentes axioma
;;;
;;;----------------------------------------------------------------------------

(defun secuente-axioma (S)
  (declare (xargs :guard (and (true-listp S)
			      (true-listp (car S)))))
  (intersectp-equal (car S) (cdr S)))

(encapsulate
 ()

 (local
  (defun elemento-compartido (Gamma Delta)
    (declare (xargs :guard (and (true-listp Gamma)
				(true-listp Delta))))
    (cond ((endp Gamma) nil)
	  ((member-equal (car Gamma) Delta) (car Gamma))
	  (t (elemento-compartido (cdr Gamma) Delta)))))

 (local
  (defthm elemento-compartido-pertenece-ambos 
    (implies (intersectp-equal Gamma Delta)
	     (and (member-equal (elemento-compartido Gamma Delta) Gamma)
		  (member-equal (elemento-compartido Gamma Delta) Delta)))
    :rule-classes :forward-chaining))
 
 (local
  (defthm elemento-compartido-modelo-secuente
    (implies (and (member-equal F Gamma)
		  (member-equal F Delta))
	     (modelo-secuente sigma (cons Gamma Delta)))))

 (defthm secuente-axioma-es-valido
   (implies (secuente-axioma S)
	    (modelo-secuente sigma S))
   :hints (("Goal" :in-theory (disable modelo-secuente))))

 )

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Reglas de inferencia
;;;
;;;----------------------------------------------------------------------------

(defun elimina-una (elt lista)
  (declare (xargs :guard (true-listp lista)))
  (cond ((endp lista) lista)
        ((equal (car lista) elt) (cdr lista))
	(t (cons (car lista) (elimina-una elt (cdr lista))))))

(defthm true-listp-elimina-una
  (implies (true-listp Gamma)
	   (true-listp (elimina-una F Gamma))))

(defthm lista-formulas-elimina-una
  (implies (lista-formulas Gamma)
	   (lista-formulas (elimina-una F Gamma))))

(defun aade-elemento (F Gamma)
  (declare (xargs :guard (true-listp Gamma)))
  (if (member-equal F Gamma)
      Gamma
      (cons F Gamma)))

(defthm lista-formulas-aade-elemento
  (iff (lista-formulas (aade-elemento F Gamma))
       (and (es-proposicional F)
	    (lista-formulas Gamma))))

(defun secuentes-expansion-izq (F Gamma Delta)
  (declare (xargs :guard (and (true-listp Gamma)
			      (true-listp Delta))))
  (let ((Gamma- (elimina-una F Gamma)))
    (cond ((es-negacion F)
	   (list
	    (cons Gamma- (aade-elemento (arg1 F) Delta))))
	  ((es-conjuncion F)
	   (list
	    (cons (aade-elemento (arg1 F)
				  (aade-elemento (arg2 F) Gamma-)) Delta)))
	  ((es-disyuncion F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma-) Delta)
	    (cons (aade-elemento (arg2 F) Gamma-) Delta)))
	  ((es-implicacion F)
	   (list
	    (cons Gamma- (aade-elemento (arg1 F) Delta))
	    (cons (aade-elemento (arg2 F) Gamma-) Delta)))
	  ((es-equivalencia F)
	   (list
	    (cons (aade-elemento (arg1 F)
				  (aade-elemento (arg2 F) Gamma-)) Delta)
	    (cons Gamma-
		  (aade-elemento (arg1 F)
				  (aade-elemento (arg2 F) Delta)))))
	  (t nil))))

(defthm es-secuente-secuentes-expansion-izq-1
  (implies 
   (and (member-equal F Gamma)
	(es-secuente (cons Gamma Delta))
	(equal (len (secuentes-expansion-izq F Gamma Delta)) 1))
   (es-secuente (car (secuentes-expansion-izq F Gamma Delta)))))

(defthm es-secuente-secuentes-expansion-izq-2
  (implies 
   (and (member-equal F Gamma)
	(es-secuente (cons Gamma Delta))
	(equal (len (secuentes-expansion-izq F Gamma Delta)) 2))
   (and (es-secuente (car (secuentes-expansion-izq F Gamma Delta)))
	(es-secuente (cadr (secuentes-expansion-izq F Gamma Delta))))))

(defun secuentes-expansion-der (F Gamma Delta)
  (declare (xargs :guard (and (true-listp Gamma)
			      (true-listp Delta))))
  (let ((Delta- (elimina-una F Delta)))
    (cond ((es-negacion F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma) Delta-)))
	  ((es-conjuncion F)
	   (list
	    (cons Gamma (aade-elemento (arg1 F) Delta-))
	    (cons Gamma (aade-elemento (arg2 F) Delta-))))
	  ((es-disyuncion F)
	   (list
	    (cons Gamma (aade-elemento (arg1 F) 
					(aade-elemento (arg2 F) Delta-)))))
	  ((es-implicacion F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma)
		  (aade-elemento (arg2 F) Delta-))))
	  ((es-equivalencia F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma)
		  (aade-elemento (arg2 F) Delta-))
	    (cons (aade-elemento (arg2 F) Gamma)
		  (aade-elemento (arg1 F) Delta-))))
	  (t nil))))

(defthm es-secuente-secuentes-expansion-der-1
  (implies 
   (and (member-equal F Delta)
	(es-secuente (cons Gamma Delta))
	(equal (len (secuentes-expansion-der F Gamma Delta)) 1))
   (es-secuente (car (secuentes-expansion-der F Gamma Delta)))))

(defthm es-secuente-secuentes-expansion-der-2
  (implies 
   (and (member-equal F Delta)
	(es-secuente (cons Gamma Delta))
	(equal (len (secuentes-expansion-der F Gamma Delta)) 2))
   (and (es-secuente (car (secuentes-expansion-der F Gamma Delta)))
	(es-secuente (cadr (secuentes-expansion-der F Gamma Delta))))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.5) Las reglas de inferencia preservan la validez
;;;
;;;----------------------------------------------------------------------------

(defthm modelo-conjuncion-aade-elemento
  (iff (modelo-conjuncion sigma (aade-elemento F Gamma))
       (and (modelo sigma F)
	    (modelo-conjuncion sigma Gamma))))

(defthm modelo-conjuncion-elimina-una
  (implies (member-equal F Gamma)
	   (iff (modelo-conjuncion sigma Gamma)
		(and (modelo-conjuncion sigma (elimina-una F Gamma))
		     (modelo sigma F)))))

(defthm modelo-disyuncion-aade-elemento
  (iff (modelo-disyuncion sigma (aade-elemento F Gamma))
       (or (modelo sigma F)
	   (modelo-disyuncion sigma Gamma))))

(defthm modelo-disyuncion-elimina-una
  (implies (member-equal F Delta)
	   (iff (modelo-disyuncion sigma Delta)
		(or (modelo-disyuncion sigma (elimina-una F Delta))
		    (modelo sigma F)))))

(defthm modelo-expansion-secuentes-izq-1
  (implies
   (and (equal (len (secuentes-expansion-izq F Gamma Delta)) 1)
	(es-proposicional F)
	(not (es-atomica F))
	(member-equal F Gamma))
   (let ((Premisas (secuentes-expansion-izq F Gamma Delta)))
     (iff (modelo-secuente sigma (car Premisas))
	  (modelo-secuente sigma (cons Gamma Delta)))))
  :hints (("Goal" :in-theory (disable aade-elemento es-atomica))))

(defthm modelo-expansion-secuentes-izq-2
  (implies
   (and (equal (len (secuentes-expansion-izq F Gamma Delta)) 2)
	(es-proposicional F)
	(not (es-atomica F))
	(member-equal F Gamma))
   (let ((Premisas (secuentes-expansion-izq F Gamma Delta)))
     (iff (modelo-secuente sigma (cons Gamma Delta))
	  (and (modelo-secuente sigma (car Premisas))
	       (modelo-secuente sigma (cadr Premisas))))))
  :hints (("Goal" :in-theory (disable aade-elemento es-atomica))))

(defthm modelo-expansion-secuentes-der-1
  (implies
   (and (equal (len (secuentes-expansion-der F Gamma Delta)) 1)
	(es-proposicional F)
	(not (es-atomica F))
	(member-equal F Delta))
   (let ((Premisas (secuentes-expansion-der F Gamma Delta)))
     (iff (modelo-secuente sigma (car Premisas))
	  (modelo-secuente sigma (cons Gamma Delta)))))
  :hints (("Goal" :in-theory (disable aade-elemento es-atomica))))

(defthm modelo-expansion-secuentes-der-2
  (implies
   (and (equal (len (secuentes-expansion-der F Gamma Delta)) 2)
	(es-proposicional F)
	(not (es-atomica F))
	(member-equal F Delta))
   (let ((Premisas (secuentes-expansion-der F Gamma Delta)))
     (iff (modelo-secuente sigma (cons Gamma Delta))
	  (and (modelo-secuente sigma (car Premisas))
	       (modelo-secuente sigma (cadr Premisas))))))
  :hints (("Goal" :in-theory (disable aade-elemento es-atomica))))

;;;============================================================================
;;;
;;; 2) Bsqueda de un contramodelo de un secuente: CONTRAMOD-secuentes
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 2.1) Funcin de seleccin genrica
;;;
;;;----------------------------------------------------------------------------

(encapsulate
 (((seleccion *) => *))

 (local
  (defun seleccion (Gamma)
    (declare (xargs :guard (true-listp Gamma)))
    (cond ((endp Gamma) nil)
	  ((es-atomica (car Gamma)) (seleccion (cdr Gamma)))
	  (t (car Gamma)))))

 (local (in-theory (disable es-atomica (:type-prescription es-atomica)
			    (:type-prescription es-negacion)
			    unicidad-es-negacion)))

 (defthm si-existe-no-atomica-seleccion-pertenece
   (implies (seleccion Gamma)
	    (member-equal (seleccion Gamma) Gamma)))
 
 (defthm si-existe-no-atomica-seleccion-no-es-atomica
   (implies (seleccion Gamma)
	    (not (es-atomica (seleccion Gamma)))))
 
 (defthm si-no-seleccion-miembros-atomicos
   (implies (and (not (seleccion Gamma))
		 (lista-formulas Gamma)
		 (member-equal F Gamma))
	    (es-atomica F)))
 
 )

;;;----------------------------------------------------------------------------
;;;
;;; 2.2) Definicin de CONTRAMOD-secuentes
;;;
;;;----------------------------------------------------------------------------

(defthm consp-aade-elemento
  (consp (aade-elemento F lista)))

(defun CONTRAMOD-secuentes (S)
  (declare (xargs :guard (and (true-listp S)
			      (true-listp (car S)))
		  :mode :program))

  (cond 
   ((endp S) nil)
   ((secuente-axioma S) nil)
   (t (let* ((Gamma (car S))
	     (Delta (cdr S))
	     (F-1 (seleccion Gamma))
	     (F-2 (seleccion Delta)))
	(cond
	 (F-1 (let ((Premisas (secuentes-expansion-izq F-1 Gamma Delta)))
		(cond ((equal (len Premisas) 1)
		       (CONTRAMOD-secuentes (car Premisas)))
		      ((equal (len Premisas) 2)
		       (or (CONTRAMOD-secuentes (car Premisas))
			   (CONTRAMOD-secuentes (cadr Premisas))))
		      (t S))))
	 (F-2 (let ((Premisas (secuentes-expansion-der F-2 Gamma Delta)))
		(cond ((equal (len Premisas) 1)
		       (CONTRAMOD-secuentes (car Premisas)))
		      ((equal (len Premisas) 2)
		       (or (CONTRAMOD-secuentes (car Premisas))
			   (CONTRAMOD-secuentes (cadr Premisas))))
		      (t S))))
	 (t S))))))

;;;----------------------------------------------------------------------------
;;;
;;; 2.3) Prueba de terminacin
;;;
;;;----------------------------------------------------------------------------

(defun medida-lista-formulas (Gamma)
  (declare (xargs :guard (true-listp Gamma)))
  (cond ((endp Gamma) 0)
	(t (+ (numero-conectivas (car Gamma))
	      (medida-lista-formulas (cdr Gamma))))))

(defun medida-secuente (S)
  (declare (xargs :guard (and (true-listp S)
			      (true-listp (car S)))))
  (+ (medida-lista-formulas (car S))
     (medida-lista-formulas (cdr S))))

(defthm medida-lista-formulas-cons
  (= (medida-lista-formulas (cons F Gamma))
     (+ (numero-conectivas F)
	(medida-lista-formulas Gamma))))

(defthm medida-lista-formulas-elimina-una-formula
  (implies (member-equal F Gamma)
	   (= (medida-lista-formulas Gamma)
	      (+ (numero-conectivas F)
		 (medida-lista-formulas (elimina-una F Gamma)))))
  :rule-classes nil)

(defthm medida-lista-formulas-elimina-una-seleccion
  (implies (seleccion Gamma)
	   (= (medida-lista-formulas Gamma)
	      (+ (numero-conectivas (seleccion Gamma))
		 (medida-lista-formulas 
		  (elimina-una (seleccion Gamma) Gamma)))))
  :hints (("Goal"
	   :use (:instance medida-lista-formulas-elimina-una-formula
			   (F (seleccion Gamma))))))

(verify-termination CONTRAMOD-secuentes
  (declare (xargs :measure (medida-secuente S))))

;;;============================================================================
;;;
;;; 3) Correccin y completitud de CONTRAMOD-secuentes
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 3.1) Prueba de la correcin
;;;
;;;----------------------------------------------------------------------------

(defun genera-asignacion-positiva (atomos)
  (declare (xargs :guard (true-listp atomos)))

  (cond ((endp atomos) nil)
	((es-simbolo-proposicional (car atomos))
	 (asume-valor (car atomos) *V*
		      (genera-asignacion-positiva (cdr atomos))))
	(t (genera-asignacion-positiva (cdr atomos)))))

(defthm genera-asignacion-positiva-es-correcta
  (implies (and (es-simbolo-proposicional F)
		(member-equal F lista))
	   (modelo (genera-asignacion-positiva lista) F)))

(defun genera-contramodelo-secuente (S)
  (declare (xargs :guard (and (true-listp S)
			      (true-listp (car S)))))
  (genera-asignacion-positiva (car S)))

(in-theory (disable es-simbolo-proposicional))

(defthm modelo-append-1
  (implies (and (es-simbolo-proposicional F)
		(modelo sigma-1 F))
	   (modelo (append sigma-1 sigma-2) F)))

(defthm genera-contramodelo-secuente-es-correcta-izq
  (implies (and (not (secuente-axioma S))
		(es-secuente S)
		(member-equal F (car S))
		(es-atomica F))
	   (modelo (genera-contramodelo-secuente S) F))
  :hints (("Goal" :in-theory (disable modelo))))

(defthm interseccion-not-member
  (implies (and (not (intersectp-equal lista1 lista2))
		(member-equal F lista2))
	   (not (member-equal F lista1))))

(defthm modelo-append-2
  (implies (and (es-simbolo-proposicional F)
		(not (modelo sigma-1 F))
		(not (modelo sigma-2 F)))
	   (not (modelo (append sigma-1 sigma-2) F))))

(defthm genera-asignacion-positiva-es-correcta-2
  (implies (and (es-simbolo-proposicional F)
		(not (member-equal F lista)))
	   (not (modelo (genera-asignacion-positiva lista) F))))

(defthm genera-contramodelo-secuente-es-correcta-der
  (implies (and (not (secuente-axioma S))
		(es-secuente S)
		(member-equal F (cdr S))
		(es-atomica F))
	   (not (modelo (genera-contramodelo-secuente S) F)))
  :hints (("Goal" :in-theory (disable modelo))))

(in-theory (disable genera-contramodelo-secuente modelo))

(include-book "../6-extensiones/listas")

(defthm genera-contramodelo-secuente-modelo-conjuncion-lista
  (implies (and (es-secuente S)
		(not (secuente-axioma S))
		(not (seleccion (car S))))
	   (modelo-conjuncion (genera-contramodelo-secuente S) (car S)))
  :hints ((extension-conjuntiva-a-listas
	   (car S)
	   (lambda (elt)
	     (modelo (genera-contramodelo-secuente S) elt))
	   (lambda (lst)
	     (modelo-conjuncion (genera-contramodelo-secuente S) lst)))))

(defthm genera-contramodelo-secuente-modelo-disyuncion-lista
  (implies (and (es-secuente S)
		(not (secuente-axioma S))
		(not (seleccion (cdr S))))
	   (not (modelo-disyuncion (genera-contramodelo-secuente S) (cdr S))))
  :hints ((extension-conjuntiva-a-listas
	   (cdr S)
	   (lambda (elt)
	     (not (modelo (genera-contramodelo-secuente S) elt)))
	   (lambda (lst)
	     (not (modelo-disyuncion (genera-contramodelo-secuente S) lst))))))

(defthm genera-contramodelo-secuente-es-correcta
  (implies (and (es-secuente S)
		(not (secuente-axioma S))
		(not (seleccion (car S)))
		(not (seleccion (cdr S))))
	   (not (modelo-secuente (genera-contramodelo-secuente S) S))))

(defthm secuentes-expansion-der-existe
  (implies
   (and (lista-formulas Delta)
	(member-equal F Delta)
	(not (es-atomica F))
	(not (equal (len (secuentes-expansion-der F Gamma Delta)) 1)))
   (equal (len (secuentes-expansion-der F Gamma Delta)) 2))
  :hints (("Goal" :in-theory (enable otra-definicion-de-es-proposicional))))

(defthm secuentes-expansion-izq-existe
  (implies
   (and (lista-formulas Gamma)
	(member-equal F Gamma)
	(not (es-atomica F))
	(not (equal (len (secuentes-expansion-izq F Gamma Delta)) 1)))
   (equal (len (secuentes-expansion-izq F Gamma Delta)) 2))
  :hints (("goal" :in-theory (enable otra-definicion-de-es-proposicional))))

(defthm seleccion-miembro-lista-formulas-es-formula
  (implies (and (lista-formulas lista)
		(seleccion lista))
	   (es-proposicional (seleccion lista))))

(defthm correccion-CONTRAMOD-secuentes
  (implies (and (es-secuente S)
		(CONTRAMOD-secuentes S))
	   (not (modelo-secuente 
		 (genera-contramodelo-secuente (CONTRAMOD-secuentes S)) S)))
  :hints (("Goal" :in-theory (disable secuentes-expansion-izq
				      secuentes-expansion-der
				      modelo modelo-secuente len)))
  :rule-classes nil)

;;;----------------------------------------------------------------------------
;;;
;;; 3.2) Prueba de la completitud
;;;
;;;----------------------------------------------------------------------------

(defthm completitud-CONTRAMOD-secuentes
  (implies (and (es-secuente S)
		(not (modelo-secuente sigma S)))
	   (CONTRAMOD-secuentes S))
  :hints (("Goal" :in-theory (disable secuentes-expansion-izq
				      secuentes-expansion-der
				      modelo modelo-secuente len)))
  :rule-classes nil)

;;;============================================================================
;;;
;;; 4) El algoritmo DAT-secuentes
;;;
;;;============================================================================

(defun DAT-secuentes (F)
  (not (CONTRAMOD-secuentes (cons nil (list F)))))

;;;----------------------------------------------------------------------------
;;;
;;; 4.1) Prueba de la correccin
;;;
;;;----------------------------------------------------------------------------

(defthm correccion-DAT-secuentes
  (implies (and (es-proposicional F)
		(DAT-secuentes F))
	   (modelo sigma F))
  :hints (("Goal"
	   :use (:instance completitud-CONTRAMOD-secuentes
			   (S (cons nil (list F)))))))

;;;----------------------------------------------------------------------------
;;;
;;; 4.2) Prueba de la completitud
;;;
;;;----------------------------------------------------------------------------

(defthm completitud-DAT-secuentes
  (implies (and (es-proposicional F)
		(not (DAT-secuentes F)))
	   (not (modelo (genera-contramodelo-secuente
			 (CONTRAMOD-secuentes (cons nil (list F)))) F)))
  :hints (("Goal"
	   :use (:instance correccion-CONTRAMOD-secuentes
			   (S (cons nil (list F)))))))

;;;============================================================================
;;;
;;; 5) El algoritmo SAT-secuentes
;;;
;;;============================================================================

(defun SAT-secuentes (F)
  (CONTRAMOD-secuentes (cons (list F) nil)))

;;;----------------------------------------------------------------------------
;;;
;;; 5.1) Prueba de la correccin
;;;
;;;----------------------------------------------------------------------------

(defthm correccion-SAT-secuentes
  (implies (and (es-proposicional F)
		(SAT-secuentes F))
	   (modelo (genera-contramodelo-secuente (SAT-secuentes F)) F))
  :hints (("Goal"
	   :use (:instance correccion-CONTRAMOD-secuentes
			   (S (cons (list F) nil))))))

;;;----------------------------------------------------------------------------
;;;
;;; 5.2) Prueba de la completitud
;;;
;;;----------------------------------------------------------------------------

(defthm completitud-SAT-secuentes
  (implies (and (es-proposicional F)
		(modelo sigma F))
	   (SAT-secuentes F))
  :hints (("Goal"
	   :use (:instance completitud-CONTRAMOD-secuentes
			   (S (cons (list F) nil))))))

;;;============================================================================
;;;
;;; 6) Consecuencia por secuentes
;;;
;;;============================================================================

(defun DEDUC-secuentes (lista-H F)
  (declare (xargs :guard (and (true-listp lista-H))))
  (not (CONTRAMOD-secuentes (cons lista-H (list F)))))

(defthm DEDUC-secuentes-sin-hipotesis
  (equal (DAT-secuentes F)
	 (DEDUC-secuentes nil F)))

;;;----------------------------------------------------------------------------
;;;
;;; 6.1) Prueba de la correccin
;;;
;;;----------------------------------------------------------------------------

(defthm correccion-DEDUC-secuentes
  (implies (and (lista-formulas lista-H)
		(es-proposicional F)
		(DEDUC-secuentes lista-H F)
		(modelo-conjuncion sigma lista-H))
	   (modelo sigma F))
  :hints (("Goal"
	   :in-theory (enable modelo)
	   :use (:instance completitud-CONTRAMOD-secuentes
			   (S (cons lista-H (list F)))))))

;;;----------------------------------------------------------------------------
;;;
;;; 6.2) Prueba de la completitud
;;;
;;;----------------------------------------------------------------------------

(defthm completitud-DEDUC-secuentes
  (let ((sigma (genera-contramodelo-secuente
		(CONTRAMOD-secuentes (cons lista-H (list F))))))
    (implies (and (lista-formulas lista-H)
		  (es-proposicional F)
		  (not (DEDUC-secuentes lista-H F)))
	     (and (modelo-conjuncion sigma lista-H)
		  (not (modelo sigma F)))))
  :hints (("Goal"
	   :in-theory (enable modelo)
	   :use (:instance correccion-CONTRAMOD-secuentes
			   (S (cons lista-H (list F)))))))

;;;============================================================================
;;;
;;; 7) Ejemplos de uso
;;;
;;;============================================================================

#|

;;;   Un ejemplo de utilizacin de este DAT es el siguiente:

(include-book "../5-secuentes/secuentes")

;;;   Necesitamos una funcin de seleccin concreta, para ello indicamos que se
;;; va a permitir redefiniciones en el sistema.

(set-ld-redefinition-action '(:warn . :overwrite) state)

(defun seleccion (lista)
  (declare (xargs :guard (true-listp lista)))
  (cond ((endp lista) nil)
	((es-atomica (car lista)) (seleccion (cdr lista)))
	(t (car lista))))

;;;   Cargamos el archivo de evaluacin de ejemplos:

(value :q)

(compile 'seleccion)

(load "../0-ejemplos/ejemplos.lisp")

;;;   Cargamos el archivo con las frmulas de Pelletier y evaluamos los
;;; ejemplos: 

(include-book "../0-ejemplos/pelletier")

(evalua-ejemplos 'DAT-secuentes *pelletier*)

;;;   Los resultados obtenidos fueron los siguientes:

;;; +---------------------------------------------------------------+--------+
;;; | Frmula                                                       | Tiempo |
;;; +---------------------------------------------------------------+--------+
;;; |  1) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  2) ((p) <-> p)                                             |  0.000 |
;;; |  3) (p -> q) -> (q -> p)                                     |  0.000 |
;;; |  4) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  5) ((p | q) -> (p | r)) -> (p | (q -> r))                    |  0.000 |
;;; |  6) (p | p)                                                  |  0.000 |
;;; |  7) (p | ((p)))                                            |  0.000 |
;;; |  8) ((p -> q) -> p) -> p                                      |  0.000 |
;;; |  9) [(p | q) & (p | q) & (p | q)] -> (p | q)             |  0.000 |
;;; | 10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q) |  0.000 |
;;; | 11) (p <-> p)                                                 |  0.000 |
;;; | 12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]                   |  0.000 |
;;; | 13) [p | (q & r)] <-> [(p | q) & (p | r)]                     |  0.000 |
;;; | 14) (p <-> q) <-> ((q | p) & (q | p))                       |  0.000 |
;;; | 15) (p -> q) <-> (p | q)                                     |  0.000 |
;;; | 16) (p -> q) | (q -> p)                                       |  0.000 |
;;; | 17) ((p & (q -> r)) -> s) <-> ((p | q | s) & (p | r | s))  |  0.000 |
;;; +---------------------------------------------------------------+--------+

;;;   Cargamos el archivo con las frmulas de Urquhart y evaluamos los
;;; ejemplos: 

(include-book "../0-ejemplos/urquhart")

(evalua-ejemplos 'DAT-secuentes *urquhart* t)

;;;   Los resultados obtenidos en tiempo con las frmulas de Urquhart son los
;;; siguientes:

;;; +=========+=====+============+
;;; | Frmula | Sol | Tiempo     |
;;; +---------+-----+------------+
;;; | U_1     | T   |      0.000 |
;;; | U_2     | T   |      0.000 |
;;; | U_3     | T   |      0.000 |
;;; | U_4     | T   |      0.000 |
;;; | U_5     | T   |      0.010 |
;;; | U_6     | T   |      0.010 |
;;; | U_7     | T   |      0.050 |
;;; | U_8     | T   |      0.110 |
;;; | U_9     | T   |      0.230 |
;;; | U_10    | T   |      0.560 |
;;; | U_11    | T   |      1.380 |
;;; | U_12    | T   |      3.170 |
;;; | U_13    | T   |      7.370 |
;;; | U_14    | T   |     16.650 |
;;; | U_15    | T   |     37.490 |
;;; +=========+=====+============+

(evalua-ejemplos 'SAT-secuentes *urquhart*)

;;; +=========+============+
;;; | Frmula | Tiempo     |
;;; +---------+------------+
;;; | U_1     |      0.000 |
;;; | U_2     |      0.000 |
;;; | U_3     |      0.000 |
;;; | U_4     |      0.000 |
;;; | U_5     |      0.000 |
;;; | U_6     |      0.000 |
;;; | U_7     |      0.000 |
;;; | U_8     |      0.000 |
;;; | U_9     |      0.010 |
;;; | U_10    |      0.000 |
;;; | U_11    |      0.000 |
;;; | U_12    |      0.000 |
;;; | U_13    |      0.000 |
;;; | U_14    |      0.010 |
;;; | U_15    |      0.010 |
;;; +=========+============+

;;;   Cargamos el archivo con las frmulas que representan el problema de las N
;;; reinas y evaluamos los ejemplos: 

(include-book "../0-ejemplos/reinas")

(evalua-ejemplos 'SAT-secuentes *n-reinas*)

;;; +==========+============+
;;; | Frmula  | Tiempo     |
;;; +----------+------------+
;;; | Reinas-2 |      0.000 |
;;; | Reinas-3 |      0.050 |
;;; | Reinas-4 |      0.420 |
;;; | Reinas-5 |      1.900 |
;;; | Reinas-6 |    170.450 |
;;; | Reinas-7 |   1037.340 |
;;; +==========+============+

|#

;;;============================================================================
