;;;============================================================================
;;;
;;; Ttulo: Un STP exitoso basado en secuentes
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(in-package "ACL2")

(certify-book "../7-generico/SAT-secuentes-K")

|#

(in-package "ACL2")

;(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

;;;   SAT proposicional basado en reglas asociadas a secuentes generado como
;;; una instancia del modelo genrico desarrollado en "SAT-generico".

(include-book "../7-generico/SAT-generico-K")

;;;============================================================================
;;;
;;; 1) Objetos proposicionales (sec-objeto)
;;;
;;;============================================================================

(defun lista-formulas (lista-F)
  (or (endp lista-F)
      (and (es-proposicional (car lista-F))
	   (lista-formulas (cdr lista-F)))))

(defthm miembro-de-lista-formulas-es-formula 
  (implies (and (lista-formulas lista-F)
		(member-equal F lista-F))
	   (es-proposicional F)))

(defun sec-objeto (S) ; = es-secuente
  (and (consp S)
       (lista-formulas (car S))
       (lista-formulas (cdr S))))

;;;============================================================================
;;;
;;; 2) Asignaciones distinguidas (sec-asignacion-distinguida)
;;;
;;;============================================================================

(defun modelo-conjuncion-K (sigma Gamma)
  (cond ((endp Gamma) t)
	(t (and (modelo-K sigma (car Gamma))
		(modelo-conjuncion-K sigma (cdr Gamma))))))

(defun no-modelo-conjuncion-K (sigma Delta)
  (cond ((endp Delta) t)
	(t (and (no-modelo-K sigma (car Delta))
		(no-modelo-conjuncion-K sigma (cdr Delta))))))

(defun sec-asignacion-distinguida-K (sigma S) ; = no-modelo-secuente-K
  (and (modelo-conjuncion-K sigma (car S))
       (no-modelo-conjuncion-K sigma (cdr S))))

;;;============================================================================
;;;
;;; 3) Regla de computacin (sec-regla-computacion)
;;;
;;;============================================================================

;;; Secuentes cerrados

(defun secuente-axioma (S)
  (intersectp-equal (car S) (cdr S)))

;;; Seleccin de regla a aplicar

(encapsulate
 (((seleccion *) => *))

 (local
  (defun seleccion (Gamma)
    (declare (xargs :guard (true-listp Gamma)))
    (cond ((endp Gamma) nil)
	  ((es-atomica (car Gamma)) (seleccion (cdr Gamma)))
	  (t (car Gamma)))))

 (local (in-theory (disable es-atomica (:type-prescription es-atomica)
			    (:type-prescription es-negacion)
			    unicidad-es-negacion)))

 (defthm si-existe-no-atomica-seleccion-pertenece
   (implies (seleccion Gamma)
	    (member-equal (seleccion Gamma) Gamma)))
 
 (defthm si-existe-no-atomica-seleccion-no-es-atomica
   (implies (seleccion Gamma)
	    (not (es-atomica (seleccion Gamma)))))
 
 (defthm si-no-seleccion-miembros-atomicos
   (implies (and (not (seleccion Gamma))
		 (lista-formulas Gamma)
		 (member-equal F Gamma))
	    (es-atomica F)))
 
 )

;;;   Funcin para aadir un elemento a una lista, si es que no estaba:

(defun aade-elemento (F Gamma)
  (if (member-equal F Gamma)
      Gamma
      (cons F Gamma)))

;;;   Finalmente podemos construir la funcin de expansin basada en las reglas
;;; anteriores:

(defun secuentes-expansion-izq (F Gamma Delta)
  (let ((Gamma- (elimina-una F Gamma)))
    (cond ((es-negacion F)
	   (list
	    (cons Gamma- (aade-elemento (arg1 F) Delta))))
	  ((es-conjuncion F)
	   (list
	    (cons (aade-elemento (arg1 F)
				  (aade-elemento (arg2 F) Gamma-)) Delta)))
	  ((es-disyuncion F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma-) Delta)
	    (cons (aade-elemento (arg2 F) Gamma-) Delta)))
	  ((es-implicacion F)
	   (list
	    (cons Gamma- (aade-elemento (arg1 F) Delta))
	    (cons (aade-elemento (arg2 F) Gamma-) Delta)))
	  ((es-equivalencia F)
	   (list
	    (cons (aade-elemento (arg1 F)
				  (aade-elemento (arg2 F) Gamma-)) Delta)
	    (cons Gamma-
		  (aade-elemento (arg1 F)
				  (aade-elemento (arg2 F) Delta)))))
	  (t nil))))

(defun secuentes-expansion-der (F Gamma Delta)
  (let ((Delta- (elimina-una F Delta)))
    (cond ((es-negacion F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma) Delta-)))
	  ((es-conjuncion F)
	   (list
	    (cons Gamma (aade-elemento (arg1 F) Delta-))
	    (cons Gamma (aade-elemento (arg2 F) Delta-))))
	  ((es-disyuncion F)
	   (list
	    (cons Gamma (aade-elemento (arg1 F) 
					(aade-elemento (arg2 F) Delta-)))))
	  ((es-implicacion F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma)
		  (aade-elemento (arg2 F) Delta-))))
	  ((es-equivalencia F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma)
		  (aade-elemento (arg2 F) Delta-))
	    (cons (aade-elemento (arg2 F) Gamma)
		  (aade-elemento (arg1 F) Delta-))))
	  (t nil))))

(defun sec-regla-computacion (S)
  (let ((Gamma (car S))
	(Delta (cdr S)))
    (if (secuente-axioma S)
	nil
	(let ((F-1 (seleccion Gamma)))
	  (if F-1
	      (secuentes-expansion-izq F-1 Gamma Delta)
	      (let ((F-2 (seleccion Delta)))
		(if F-2
		    (secuentes-expansion-der F-2 Gamma Delta)
		    t)))))))

;;;   La regla de computacin mantiene la estructura de los objetos:

(in-theory (disable definiciones-conectivas
		    es-proposicional
		    otra-definicion-de-es-proposicional))

(encapsulate
 ()

 (local
  (defthm lista-formulas-aade-elemento
    (iff (lista-formulas (aade-elemento F Gamma))
	 (and (es-proposicional F)
	      (lista-formulas Gamma)))))

 (local
  (defthm lista-formulas-elimina-una
    (implies (lista-formulas Gamma)
	     (lista-formulas (elimina-una F Gamma)))))

 (defthm member-sec-regla-computacion-es-sec-objeto
   (implies (and (sec-objeto S-1)
		 (member-equal S-2 (sec-regla-computacion S-1)))
	    (sec-objeto S-2))
   :hints (("Goal" :in-theory (disable aade-elemento))))

 )

;;;============================================================================
;;;
;;; 4) Funcin de representacin (sec-representacion)
;;;
;;;============================================================================

(defun sec-representacion (F)
  (list (list F)))

(defthm sec-representacion-es-sec-objeto
  (implies (es-proposicional formula)
	   (sec-objeto (sec-representacion formula))))

;;;============================================================================
;;;
;;; 5) Funcin de medida (sec-medida)
;;;
;;;============================================================================

(defun medida-lista-formulas (Gamma)
  (cond ((endp Gamma) 0)
	(t (+ (numero-conectivas (car Gamma))
	      (medida-lista-formulas (cdr Gamma))))))

(defun sec-medida (S)
  (+ (medida-lista-formulas (car S))
     (medida-lista-formulas (cdr S))))

(defthm sec-medida-es-e0-ordinalp
  (e0-ordinalp (sec-medida S)))

;;;============================================================================
;;;
;;; 6) Funcin modelo (sec-modelo)
;;;
;;;============================================================================

(defun genera-asignacion-positiva-K (Xs)
  (cond ((endp Xs) nil)
	((es-simbolo-proposicional (car Xs))
	 (asume-valor (car Xs) *V*
		      (genera-asignacion-positiva-K (cdr Xs))))
	(t (genera-asignacion-positiva-K (cdr Xs)))))

(defthm genera-asignacion-positiva-K-es-correcta
  (implies (and (es-simbolo-proposicional F)
		(member-equal F lista))
	   (modelo-K (genera-asignacion-positiva-K lista) F)))

(defun genera-asignacion-negativa-K (Xs)
  (cond ((endp Xs) nil)
	((es-simbolo-proposicional (car Xs))
	 (asume-valor (car Xs) *F*
		      (genera-asignacion-negativa-K (cdr Xs))))
	(t (genera-asignacion-negativa-K (cdr Xs)))))

(defthm genera-asignacion-negativa-K-es-correcta
  (implies (and (es-simbolo-proposicional F)
		(member-equal F lista))
	   (no-modelo-K (genera-asignacion-negativa-K lista) F)))

(defun sec-modelo-K (S) ; = genera-no-modelo-secuente-K
  (cond ((sec-objeto S)
	 (append (genera-asignacion-positiva-K (car S))
		 (genera-asignacion-negativa-K (cdr S))))
	(t nil)))
 
;;;============================================================================
;;;
;;; 7) Propiedad P1
;;;
;;;============================================================================

(encapsulate
 ()

 (local
  (defthm medida-lista-formulas-elimina-una-formula
    (implies (member-equal F Gamma)
	     (= (medida-lista-formulas Gamma)
		(+ (numero-conectivas F)
		   (medida-lista-formulas (elimina-una F Gamma)))))
    :rule-classes nil))
  
 (local
  (defthm medida-lista-formulas-elimina-una-seleccion
    (implies (seleccion Gamma)
	     (= (medida-lista-formulas Gamma)
		(+ (numero-conectivas (seleccion Gamma))
		   (medida-lista-formulas 
		    (elimina-una (seleccion Gamma) Gamma)))))
    :hints (("Goal"
	     :use (:instance medida-lista-formulas-elimina-una-formula
			     (F (seleccion Gamma)))))))

 (defthm P1-sec
   (implies (member-equal S-2 (sec-regla-computacion S-1))
	    (e0-ord-< (sec-medida S-2)
		      (sec-medida S-1))))

 )

;;;============================================================================
;;;
;;; 8) Propiedad P2
;;;
;;;============================================================================

(defthm P2-sec
  (implies (es-proposicional F)
	   (iff (sec-asignacion-distinguida-K sigma (sec-representacion F))
		(modelo-K sigma F))))

;;;============================================================================
;;;
;;; 9) Propiedad P3
;;;
;;;============================================================================

;;;   La propiedad fundamental de los secuentes cerrados es que no tienen
;;; modelos: 

(encapsulate
 ()

 (local
  (defun elemento-compartido (Gamma Delta)
    (cond ((endp Gamma) nil)
	  ((member-equal (car Gamma) Delta) (car Gamma))
	  (t (elemento-compartido (cdr Gamma) Delta)))))

 (local
  (defthm elemento-compartido-pertenece-ambos 
    (implies (intersectp-equal Gamma Delta)
	     (and (member-equal (elemento-compartido Gamma Delta) Gamma)
		  (member-equal (elemento-compartido Gamma Delta) Delta)))
    :rule-classes :forward-chaining))
 
 (local
  (defthm elemento-compartido-modelo-secuente
    (implies (and (member-equal F Gamma)
		  (member-equal F Delta))
	     (not (sec-asignacion-distinguida-K sigma (cons Gamma Delta))))))

 (defthm secuente-axioma-no-tiene-no-modelos
   (implies (secuente-axioma S)
	    (not (sec-asignacion-distinguida-K sigma S)))
   :hints (("Goal" :in-theory (disable sec-asignacion-distinguida-K))))

 )

;;;   La regla de computacin preserva los modelos.

(encapsulate
 ()

 (local
  (defthm modelo-conjuncion-K-aade-elemento
    (iff (modelo-conjuncion-K sigma (aade-elemento F Gamma))
	 (and (modelo-K sigma F)
	      (modelo-conjuncion-K sigma Gamma))))
  )

 (local
  (defthm modelo-conjuncion-K-elimina-una
    (implies (member-equal F Gamma)
	     (iff (modelo-conjuncion-K sigma Gamma)
		  (and (modelo-conjuncion-K sigma (elimina-una F Gamma))
		       (modelo-K sigma F)))))
  )

 (local
  (defthm no-modelo-conjuncion-K-aade-elemento
    (iff (no-modelo-conjuncion-K sigma (aade-elemento F Gamma))
	 (and (no-modelo-K sigma F)
	      (no-modelo-conjuncion-K sigma Gamma))))
  )

 (local
  (defthm no-modelo-conjuncion-K-elimina-una
    (implies (member-equal F Gamma)
	     (iff (no-modelo-conjuncion-K sigma Gamma)
		  (and (no-modelo-conjuncion-K sigma (elimina-una F Gamma))
		       (no-modelo-K sigma F)))))
  )

 (defun sec-asignacion-distinguida-K-lista (sigma lista-S)
   (cond ((endp lista-S) nil)
	 (t (or (sec-asignacion-distinguida-K sigma (car lista-S))
		(sec-asignacion-distinguida-K-lista sigma (cdr lista-S))))))
 
 (local
  (defthm sec-asignacion-distinguida-secuentes-expansion-izq
    (implies (and (es-proposicional F)
		  (not (es-atomica F))
		  (member-equal F Gamma))
	     (iff (sec-asignacion-distinguida-K-lista 
		   sigma (secuentes-expansion-izq F Gamma Delta))
		  (sec-asignacion-distinguida-K sigma (cons Gamma Delta))))
    :hints (("Goal" :in-theory (disable aade-elemento es-atomica))))
  )
  
 (local
  (defthm sec-asignacion-distinguida-secuentes-expansion-der
    (implies (and (es-proposicional F)
		  (not (es-atomica F))
		  (member-equal F Delta))
	     (iff (sec-asignacion-distinguida-K-lista 
		   sigma (secuentes-expansion-der F Gamma Delta))
		  (sec-asignacion-distinguida-K sigma (cons Gamma Delta))))
    :hints (("Goal" :in-theory (disable aade-elemento es-atomica))))
  )

 (local
  (defthm seleccion-miembro-lista-formulas-es-formula
    (implies (and (lista-formulas lista)
		  (seleccion lista))
	     (es-proposicional (seleccion lista))))
  )

 (defthm P3-sec
   (implies (and (sec-objeto S)
		 (not (equal (sec-regla-computacion S) t)))
	    (iff (sec-asignacion-distinguida-K-lista 
		  sigma (sec-regla-computacion S))
		 (sec-asignacion-distinguida-K sigma S)))
   :hints (("Goal" :in-theory (disable aade-elemento
				       secuentes-expansion-izq
				       secuentes-expansion-der
				       secuente-axioma
				       sec-asignacion-distinguida-K))))

 )

;;;============================================================================
;;;
;;; 10) Propiedad P4
;;;
;;;============================================================================

(encapsulate
 ()

 (local
  (defthm sec-modelo-asigna-modelo
    (implies (and (member-equal F Gamma)
		  (not (secuente-axioma (cons Gamma Delta)))
		  (es-simbolo-proposicional F))
	     (modelo-K (append (genera-asignacion-positiva-K Gamma)
			       (genera-asignacion-negativa-K Delta)) F))))

 (local
  (defthm sec-modelo-asigna-no-modelo
    (implies (and (member-equal F Delta)
		  (not (secuente-axioma (cons Gamma Delta)))
		  (es-simbolo-proposicional F))
	     (no-modelo-K (append (genera-asignacion-positiva-K Gamma)
				  (genera-asignacion-negativa-K Delta)) F))))
  
 (local
  (include-book "../6-extensiones/listas"))

 (local (in-theory (disable modelo-K no-modelo-K)))

 (local
  (defthm modelo-conjuncion-K-funciona-aux
    (implies (and (lista-formulas Gamma)
		  (not (seleccion Gamma))
		  (not (secuente-axioma (cons Gamma Delta)))
		  (member-equal F Gamma))
	     (modelo-K (append (genera-asignacion-positiva-K Gamma)
			       (genera-asignacion-negativa-K Delta)) F))
    :hints (("Goal" :use (:instance si-no-seleccion-miembros-atomicos)))))

 (local
  (defthm modelo-conjuncion-K-funciona
    (implies (and (lista-formulas Gamma)
		  (not (seleccion Gamma))
		  (not (secuente-axioma (cons Gamma Delta))))
	     (modelo-conjuncion-K 
	      (append (genera-asignacion-positiva-K Gamma)
		      (genera-asignacion-negativa-K Delta))
	      Gamma))
    :hints ((extension-conjuntiva-a-listas
	     Gamma
	     (lambda (F)
	       (modelo-K (append (genera-asignacion-positiva-K Gamma)
				 (genera-asignacion-negativa-K Delta)) F))
	     (lambda (lista-F)
	       (modelo-conjuncion-K 
		(append (genera-asignacion-positiva-K Gamma)
			(genera-asignacion-negativa-K Delta))
		lista-F))))))

 (local
  (defthm no-modelo-conjuncion-K-funciona-aux
    (implies (and (lista-formulas Delta)
		  (not (seleccion Delta))
		  (not (secuente-axioma (cons Gamma Delta)))
		  (member-equal F Delta))
	     (no-modelo-K (append (genera-asignacion-positiva-K Gamma)
				  (genera-asignacion-negativa-K Delta)) F))
    :hints (("Goal" :use (:instance si-no-seleccion-miembros-atomicos
				    (Gamma Delta))))))
  
 (local
  (defthm no-modelo-conjuncion-K-funciona
    (implies (and (lista-formulas Delta)
		  (not (seleccion Delta))
		  (not (secuente-axioma (cons Gamma Delta))))
	     (no-modelo-conjuncion-K 
	      (append (genera-asignacion-positiva-K Gamma)
		      (genera-asignacion-negativa-K Delta))
	      Delta))
    :hints ((extension-conjuntiva-a-listas
	       Delta
	       (lambda (F)
		 (no-modelo-K 
		  (append (genera-asignacion-positiva-K Gamma)
			  (genera-asignacion-negativa-K Delta)) F))
	       (lambda (lista-F)
		 (no-modelo-conjuncion-K 
		  (append (genera-asignacion-positiva-K Gamma)
			  (genera-asignacion-negativa-K Delta))
		  lista-F))))))

 (defthm P4-sec
   (implies (and (sec-objeto S)
		 (equal (sec-regla-computacion S) t))
	    (sec-asignacion-distinguida-K (sec-modelo-K S) S)))

 )

;;;============================================================================
;;;
;;; 11) Instancia
;;;
;;;============================================================================

;;;   A continuacin instanciamos la teora (definiciones y teoremas) del
;;; demostrador sat genrico con la siguiente correspondencia entre las
;;; funciones:

;;;   gen-objeto                       -- sec-objeto
;;;   gen-representacion               -- sec-representacion
;;;   gen-asignacion-distinguida       -- sec-asignacion-distinguida
;;;   gen-asignacion-distinguida-lista -- sec-asignacion-distinguida-lista
;;;   gen-regla-computacion            -- sec-regla-computacion
;;;   gen-seleccion                    -- car
;;;   gen-medida                       -- sec-medida
;;;   gen-modelo                       -- sec-modelo

;;;   Falta por demostrar la propiedad asociada a la funcin obj-seleccion,
;;; que aplicada a un lista de ramas no vaca devuelva un elemento de dicha
;;; lista.  No es necesario establecerla explcitamente dado la simplicidad de
;;; la funcin que usamos en su lugar.

(in-theory (disable sec-regla-computacion
		    (:type-prescription SEC-REGLA-COMPUTACION)
		    sec-representacion
		    sec-medida
		    sec-asignacion-distinguida-K
		    sec-modelo-K
		    sec-objeto))

(definstance-*sat-generico*
  ((gen-objeto                       sec-objeto)
   (gen-representacion               sec-representacion)
   (gen-asignacion-distinguida       sec-asignacion-distinguida-K)
   (gen-asignacion-distinguida-lista sec-asignacion-distinguida-K-lista)
   (gen-regla-computacion            sec-regla-computacion)
   (gen-seleccion                    car)
   (gen-medida                       sec-medida)
   (gen-modelo                       sec-modelo-K))
  "-secuentes")

;;;============================================================================
;;;
;;; 12) Ejemplos
;;;
;;;============================================================================

#|

;;;   Un ejemplo de utilizacin de este ATP es el siguiente:

;;;   Necesitamos una funcin de seleccin concreta, para ello indicamos que se
;;; va a permitir redefiniciones en el sistema.

(set-ld-redefinition-action '(:warn . :overwrite) state)

(defun seleccion (lista)
  (declare (xargs :guard (true-listp lista)))
  (cond ((endp lista) nil)
	((es-atomica (car lista)) (seleccion (cdr lista)))
	(t (car lista))))

;;;   Cargamos el archivo de evaluacin de ejemplos:

(value :q)

(compile 'seleccion)

(load "../0-ejemplos/ejemplos.lisp")

;;;   Cargamos el archivo con las frmulas de Pelletier y evaluamos los
;;; ejemplos: 

(lp)

(include-book "../0-ejemplos/pelletier")

(value :q)

(evalua-ejemplos 'atp-generico-secuentes *pelletier*)

;;;   Los resultados obtenidos fueron los siguientes:

;;; +---------------------------------------------------------------+--------+
;;; | Frmula                                                       | Tiempo |
;;; +---------------------------------------------------------------+--------+
;;; |  1) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  2) ((p) <-> p)                                             |  0.000 |
;;; |  3) (p -> q) -> (q -> p)                                     |  0.000 |
;;; |  4) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  5) ((p | q) -> (p | r)) -> (p | (q -> r))                    |  0.000 |
;;; |  6) (p | p)                                                  |  0.000 |
;;; |  7) (p | ((p)))                                            |  0.000 |
;;; |  8) ((p -> q) -> p) -> p                                      |  0.000 |
;;; |  9) [(p | q) & (p | q) & (p | q)] -> (p | q)             |  0.010 |
;;; | 10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q) |  0.000 |
;;; | 11) (p <-> p)                                                 |  0.000 |
;;; | 12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]                   |  0.010 |
;;; | 13) [p | (q & r)] <-> [(p | q) & (p | r)]                     |  0.000 |
;;; | 14) (p <-> q) <-> ((q | p) & (q | p))                       |  0.000 |
;;; | 15) (p -> q) <-> (p | q)                                     |  0.000 |
;;; | 16) (p -> q) | (q -> p)                                       |  0.000 |
;;; | 17) ((p & (q -> r)) -> s) <-> ((p | q | s) & (p | r | s))  |  0.010 |
;;; +---------------------------------------------------------------+--------+

;;;   Cargamos el archivo con las frmulas de Urquhart y evaluamos los
;;; ejemplos: 

(lp)

(include-book "../0-ejemplos/urquhart")

(value :q)

(evalua-ejemplos 'sat-generico-secuentes *urquhart-v*)

;;;   Los resultados obtenidos en tiempo con las frmulas de Urquhart son los
;;; siguientes:

;;; +---------+------------+
;;; | Frmula | Tiempo     |
;;; +---------+------------+
;;; | U_1     |      0.000 |
;;; | U_2     |      0.000 |
;;; | U_3     |      0.000 |
;;; | U_4     |      0.000 |
;;; | U_5     |      0.000 |
;;; | U_6     |      0.000 |
;;; | U_7     |      0.000 |
;;; | U_8     |      0.000 |
;;; | U_9     |      0.000 |
;;; | U_10    |      0.000 |
;;; | U_11    |      0.000 |
;;; | U_12    |      0.000 |
;;; | U_13    |      0.000 |
;;; | U_14    |      0.000 |
;;; | U_15    |      0.010 |
;;; | U_16    |      0.000 |
;;; | U_17    |      0.000 |
;;; | U_18    |      0.000 |
;;; | U_19    |      0.000 |
;;; | U_20    |      0.010 |
;;; +---------+------------+

(evalua-ejemplos 'sat-generico-secuentes *urquhart-f* t)

;;; +---------+-----+------------+
;;; | Frmula | Sol | Tiempo     |
;;; +---------+-----+------------+
;;; |  U_1   | NIL |      0.000 |
;;; |  U_2   | NIL |      0.000 |
;;; |  U_3   | NIL |      0.000 |
;;; |  U_4   | NIL |      0.000 |
;;; |  U_5   | NIL |      0.010 |
;;; |  U_6   | NIL |      0.000 |
;;; |  U_7   | NIL |      0.020 |
;;; |  U_8   | NIL |      0.030 |
;;; |  U_9   | NIL |      0.100 |
;;; |  U_10  | NIL |      0.220 |
;;; |  U_11  | NIL |      0.510 |
;;; |  U_12  | NIL |      1.300 |
;;; |  U_13  | NIL |      2.950 |
;;; |  U_14  | NIL |      6.830 |
;;; |  U_15  | NIL |     15.220 |
;;; |  U_16  | NIL |     34.210 |
;;; |  U_17  | NIL |     76.610 |
;;; |  U_18  | NIL |    170.380 |
;;; |  U_19  | NIL |    376.120 |
;;; |  U_20  | NIL |    828.710 |
;;; +---------+-----+------------+

|#

;;;============================================================================
