;;;============================================================================
;;;
;;; Ttulo: Un STP exitoso basado en secuentes
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(in-package "ACL2")

(certify-book "../7-generico/SAT-secuentes")

|#

(in-package "ACL2")

;(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

;;;   SAT proposicional basado en reglas asociadas a secuentes generado como
;;; una instancia del modelo genrico desarrollado en "SAT-generico".

(include-book "../7-generico/SAT-generico")

;;;============================================================================
;;;
;;; 1) Objetos proposicionales (sec-objeto)
;;;
;;;============================================================================

(defun lista-formulas (Gamma)
  (cond ((endp Gamma) t)
	(t (and (es-proposicional (car Gamma))
		(lista-formulas (cdr Gamma))))))

(defthm lista-formulas-member
  (implies (and (lista-formulas Gamma)
		(member-equal F Gamma))
	   (es-proposicional F)))

(defthm lista-formulas-append
  (iff (lista-formulas (append Gamma-1 Gamma-2))
       (and (lista-formulas Gamma-1)
	    (lista-formulas Gamma-2))))

(defun sec-objeto (S)
  (lista-formulas (append (car S) (cdr S))))

;;;============================================================================
;;;
;;; 2) Asignaciones distinguidas (sec-asignacion-distinguida)
;;;
;;;============================================================================

(defun modelo-lista-formulas (sigma Gamma)
  (cond ((endp Gamma) t)
	((modelo sigma (car Gamma))
	 (modelo-lista-formulas sigma (cdr Gamma)))
	(t nil)))

(defun contramodelo-lista-formulas (sigma Gamma)
  (cond ((endp Gamma) t)
	((not (modelo sigma (car Gamma)))
	 (contramodelo-lista-formulas sigma (cdr Gamma)))
	(t nil)))

(defun sec-asignacion-distinguida (sigma S)
  (and (modelo-lista-formulas sigma (car S))
       (contramodelo-lista-formulas sigma (cdr S))))

;;;============================================================================
;;;
;;; 3) Regla de computacin (sec-regla-computacion)
;;;
;;;============================================================================

;;; Secuentes cerrados

(defun interseccion (lista-1 lista-2)
  (cond ((endp lista-1) nil)
	((member-equal (car lista-1) lista-2) t)
	(t (interseccion (cdr lista-1) lista-2))))

;;; Seleccin de regla a aplicar

(encapsulate
 (((seleccion *) => *))

 (local
  (defun seleccion (Gamma)
    (cond ((endp Gamma) nil)
	  ((es-atomica (car Gamma)) (seleccion (cdr Gamma)))
	  (t (car Gamma)))))

 (local (in-theory (disable es-atomica 
			    (:type-prescription es-atomica)
			    (:type-prescription es-negacion)
			    unicidad-es-negacion)))

 (defthm si-existe-no-atomica-seleccion-pertenece
   (implies (seleccion Gamma)
	    (member-equal (seleccion Gamma) Gamma)))
 
 (defthm seleccion-es-compuesta
   (implies (and (lista-formulas Gamma)
		 (seleccion Gamma))
	    (or (es-negacion (seleccion Gamma))
		(es-conjuncion (seleccion Gamma))
		(es-disyuncion (seleccion Gamma))
		(es-implicacion (seleccion Gamma))
		(es-equivalencia (seleccion Gamma))))
   :rule-classes :forward-chaining)

 (defthm si-no-seleccion-miembros-atomicos
   (implies (and (lista-formulas Gamma)
		 (not (seleccion Gamma))
		 (member-equal F Gamma))
	    (es-atomica F))
   :rule-classes :forward-chaining)

 )

;;;   Funcin para aadir un elemento a una lista, si es que no estaba:

(defun aade-elemento (elt lista)
  (if (member-equal elt lista)
      lista
      (cons elt lista)))

;;;   Finalmente podemos construir la funcin de expansin basada en las reglas
;;; anteriores:

(defun sec-regla-computacion-izq (F Gamma Delta)
  (let ((Gamma- (elimina-una F Gamma)))
    (cond ((es-negacion F)
	   (list
	    (cons Gamma- (aade-elemento (arg1 F) Delta))))
	  ((es-conjuncion F)
	   (list
	    (cons (aade-elemento (arg1 F)
				  (aade-elemento (arg2 F) Gamma-)) Delta)))
	  ((es-disyuncion F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma-) Delta)
	    (cons (aade-elemento (arg2 F) Gamma-) Delta)))
	  ((es-implicacion F)
	   (list
	    (cons Gamma- (aade-elemento (arg1 F) Delta))
	    (cons (aade-elemento (arg2 F) Gamma-) Delta)))
	  ((es-equivalencia F)
	   (list
	    (cons (aade-elemento (arg1 F)
				  (aade-elemento (arg2 F) Gamma-)) Delta)
	    (cons Gamma-
		  (aade-elemento (arg1 F)
				  (aade-elemento (arg2 F) Delta)))))
	  (t nil))))

(defun sec-regla-computacion-der (F Gamma Delta)
  (let ((Delta- (elimina-una F Delta)))
    (cond ((es-negacion F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma) Delta-)))
	  ((es-conjuncion F)
	   (list
	    (cons Gamma (aade-elemento (arg1 F) Delta-))
	    (cons Gamma (aade-elemento (arg2 F) Delta-))))
	  ((es-disyuncion F)
	   (list
	    (cons Gamma (aade-elemento (arg1 F) 
				      (aade-elemento (arg2 F) Delta-)))))
	  ((es-implicacion F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma)
		  (aade-elemento (arg2 F) Delta-))))
	  ((es-equivalencia F)
	   (list
	    (cons (aade-elemento (arg1 F) Gamma)
		  (aade-elemento (arg2 F) Delta-))
	    (cons (aade-elemento (arg2 F) Gamma)
		  (aade-elemento (arg1 F) Delta-))))
	  (t nil))))

(defun sec-regla-computacion (S)
  (let ((Gamma (car S))
	(Delta (cdr S)))
    (if (interseccion Gamma Delta)
	nil
	(let ((F-1 (seleccion Gamma)))
	  (if F-1
	      (sec-regla-computacion-izq F-1 Gamma Delta)
	      (let ((F-2 (seleccion Delta)))
		(if F-2
		    (sec-regla-computacion-der F-2 Gamma Delta)
		    t)))))))

;;;   La regla de computacin mantiene la estructura de los objetos:

(in-theory (disable definiciones-conectivas
		    es-proposicional
		    otra-definicion-de-es-proposicional))

(encapsulate
 ()

 (local
  (defthm lista-formulas-aade-elemento
    (iff (lista-formulas (aade-elemento F Gamma))
	 (and (es-proposicional F)
	      (lista-formulas Gamma)))))

 (local
  (defthm lista-formulas-elimina-una
    (implies (lista-formulas Gamma)
	     (lista-formulas (elimina-una F Gamma)))))

 (defthm member-sec-regla-computacion-es-sec-objeto
   (implies (and (sec-objeto S-1)
		 (member-equal S-2 (sec-regla-computacion S-1)))
	    (sec-objeto S-2))
   :hints (("Goal" :in-theory (disable aade-elemento))))

 )

;;;============================================================================
;;;
;;; 4) Funcin de representacin (sec-representacion)
;;;
;;;============================================================================

(defun sec-representacion (F)
  (list (list F)))

(defthm sec-representacion-es-sec-objeto
  (implies (es-proposicional formula)
	   (sec-objeto (sec-representacion formula))))

;;;============================================================================
;;;
;;; 5) Funcin de medida (sec-medida)
;;;
;;;============================================================================

(defun medida-lista-formulas (Gamma)
  (cond ((endp Gamma) 0)
	(t (+ (numero-conectivas (car Gamma))
	      (medida-lista-formulas (cdr Gamma))))))

(defun sec-medida (S)
  (+ (medida-lista-formulas (car S))
     (medida-lista-formulas (cdr S))))

(defthm sec-medida-es-e0-ordinalp
  (e0-ordinalp (sec-medida S)))

;;;============================================================================
;;;
;;; 6) Funcin modelo (sec-modelo)
;;;
;;;============================================================================

(defun asignacion-positiva (lista)
  (cond ((endp lista) nil)
	((es-simbolo-proposicional (car lista))
	 (asume-valor (car lista) *V* (asignacion-positiva (cdr lista))))
	(t (asignacion-positiva (cdr lista)))))
 
(defun asignacion-negativa (lista)
  (cond ((endp lista) nil)
	((es-simbolo-proposicional (car lista))
	 (asume-valor (car lista) *F* (asignacion-negativa (cdr lista))))
	(t (asignacion-negativa (cdr lista)))))

(defun sec-modelo (S)
  (append (asignacion-positiva (car S))
	  (asignacion-negativa (cdr S))))
 
;;;============================================================================
;;;
;;; 7) Propiedad P1
;;;
;;;============================================================================

(encapsulate
 ()

 (local
  (defthm medida-lista-formulas-elimina-una-formula
    (implies (member-equal F Gamma)
	     (= (medida-lista-formulas Gamma)
		(+ (numero-conectivas F)
		   (medida-lista-formulas (elimina-una F Gamma)))))
    :rule-classes nil))
  
 (local
  (defthm medida-lista-formulas-elimina-una-seleccion
    (implies (seleccion Gamma)
	     (= (medida-lista-formulas Gamma)
		(+ (numero-conectivas (seleccion Gamma))
		   (medida-lista-formulas 
		    (elimina-una (seleccion Gamma) Gamma)))))
    :hints (("Goal"
	     :use (:instance medida-lista-formulas-elimina-una-formula
			     (F (seleccion Gamma)))))))

 (defthm P1-sec
   (implies (member-equal S-2 (sec-regla-computacion S-1))
	    (e0-ord-< (sec-medida S-2)
		      (sec-medida S-1))))

 )

;;;============================================================================
;;;
;;; 8) Propiedad P2
;;;
;;;============================================================================

(defthm P2-sec
  (implies (es-proposicional F)
	   (iff (sec-asignacion-distinguida sigma (sec-representacion F))
		(modelo sigma F))))

;;;============================================================================
;;;
;;; 9) Propiedad P3
;;;
;;;============================================================================

;;;   La propiedad fundamental de los secuentes cerrados es que no tienen
;;; modelos: 

(encapsulate
 ()

 (local
  (defun elemento-compartido (lista-1 lista-2)
    (cond ((endp lista-1) nil)
	  ((member-equal (car lista-1) lista-2) (car lista-1))
	  (t (elemento-compartido (cdr lista-1) lista-2)))))

 (local
  (defthm elemento-compartido-pertenece-ambos 
    (implies (interseccion lista-1 lista-2)
	     (and (member-equal (elemento-compartido lista-1 lista-2) 
				lista-1)
		  (member-equal (elemento-compartido lista-1 lista-2) 
				lista-2)))
    :rule-classes :forward-chaining))

 (local
  (defthm modelo-lista-formulas-member
    (implies (and (modelo-lista-formulas sigma Gamma)
		  (member-equal F Gamma))
	     (modelo sigma F))))

 (local
  (defthm contramodelo-lista-formulas-member
    (implies (and (member-equal F Gamma)
		  (modelo sigma F))
	     (not (contramodelo-lista-formulas sigma Gamma)))))

 (defthm interseccion-not-sec-asignacion-distinguida
   (implies (and (interseccion (car S) (cdr S))
		 (modelo-lista-formulas sigma (car S)))
	    (not (contramodelo-lista-formulas sigma (cdr S)))))

 )

;;;   La regla de computacin preserva los modelos.

(encapsulate
 ()

 (local
  (defthm modelo-lista-formulas-aade-elemento
    (iff (modelo-lista-formulas sigma (aade-elemento F Gamma))
	 (and (modelo sigma F)
	      (modelo-lista-formulas sigma Gamma))))
  )

 (local
  (defthm modelo-lista-formulas-elimina-una-1
    (implies (modelo-lista-formulas sigma Gamma)
	     (modelo-lista-formulas sigma (elimina-una F Gamma))))
  )

 (local
  (defthm modelo-lista-formulas-elimina-una-2
    (implies (and (modelo-lista-formulas sigma (elimina-una F Gamma))
		  (member-equal F Gamma)
		  (modelo sigma F))
	     (modelo-lista-formulas sigma Gamma)))
  )

 (local
  (defthm contramodelo-lista-formulas-aade-elemento
    (iff (contramodelo-lista-formulas sigma (aade-elemento F Gamma))
	 (and (not (modelo sigma F))
	      (contramodelo-lista-formulas sigma Gamma))))
  )

 (local
  (defthm contramodelo-lista-formulas-elimina-una-1
    (implies (contramodelo-lista-formulas sigma Gamma)
	     (contramodelo-lista-formulas sigma (elimina-una F Gamma))))
  )

 (local
  (defthm contramodelo-lista-formulas-elimina-una-2
    (implies (and (contramodelo-lista-formulas sigma (elimina-una F Gamma))
		  (member-equal F Gamma)
		  (not (modelo sigma F)))
	     (contramodelo-lista-formulas sigma Gamma)))
  )

 (local
  (defthm modelo-lista-formulas-member
    (implies (and (not (modelo sigma F))
		  (member-equal F Gamma))
	     (not (modelo-lista-formulas sigma Gamma))))
  )

 (local
  (defthm modelo-lista-formulas-seleccion
    (implies (and (not (modelo sigma (seleccion Gamma)))
		  (seleccion Gamma))
	     (not (modelo-lista-formulas sigma Gamma)))
    :hints (("Goal" :in-theory (disable modelo))))
  )

 (local
  (defthm contramodelo-lista-formulas-member
    (implies (and (modelo sigma F)
		  (member-equal F Gamma))
	     (not (contramodelo-lista-formulas sigma Gamma))))
  )

 (local
  (defthm contramodelo-lista-formulas-seleccion
    (implies (and (modelo sigma (seleccion Gamma))
		  (seleccion Gamma))
	     (not (contramodelo-lista-formulas sigma Gamma)))
    :hints (("Goal" :in-theory (disable modelo))))
  )

 (defun sec-asignacion-distinguida-lista (sigma lista-S)
   (cond ((endp lista-S) nil)
	 (t (or (sec-asignacion-distinguida sigma (car lista-S))
		(sec-asignacion-distinguida-lista sigma (cdr lista-S))))))
 
 (defthm P3-sec
   (implies (and (sec-objeto S)
		 (not (equal (sec-regla-computacion S) t)))
	    (iff (sec-asignacion-distinguida-lista 
		  sigma (sec-regla-computacion S))
		 (sec-asignacion-distinguida sigma S)))
   :hints (("Goal" :in-theory (disable aade-elemento))))

 )

;;;============================================================================
;;;
;;; 10) Propiedad P4
;;;
;;;============================================================================

(encapsulate
 ()

 (local
  (defthm sec-modelo-asigna-modelo
    (implies (and (member-equal F Gamma)
		  (not (interseccion Gamma Delta))
		  (es-simbolo-proposicional F))
	     (modelo (append (asignacion-positiva Gamma)
			     (asignacion-negativa Delta)) F))))

 (local
  (defthm sec-modelo-asigna-no-modelo
    (implies (and (member-equal F Delta)
		  (not (interseccion Gamma Delta))
		  (es-simbolo-proposicional F))
	     (not (modelo (append (asignacion-positiva Gamma)
				  (asignacion-negativa Delta)) F)))))
  
 (local
  (include-book "../6-extensiones/listas"))

 (local (in-theory (disable modelo)))

 (local
  (defthm modelo-lista-formulas-funciona-aux
    (implies (and (lista-formulas Gamma)
		  (not (seleccion Gamma))
		  (not (interseccion Gamma Delta))
		  (member-equal F Gamma))
	     (modelo (append (asignacion-positiva Gamma)
			     (asignacion-negativa Delta)) F))
    :hints (("Goal" :use (:instance si-no-seleccion-miembros-atomicos)))))

 (local
  (defthm modelo-lista-formulas-funciona
    (implies (and (lista-formulas Gamma)
		  (not (seleccion Gamma))
		  (not (interseccion Gamma Delta)))
	     (modelo-lista-formulas (append (asignacion-positiva Gamma)
					    (asignacion-negativa Delta))
				    Gamma))
    :hints ((extension-conjuntiva-a-listas
	     Gamma
	     (lambda (F)
	       (modelo (append (asignacion-positiva Gamma)
			       (asignacion-negativa Delta)) F))
	     (lambda (lista-F)
	       (modelo-lista-formulas (append (asignacion-positiva Gamma)
					      (asignacion-negativa Delta))
				      lista-F))))))

 (local
  (defthm contramodelo-lista-formulas-funciona-aux
    (implies (and (lista-formulas Delta)
		  (not (seleccion Delta))
		  (not (interseccion Gamma Delta))
		  (member-equal F Delta))
	     (not (modelo (append (asignacion-positiva Gamma)
				  (asignacion-negativa Delta)) F)))
    :hints (("Goal" :use (:instance si-no-seleccion-miembros-atomicos
				    (Gamma Delta))))))
  
 (local
  (defthm contramodelo-lista-formulas-funciona
    (implies (and (lista-formulas Delta)
		  (not (seleccion Delta))
		  (not (interseccion Gamma Delta)))
	     (contramodelo-lista-formulas (append (asignacion-positiva Gamma)
						  (asignacion-negativa Delta))
					  Delta))
    :hints ((extension-conjuntiva-a-listas
	       Delta
	       (lambda (F)
		 (not (modelo (append (asignacion-positiva Gamma)
				      (asignacion-negativa Delta)) F)))
	       (lambda (lista-F)
		 (contramodelo-lista-formulas 
		  (append (asignacion-positiva Gamma)
			  (asignacion-negativa Delta))
		  lista-F))))))

 (defthm P4-sec
   (implies (and (sec-objeto S)
		 (equal (sec-regla-computacion S) t))
	    (sec-asignacion-distinguida (sec-modelo S) S)))

 )

;;;============================================================================
;;;
;;; 11) Instancia
;;;
;;;============================================================================

;;;   A continuacin instanciamos la teora (definiciones y teoremas) del
;;; demostrador sat genrico con la siguiente correspondencia entre las
;;; funciones:

;;;   gen-objeto                       -- sec-objeto
;;;   gen-representacion               -- sec-representacion
;;;   gen-asignacion-distinguida       -- sec-asignacion-distinguida
;;;   gen-asignacion-distinguida-lista -- sec-asignacion-distinguida-lista
;;;   gen-regla-computacion            -- sec-regla-computacion
;;;   gen-seleccion                    -- car
;;;   gen-medida                       -- sec-medida
;;;   gen-modelo                       -- sec-modelo

;;;   Falta por demostrar la propiedad asociada a la funcin obj-seleccion,
;;; que aplicada a un lista de ramas no vaca devuelva un elemento de dicha
;;; lista.  No es necesario establecerla explcitamente dado la simplicidad de
;;; la funcin que usamos en su lugar.

(in-theory (disable sec-regla-computacion
		    (:type-prescription SEC-REGLA-COMPUTACION)
		    sec-representacion
		    sec-medida
		    sec-asignacion-distinguida
		    sec-modelo
		    sec-objeto))

(definstance-*sat-generico*
  ((gen-objeto                       sec-objeto)
   (gen-representacion               sec-representacion)
   (gen-asignacion-distinguida       sec-asignacion-distinguida)
   (gen-asignacion-distinguida-lista sec-asignacion-distinguida-lista)
   (gen-regla-computacion            sec-regla-computacion)
   (gen-seleccion                    car)
   (gen-medida                       sec-medida)
   (gen-modelo                       sec-modelo))
  "-secuentes")

;;;============================================================================
;;;
;;; 12) Ejemplos
;;;
;;;============================================================================

#|

(include-book "../7-generico/SAT-secuentes")

;;;   Un ejemplo de utilizacin de este ATP es el siguiente:

;;;   Necesitamos una funcin de seleccin concreta, para ello indicamos que se
;;; va a permitir redefiniciones en el sistema.

(set-ld-redefinition-action '(:warn . :overwrite) state)

(defun seleccion (lista)
  (declare (xargs :guard (true-listp lista)))
  (cond ((endp lista) nil)
	((es-atomica (car lista)) (seleccion (cdr lista)))
	(t (car lista))))

;;;   Cargamos el archivo de evaluacin de ejemplos:

(value :q)

(compile 'seleccion)

(load "../0-ejemplos/ejemplos.lisp")

;;;   Cargamos el archivo con las frmulas de Pelletier y evaluamos los
;;; ejemplos: 

(include-book "../0-ejemplos/pelletier")

(evalua-ejemplos 'DAT-generico-secuentes *pelletier*)

;;;   Los resultados obtenidos fueron los siguientes:

;;; +---------------------------------------------------------------+--------+
;;; | Frmula                                                       | Tiempo |
;;; +---------------------------------------------------------------+--------+
;;; |  1) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  2) ((p) <-> p)                                             |  0.000 |
;;; |  3) (p -> q) -> (q -> p)                                     |  0.000 |
;;; |  4) (p -> q) <-> (q -> p)                                   |  0.000 |
;;; |  5) ((p | q) -> (p | r)) -> (p | (q -> r))                    |  0.000 |
;;; |  6) (p | p)                                                  |  0.000 |
;;; |  7) (p | ((p)))                                            |  0.000 |
;;; |  8) ((p -> q) -> p) -> p                                      |  0.000 |
;;; |  9) [(p | q) & (p | q) & (p | q)] -> (p | q)             |  0.000 |
;;; | 10) [(q -> r) & (r -> (p & q)) & (p -> (q | r))] -> (p <-> q) |  0.000 |
;;; | 11) (p <-> p)                                                 |  0.000 |
;;; | 12) [(p <-> q) <-> r] <-> [p <-> (q <-> r)]                   |  0.000 |
;;; | 13) [p | (q & r)] <-> [(p | q) & (p | r)]                     |  0.000 |
;;; | 14) (p <-> q) <-> ((q | p) & (q | p))                       |  0.000 |
;;; | 15) (p -> q) <-> (p | q)                                     |  0.000 |
;;; | 16) (p -> q) | (q -> p)                                       |  0.000 |
;;; | 17) ((p & (q -> r)) -> s) <-> ((p | q | s) & (p | r | s))  |  0.000 |
;;; +---------------------------------------------------------------+--------+

;;;   Cargamos el archivo con las frmulas de Urquhart y evaluamos los
;;; ejemplos: 

(include-book "../0-ejemplos/urquhart")

(evalua-ejemplos 'dat-generico-secuentes *urquhart*)

;;;   Los resultados obtenidos en tiempo con las frmulas de Urquhart son los
;;; siguientes:

;;; +=========+============+
;;; | Frmula | Tiempo     |
;;; +---------+------------+
;;; | U_1     |      0.000 |
;;; | U_2     |      0.000 |
;;; | U_3     |      0.000 |
;;; | U_4     |      0.000 |
;;; | U_5     |      0.010 |
;;; | U_6     |      0.010 |
;;; | U_7     |      0.040 |
;;; | U_8     |      0.100 |
;;; | U_9     |      0.310 |
;;; | U_10    |      0.630 |
;;; | U_11    |      1.530 |
;;; | U_12    |      3.570 |
;;; | U_13    |      8.230 |
;;; | U_14    |     18.540 |
;;; | U_15    |     43.290 |
;;; +=========+============+

(evalua-ejemplos 'sat-generico-secuentes *urquhart*)

;;; +=========+============+
;;; | Frmula | Tiempo     |
;;; +---------+------------+
;;; | U_1     |      0.000 |
;;; | U_2     |      0.000 |
;;; | U_3     |      0.000 |
;;; | U_4     |      0.000 |
;;; | U_5     |      0.000 |
;;; | U_6     |      0.000 |
;;; | U_7     |      0.000 |
;;; | U_8     |      0.000 |
;;; | U_9     |      0.010 |
;;; | U_10    |      0.000 |
;;; | U_11    |      0.000 |
;;; | U_12    |      0.000 |
;;; | U_13    |      0.000 |
;;; | U_14    |      0.010 |
;;; | U_15    |      0.000 |
;;; +=========+============+

;;;   Cargamos el archivo con las frmulas que representan el problema de las N
;;; reinas y evaluamos los ejemplos: 

(include-book "../0-ejemplos/reinas")

(evalua-ejemplos 'SAT-generico-secuentes *n-reinas*)

;;; +==========+============+
;;; | Frmula  | Tiempo     |
;;; +----------+------------+
;;; | Reinas-2 |      0.000 |
;;; | Reinas-3 |      0.040 |
;;; | Reinas-4 |      0.430 |
;;; | Reinas-5 |      1.900 |
;;; | Reinas-6 |    189.990 |
;;; +==========+============+

|#

;;;============================================================================
