;;;============================================================================
;;;
;;; Ttulo: Conjuntos
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../9-resolucion/conjuntos")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;;
;;; 0) Introduccin
;;;
;;;============================================================================

;;;   En este libro (conjuntos.lisp) se establecern los conceptos necesarios
;;; acerca de los conjuntos en ACL2. Los eventos en MAYUSCULAS se encuentran en
;;; el sistema, pero se han incluido para completar el desarrollo aqu
;;; expuesto.

;;;============================================================================
;;;
;;; 1) Conjuntos
;;;
;;;============================================================================

;;;   Representaremos los conjuntos como listas y utilizaremos las funciones
;;; predefinidas en el sistema para trabajar con ellos. Incluimos aqu las
;;; definiciones originales de estas funciones en ACL2, escritas en maysculas,
;;; para dar una idea de su comportamiento.

;;;----------------------------------------------------------------------------
;;;
;;; 1.1) Relacin de pertenencia
;;;
;;;----------------------------------------------------------------------------

;;;   La relacin de pertenencia de un elemento a un conjunto viene dada por la
;;; funcin member-equal, definida como sigue:

(DEFUN MEMBER-EQUAL (X LST)
  (DECLARE (XARGS :GUARD (TRUE-LISTP LST)))
  (COND ((ENDP LST) NIL)
	((EQUAL X (CAR LST)) LST)
	(T (MEMBER-EQUAL X (CDR LST)))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Relacin de contencin
;;;
;;;----------------------------------------------------------------------------

;;;   La relacin de contencin entre conjuntos viene dada por la funcin
;;; subsetp-equal, definida como sigue:

(DEFUN SUBSETP-EQUAL (X Y)
  (DECLARE (XARGS :GUARD (AND (TRUE-LISTP Y) 
			      (TRUE-LISTP X))))
  (COND ((ENDP X) T)
	((MEMBER-EQUAL (CAR X) Y)
	 (SUBSETP-EQUAL (CDR X) Y))
	(T NIL)))

(defthm subsetp-equal-member 
  (implies (and (subsetp-equal conjunto1 conjunto2)
		(member-equal elemento conjunto1))
	   (member-equal elemento conjunto2)))

;;;   Aunque en ACL2 existe la funcin subsetp-equal, no se han establecido
;;; (por lo menos los autores no las han encontrado) sus principales
;;; propiedades: reflexividad y transitividad.

(defthm subsetp-equal-cons
  (implies (subsetp-equal conj-1 conj-2)
	   (subsetp-equal conj-1 (cons elemento conj-2))))

(defthm subsetp-equal-reflexivo
  (subsetp-equal conjunto conjunto))

(defthm subsetp-equal-transitivo
  (implies (and (subsetp-equal conjunto1 conjunto2)
		(subsetp-equal conjunto2 conjunto3))
	   (subsetp-equal conjunto1 conjunto3)))

(defthm member-subsetp-aux
  (implies (and (member-equal elemento conjunto1)
		(not (member-equal elemento conjunto2)))
	   (not (subsetp-equal conjunto1 conjunto2))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Estrategia para demostrar la contencin de conjuntos
;;;
;;;----------------------------------------------------------------------------

(encapsulate
 (((subsetp-strategy-c1) => *)
  ((subsetp-strategy-c2) => *)
  ((subsetp-strategy-hyp) => *))

 (local (defun subsetp-strategy-c1 () nil))
		       
 (local (defun subsetp-strategy-c2 () nil))
		       
 (local (defun subsetp-strategy-hyp () t))

 (defthm subsetp-equal-strategy-constraint
   (implies (and (subsetp-strategy-hyp)
		 (member-equal strategy-e (subsetp-strategy-c1)))
	    (member-equal strategy-e (subsetp-strategy-c2)))))

;;;   A partir de esta propiedad tenemos que demostrar que $C_1$ est contenido
;;; en $C_2$. Para probar esto, demostramos que si para todo elemento de un
;;; multiconjunto m1, el nmero de ocurrencias de dicho elemento en m1 es
;;; menor o igual al nmero de ocurrencias en m2, entonces m1 est
;;; contenido en m2.

(encapsulate
 ()
 
 (local
  (defthm subsetp-equal-strategy-aux
    (implies (and (subsetp-strategy-hyp)
		  (subsetp-equal conjunto (subsetp-strategy-c1)))
	     (subsetp-equal conjunto (subsetp-strategy-c2)))))

 (defthm subsetp-equal-strategy
   (implies (subsetp-strategy-hyp)
	    (subsetp-equal (subsetp-strategy-c1) (subsetp-strategy-c2))))

 )

(defun components-equal-subsetp (form)
  (declare (xargs :mode :program))

  (case-match form
	      
	      (('IMPLIES form-hyp
			 ('IGUAL-CONJUNTO form-c1 form-c2))
	       (mv form-hyp form-c1 form-c2))
	      
	      (('IGUAL-CONJUNTO form-c1 form-c2)
	       (mv t form-c1 form-c2))
	      
	      (('IMPLIES form-hyp
			 ('SUBSETP-EQUAL form-c1 form-c2))
	       (mv form-hyp form-c1 form-c2))
	      
	      (('SUBSETP-EQUAL form-c1 form-c2)
	       (mv t form-c1 form-c2))
	      
	      (& (mv nil nil nil))))
 
(defun defstrategy-subsetp-hint (id clause world)
  (declare (xargs :mode :program)
	   (ignore id world))

  (mv-let (form-hyp form-c1 form-c2)
	  (components-equal-subsetp (car clause))
	  (if form-hyp
	      `(:use (:functional-instance 
		      subsetp-equal-strategy
		      (subsetp-strategy-c1 (lambda () ,form-c1))
		      (subsetp-strategy-c2 (lambda () ,form-c2))
		      (subsetp-strategy-hyp (lambda () ,form-hyp))))
	      nil)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Igualdad de conjuntos
;;;
;;;----------------------------------------------------------------------------

;;;   Definimos la igualdad de conjuntos como un test de doble contencin:

(defun igual-conjunto (conjunto1 conjunto2)
  (declare (xargs :guard (and (true-listp conjunto1)
			      (true-listp conjunto2))))
  (and (subsetp-equal conjunto1 conjunto2)
       (subsetp-equal conjunto2 conjunto1)))

(defthm member-igual-conjunto
  (implies (igual-conjunto conjunto1 conjunto2)
	   (iff (member-equal elemento conjunto1)
		(member-equal elemento conjunto2))))

;;;   Gracias a las propiedades demostradas sobre la contencin de conjuntos,
;;; se demuestra fcilmente que la igualdad es una relacin de equivalencia:

(defequiv igual-conjunto)

;;;----------------------------------------------------------------------------
;;;
;;; 1.5) Estrategia para demostrar la igualdad de conjuntos
;;;
;;;----------------------------------------------------------------------------

(encapsulate
 (((igual-conjunto-strategy-c1) => *)
  ((igual-conjunto-strategy-c2) => *)
  ((igual-conjunto-strategy-hyp) => *))

 (local (defun igual-conjunto-strategy-c1 () nil))
		       
 (local (defun igual-conjunto-strategy-c2 () nil))
		       
 (local (defun igual-conjunto-strategy-hyp () t))

 (defthm igual-conjunto-strategy-constraint
   (implies (igual-conjunto-strategy-hyp)
	    (iff (member-equal strategy-e (igual-conjunto-strategy-c1))
		 (member-equal strategy-e (igual-conjunto-strategy-c2))))))

;;;   A partir de esta propiedad tenemos que demostrar que $C_1$ es igual a
;;; $C_2$. Para probar esto, demostramos que si para todo elemento de un
;;; multiconjunto m1, el nmero de ocurrencias de dicho elemento en m1 es
;;; menor o igual al nmero de ocurrencias en m2, entonces m1 est
;;; contenido en m2.

(encapsulate
 ()
 
 (local
  (defthm igual-conjunto-strategy-aux-1
    (implies (igual-conjunto-strategy-hyp)
	     (subsetp-equal (igual-conjunto-strategy-c1)
			    (igual-conjunto-strategy-c2)))
    :hints (defstrategy-subsetp-hint)))

 (local
  (defthm igual-conjunto-strategy-aux-2
    (implies (igual-conjunto-strategy-hyp)
	     (subsetp-equal (igual-conjunto-strategy-c2)
			    (igual-conjunto-strategy-c1)))
    :hints (defstrategy-subsetp-hint)))

 (defthm igual-conjunto-strategy
   (implies (igual-conjunto-strategy-hyp)
	    (igual-conjunto (igual-conjunto-strategy-c1)
			    (igual-conjunto-strategy-c2))))

 )

(defun defstrategy-igual-conjunto-hint (id clause world)
  (declare (xargs :mode :program)
	   (ignore id world))

  (mv-let (form-hyp form-c1 form-c2)
	  (components-equal-subsetp (car clause))
	  (if form-hyp
	      `(:use (:functional-instance 
		      igual-conjunto-strategy
		      (igual-conjunto-strategy-c1 (lambda () ,form-c1))
		      (igual-conjunto-strategy-c2 (lambda () ,form-c2))
		      (igual-conjunto-strategy-hyp (lambda () ,form-hyp))))
	      nil)))

;;;----------------------------------------------------------------------------
;;;
;;; 1.6) Conjunto vacio
;;;
;;;----------------------------------------------------------------------------

(defun conjunto-vacio (conjunto)
  (declare (xargs :guard (true-listp conjunto)))
  (not (consp conjunto)))

(defcong igual-conjunto iff (conjunto-vacio c) 1)

;;;----------------------------------------------------------------------------
;;;
;;; 1.7) Unin de conjuntos
;;;
;;;----------------------------------------------------------------------------

;;;   La unin de conjuntos se calcula en el sistema con la funcin
;;; union-equal: 

(DEFUN UNION-EQUAL (X Y)
  (DECLARE (XARGS :GUARD (AND (TRUE-LISTP X) 
			      (TRUE-LISTP Y))))
  (COND ((ENDP X) Y)
	((MEMBER-EQUAL (CAR X) Y)
	 (UNION-EQUAL (CDR X) Y))
	(T (CONS (CAR X)
		 (UNION-EQUAL (CDR X) Y)))))

;;;   La siguiente caracterizacin de la unin de conjuntos es quiz ms
;;; importante que la definicin en s. De hecho, cualquier funcin que cumpla
;;; la siguiente propiedad sera adecuada para calcular la unin de conjuntos:

(defthm member-equal-union-equal
  (iff (member-equal elemento (union-equal conjunto1 conjunto2))
       (or (member-equal elemento conjunto1)
	   (member-equal elemento conjunto2))))

(defthm true-listp-union-equal
  (implies (and (true-listp conjunto1)
		(true-listp conjunto2))
	   (true-listp (union-equal conjunto1 conjunto2))))

(defcong igual-conjunto igual-conjunto (union-equal c1 c2) 1
  :hints (defstrategy-igual-conjunto-hint))

(defcong igual-conjunto igual-conjunto (union-equal c1 c2) 2
  :hints (defstrategy-igual-conjunto-hint))

(in-theory (disable union-equal))

;;;----------------------------------------------------------------------------
;;;
;;; 1.8) Interseccin de conjuntos
;;;
;;;----------------------------------------------------------------------------

;;;   En ACL2 no existe una funcin que sirva para calcular la interseccin de
;;; conjuntos con respecto a la relacin equal, pero si existe una que
;;; calcula la interseccin de conjuntos con respecto a la relacin eq.
;;; Tomaremos esta ltima como modelo para definir la funcin interseccion:

(defun interseccion (conjunto1 conjunto2)
  (declare (xargs :guard (and (true-listp conjunto1)
			      (true-listp conjunto2))))
  (cond ((endp conjunto1) nil)
	((member-equal (car conjunto1) conjunto2)
	 (cons (car conjunto1)
	       (interseccion (cdr conjunto1) conjunto2)))
	(t (interseccion (cdr conjunto1) conjunto2))))

;;;   Al igual que antes, no es tan importante la definicin de la funcin que
;;; calcula la interseccin como su caracterizacin. De forma que cualquier
;;; otra funcin que cumpla la siguiente propiedad sera adecuada para calcular
;;; la interseccin de conjuntos:

(defthm member-equal-interseccion
  (iff (member-equal elemento (interseccion conjunto1 conjunto2))
       (and (member-equal elemento conjunto1)
	    (member-equal elemento conjunto2))))

(defcong igual-conjunto igual-conjunto (interseccion c1 c2) 1
  :hints (defstrategy-igual-conjunto-hint))

(defcong igual-conjunto igual-conjunto (interseccion c1 c2) 2
  :hints (defstrategy-igual-conjunto-hint))

(defthm interseccion-existe
  (implies (and (member-equal elemento conjunto1)
		(member-equal elemento conjunto2))
	   (not (conjunto-vacio (interseccion conjunto1 conjunto2))))
  :rule-classes :type-prescription)

(defthm consp-interseccion
  (implies (and (member-equal elemento conjunto1)
		(member-equal elemento conjunto2))
	   (consp (interseccion conjunto1 conjunto2))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.9) Eliminacin de un elemento de un conjunto
;;;
;;;----------------------------------------------------------------------------

;;;   Par eliminar un elemento de un conjunto, tenemos que recorrer el conjunto
;;; completo para eliminar todas las ocurrencias de dicho elemento. La funcin
;;; que implementa ese proceso es la siguiente:

(defun elimina-elemento (elemento conjunto)
  (declare (xargs :guard (true-listp conjunto)))
  (cond ((endp conjunto) nil)
	((equal elemento (car conjunto))
	 (elimina-elemento elemento (cdr conjunto)))
	(t (cons (car conjunto)
		 (elimina-elemento elemento (cdr conjunto))))))

(defthm member-equal-elimina-elemento
  (iff (member-equal elemento1 (elimina-elemento elemento2 conjunto))
       (and (member-equal elemento1 conjunto)
	    (not (equal elemento1 elemento2))))
  :hints (("Goal" :induct (elimina-elemento elemento2 conjunto))))

(defcong igual-conjunto igual-conjunto (elimina-elemento e c) 2
  :hints (defstrategy-igual-conjunto-hint))

(defthm true-listp-elimina-elemento
  (implies (true-listp conjunto)
	   (true-listp (elimina-elemento elemento conjunto))))

(defthm e0-ord-<-elimina-elemento
  (implies (member-equal elemento conjunto)
	   (e0-ord-< (acl2-count (elimina-elemento elemento conjunto))
		     (acl2-count conjunto)))
  :hints (("Goal" :induct (elimina-elemento elemento conjunto))))

(defthm subsetp-equal-elimina-elemento-1
  (subsetp-equal (elimina-elemento elemento conjunto) conjunto))

(defthm interseccion-elimina-elemento
  (equal (elimina-elemento elemento (interseccion conjunto1 conjunto2))
	 (interseccion (elimina-elemento elemento conjunto1)
		       (elimina-elemento elemento conjunto2))))

(defthm consp-interseccion-elimina-elemento
  (implies (not (conjunto-vacio 
		 (interseccion conjunto1 
			       (elimina-elemento elemento conjunto2))))
	   (not (conjunto-vacio (interseccion conjunto1 conjunto2))))
  :rule-classes :type-prescription)

(defthm consp-interseccion-elimina-elemento-no-perteneciente
  (implies 
   (not (member-equal elemento conjunto2))
   (equal (interseccion (elimina-elemento elemento conjunto1) conjunto2)
	  (interseccion conjunto1 conjunto2))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.10) Conjuntos unitarios
;;;
;;;----------------------------------------------------------------------------

;;;   La siguiente funcin comprueba si un conjunto es unitario:

(defun conjunto-unitario (conjunto)
  (declare (xargs :guard (and (true-listp conjunto))))
  (and (consp conjunto)
       (endp (elimina-elemento (car conjunto) conjunto))))

(defthm subsetp-equal-elimina-elemento-2
  (implies (member-equal elemento conjunto)
	   (not (subsetp-equal conjunto 
			       (elimina-elemento elemento conjunto)))))

(defthm unicidad-en-conjunto-unitario ;1
  (implies (and (not (equal elemento1 elemento2))
		(member-equal elemento1 conjunto)
		(member-equal elemento2 conjunto))
	   (not (conjunto-unitario conjunto))))

(defthm unicidad-en-conjunto-unitario-interseccion
  (implies (and (member-equal elemento conjunto1)
		(member-equal elemento conjunto3)
		(subsetp-equal conjunto2 conjunto1)
		(conjunto-unitario (interseccion conjunto1 conjunto3))
		(not (conjunto-vacio (interseccion conjunto2 conjunto3))))
	   (member-equal elemento conjunto2))
  :hints (("Goal"
	   :in-theory (disable conjunto-unitario)))
  :rule-classes nil)

(defthm consp-interseccion-elimina-elemento-aux
  (implies 
   (and (not (conjunto-vacio (interseccion representante conjunto)))
	(not (conjunto-unitario (interseccion representante conjunto))))
   (not (conjunto-vacio 
	 (interseccion (elimina-elemento elemento representante) conjunto)))))

(defthm interseccion-unitaria-interseccion-vacia
  (implies (and (member-equal elemento conjunto1)
		(member-equal elemento conjunto2)
		(conjunto-unitario (interseccion conjunto1 conjunto2)))
	   (conjunto-vacio
	    (interseccion conjunto1 (elimina-elemento elemento conjunto2)))))

(in-theory (disable interseccion elimina-elemento))

;;;----------------------------------------------------------------------------
;;;
;;; 1.12) Conjunto miembro de una coleccin de conjuntos
;;;
;;;----------------------------------------------------------------------------

;;;   Usaremos dos formas de establecer la pertenencia de un conjunto a una
;;; coleccin. La primera de ellas, mediante la funcin member-equal, es til
;;; cuando es necesario recorrer todos los elementos de la coleccin o escoger
;;; un elemento de la misma. La segunda de ellas, mediante la funcin
;;; conjunto-miembro, se usar para establecer dicha propiedad en otras
;;; circustancias:

(defun conjunto-miembro (conjunto S)
  (declare (xargs :guard (and (true-listp conjunto)
			      (true-list-listp S))))
  (cond ((endp S) nil)
	((igual-conjunto conjunto (car S))
	 t)
	(t (conjunto-miembro conjunto (cdr S)))))

(defcong igual-conjunto iff (conjunto-miembro conjunto S) 1)

(defun conjunto-miembro-testigo (conjunto S)
  (declare (xargs :guard (and (true-listp conjunto)
			      (true-list-listp S))))
  (cond ((endp S) nil)
	((igual-conjunto conjunto (car S))
	 (car S))
	(t (conjunto-miembro-testigo conjunto (cdr S)))))

(defthm conjunto-miembro-testigo-member-equal
  (implies (conjunto-miembro conjunto S)
	   (member-equal (conjunto-miembro-testigo conjunto S) S)))

(defthm conjunto-miembro-testigo-igual-conjunto
  (implies (conjunto-miembro conjunto S)
	   (igual-conjunto (conjunto-miembro-testigo conjunto S) 
			   conjunto)))

(defthm member-equal->conjunto-miembro
  (implies (member-equal conjunto S)
	   (conjunto-miembro conjunto S)))

;;;============================================================================
