;;;============================================================================
;;;
;;; Ttulo: Resolucin: Saturacin por resolucin condicionada
;;; Parte de la tesis:
;;;    "Teora computacional (en ACL2) sobre clculos proposicionales"
;;; FJMM 08-06-02
;;;
;;;============================================================================

#| Cdigo para la certificacin:

(certify-book "../9-resolucion/resolucion-sat")

|#

(in-package "ACL2")

(set-verify-guards-eagerness 2)

;;;============================================================================
;;; 
;;; 0) Introduccin
;;;
;;;============================================================================

(include-book "../8-davisputnam/clausulas")

(include-book "../9-resolucion/conjuntos")

(in-theory (disable definiciones-conectivas literales-iguales-2))

;;;============================================================================
;;;
;;; 1) Resolucin condicionada
;;;
;;;============================================================================

(defthm complementario-equal-2
  (implies (and (es-literal L1)
                (es-literal L2))
           (iff (equal (complementario L1) L2)
                (equal L1 (complementario L2)))))

(in-theory (disable complementario es-literal))

(defun resolvente (L C1 C2)
  (declare (xargs :guard (and (true-listp C1)
                              (true-listp C2))))

  (union-equal (elimina-elemento L C1)
               (elimina-elemento (complementario L) C2)))

(defthm member-resolvente
  (iff (member-equal L1 (resolvente L2 C1 C2))
       (or (and (member-equal L1 C1)
                (not (equal L1 L2)))
           (and (member-equal L1 C2)
                (not (equal L1 (complementario L2)))))))

(defcong igual-conjunto igual-conjunto 
  (resolvente L C1 C2) 2
  :hints (("Goal" :in-theory (disable igual-conjunto complementario))))

(defcong igual-conjunto igual-conjunto 
  (resolvente L C1 C2) 3
  :hints (("Goal" :in-theory (disable igual-conjunto complementario))))

(defthm true-listp-resolvente
  (implies (and (true-listp C1)
                (true-listp C2))
           (true-listp (resolvente L C1 C2))))

(defthm elimina-elemento-es-clausula
  (implies (es-clausula C)
           (es-clausula (elimina-elemento L C)))
  :hints (("Goal" :in-theory (enable elimina-elemento))))

(defthm union-equal-es-clausula
  (implies (and (es-clausula C1)
                (es-clausula C2))
           (es-clausula (union-equal C1 C2)))
  :hints (("Goal" :in-theory (enable union-equal))))

(defthm resolvente-es-clausula
  (implies (and (es-clausula C1) 
                (es-clausula C2))
           (es-clausula (resolvente L C1 C2))))

;;;----------------------------------------------------------------------------
;;;
;;; 1.2) Bsqueda de un literal para hacer resolucin
;;;
;;;----------------------------------------------------------------------------

(defun literal-resolucion-clausulas (C1 C2)
  (declare (xargs :guard (and (true-listp C1)
                              (true-listp C2))))

  (cond ((endp C1) nil)
        ((member-equal (complementario (car C1)) C2)
         (car C1))
        (t (literal-resolucion-clausulas (cdr C1) C2))))

(defthm literal-resolucion-clausulas-propiedades-1
  (implies 
   (literal-resolucion-clausulas C1 C2)
   (and (member-equal (literal-resolucion-clausulas C1 C2) C1)
        (member-equal (complementario (literal-resolucion-clausulas C1 C2))
                      C2)))
  :hints (("Goal" :induct (literal-resolucion-clausulas C1 C2)))
  :rule-classes :forward-chaining)

(defthm literal-resolucion-clausulas-suficiente-clausulas
  (implies (and (es-clausula C1)
                (es-clausula C2)
                (member-equal L C1)
                (member-equal (complementario L) C2))
           (literal-resolucion-clausulas C1 C2)))

(defthm literal-resolucion-clausulas-conmuta-clausulas
  (implies (and (literal-resolucion-clausulas C1 C2)
                (es-clausula C1)
                (es-clausula C2))
           (literal-resolucion-clausulas C2 C1))
  :hints (("Goal" 
           :use (:instance literal-resolucion-clausulas-suficiente-clausulas
                            (L (complementario 
                                (literal-resolucion-clausulas C1 C2)))
                            (C1 C2)
                            (C2 C1))))
  :otf-flg t
  :rule-classes :forward-chaining)

(defthm clausula-no-contiene-nil
  (implies (es-clausula C)
           (not (member-equal nil C))))

(defthm literal-resolucion-clausulas-propiedades-2
  (implies
   (and (es-clausula C1)
        (member-equal (literal-resolucion-clausulas C1 C2) C1))
   (member-equal (complementario (literal-resolucion-clausulas C1 C2)) C2))
  :hints (("Goal" :use literal-resolucion-clausulas-propiedades-1)))

(defthm literal-resolucion-clausulas-member-equal-1
  (implies (and (member-equal l c1)
                (member-equal (complementario l) c2))
           (member-equal (literal-resolucion-clausulas c1 c2) c1)))

(defthm literal-resolucion-clausulas-subsetp-equal
  (implies (and (subsetp-equal c1 c2)
                (literal-resolucion-clausulas c1 c3))
           (member-equal (literal-resolucion-clausulas c2 c3) c2))
  :hints (("Goal"
           :use (:instance literal-resolucion-clausulas-member-equal-1
                           (l (literal-resolucion-clausulas c1 c3))
                           (c1 c2)
                           (c2 c3)))))

(defthm literal-resolucion-clausulas-igual-conjunto-1
  (implies (and (igual-conjunto c1 c2)
                (literal-resolucion-clausulas c1 c3))
           (member-equal (literal-resolucion-clausulas c2 c3) c2))
  :rule-classes :forward-chaining)

(defcong igual-conjunto iff (member-equal l c) 2)

(defthm literal-resolucion-clausulas-igual-conjunto-2
  (implies (and (igual-conjunto c1 c2)
                (literal-resolucion-clausulas c1 c3))
           (member-equal (literal-resolucion-clausulas c2 c3) c1))
  :rule-classes :forward-chaining)

;;;----------------------------------------------------------------------------
;;;
;;; 1.3) Condicin de resolucin
;;;
;;;----------------------------------------------------------------------------

(encapsulate
 (((condicion * *) => *)
  ((asignacion) => *))

 (local
  (defun asignacion ()
    nil))

 (local
  (defun condicion (C1 C2)
    (declare (ignore C1 C2))
    t))

 (defthm condicion-boolean-p
   (booleanp (condicion C1 C2)))

 (defcong igual-conjunto iff (condicion c1 c2) 1)

 (defcong igual-conjunto iff (condicion c1 c2) 2)

 (defthm condicion-conmutativa
   (iff (condicion c1 c2)
        (condicion c2 c1)))

 (defthm propiedad-fundamental-condicion
   (implies (and (es-clausula D)
                 (es-clausula C)
                 (not (modelo-clausula (asignacion) D))
                 (modelo-clausula (asignacion) C))
            (condicion C D)))

 )

;;;----------------------------------------------------------------------------
;;;
;;; 1.4) Resolucin condicionada entre dos clusulas
;;; 
;;;----------------------------------------------------------------------------

(defun resolucion-cond-clausulas (C1 C2)
  (declare (xargs :guard (and (true-listp C1)
                              (true-listp C2))))

  (if (condicion C1 C2)
      (let ((L (literal-resolucion-clausulas C1 C2)))
        (if L
            (let ((res (resolvente L C1 C2)))
              (if (tiene-literales-complementarios res)
                  nil
                  (list res)))
            nil))
      nil))

(defthm es-clausula-subsetp-equal
  (implies (and (subsetp-equal C1 C2)
                (es-clausula C2))
           (es-clausula C1)))

(defthm es-clausula-congruencia
  (implies (igual-conjunto c1 c2)
           (iff (es-clausula c1)
                (es-clausula c2)))
  :rule-classes :congruence)

(defthm tiene-literales-complementarios-member
  (implies (and (es-literal l)
                (member-equal l c)
                (member-equal (complementario l) c))
           (tiene-literales-complementarios c)))

(defthm member-clausula-literal
  (implies (and (member-equal L C)
                (es-clausula C))
           (es-literal L)))

(defthm tiene-literales-complementarios-subsetp-equal
  (implies (and (subsetp-equal c1 c2)
                (tiene-literales-complementarios c1)
                (es-clausula c1))
           (tiene-literales-complementarios c2)))

(defthm tiene-literales-complementarios-congruencia
  (implies (and (igual-conjunto c1 c2)
                (es-clausula c1))
           (iff (tiene-literales-complementarios c1)
                (tiene-literales-complementarios c2))))

(defthm tiene-literales-complementarios-resolvente
  (implies (and (igual-conjunto c1 c2)
                (es-clausula c1)
                (es-clausula c))
           (iff (tiene-literales-complementarios (resolvente l c1 c))
                (tiene-literales-complementarios (resolvente l c2 c))))
  :hints (("Goal"
           :in-theory (disable igual-conjunto resolvente)
           :use (:instance tiene-literales-complementarios-congruencia
                           (c1 (resolvente l c1 c))
                           (c2 (resolvente l c2 c))))))

(defthm literal-resolucion-clausulas-not-igual-conjunto-1
  (implies (and (not (equal (literal-resolucion-clausulas c1 c3)
                            (literal-resolucion-clausulas c2 c3)))
                (es-clausula c1)
                (es-clausula c3)
                (igual-conjunto c1 c2))
           (not (resolucion-cond-clausulas c1 c3)))
  :hints (("Goal" :in-theory (disable resolvente igual-conjunto)
           :USE (:INSTANCE 
                 TIENE-LITERALES-COMPLEMENTARIOS-MEMBER
                 (L (LITERAL-RESOLUCION-CLAUSULAS C2 C3))
                 (C (RESOLVENTE (LITERAL-RESOLUCION-CLAUSULAS C1 C3)
                                C1 C3))))))

(defthm resolucion-cond-clausulas-congruencia-1
  (implies (and (es-clausula C1)
                (es-clausula C3)
                (igual-conjunto C1 C2))
           (igual-conjunto (car (resolucion-cond-clausulas C1 C3))
                           (car (resolucion-cond-clausulas C2 C3))))
  :hints (("Goal" :in-theory (disable resolvente igual-conjunto)
           :cases ((equal (literal-resolucion-clausulas C1 C3)
                          (literal-resolucion-clausulas C2 C3))))))

(defthm resolucion-cond-clausulas-consp-1
  (implies (and (es-clausula C1)
                (es-clausula C3)
                (igual-conjunto C1 C2))
           (iff (consp (resolucion-cond-clausulas C1 C3))
                (consp (resolucion-cond-clausulas C2 C3))))
  :hints (("Goal" :in-theory (disable resolvente igual-conjunto)
           :cases ((equal (literal-resolucion-clausulas C1 C3)
                          (literal-resolucion-clausulas C2 C3))))))

(defthm literal-resolucion-clausulas-conmuta-not-equal-1
  (implies (and (not (equal (literal-resolucion-clausulas c1 c2)
                            (complementario 
                             (literal-resolucion-clausulas c2 c1))))
                (es-clausula c1)
                (es-clausula c2))
           (not (resolucion-cond-clausulas c1 c2)))
  :otf-flg t
  :hints (("Goal" :in-theory (disable resolvente igual-conjunto)
           :use (:INSTANCE 
                 TIENE-LITERALES-COMPLEMENTARIOS-MEMBER
                 (L (COMPLEMENTARIO (LITERAL-RESOLUCION-CLAUSULAS C2 C1)))
                 (C (RESOLVENTE (LITERAL-RESOLUCION-CLAUSULAS C1 C2)
                                C1 C2))))))

(defthm literal-resolucion-clausulas-conmuta-not-equal-2
  (implies (and (not (equal (literal-resolucion-clausulas c1 c2)
                            (complementario 
                             (literal-resolucion-clausulas c2 c1))))
                (es-clausula c1)
                (es-clausula c2))
           (not (resolucion-cond-clausulas c2 c1)))
  :otf-flg t
  :hints (("Goal" :in-theory (disable resolvente igual-conjunto)
           :use (:INSTANCE 
                 TIENE-LITERALES-COMPLEMENTARIOS-MEMBER
                 (L (complementario (LITERAL-RESOLUCION-CLAUSULAS C1 C2)))
                 (C (RESOLVENTE (LITERAL-RESOLUCION-CLAUSULAS C2 C1)
                                C2 C1))))))
  
(defthm resolvente-conmuta
  (implies (and (equal l2 (complementario l1))
                (es-literal l1))
           (igual-conjunto (resolvente l1 c1 c2)
                           (resolvente l2 c2 c1)))
  :hints (defstrategy-igual-conjunto-hint))

(defthm resolucion-cond-clausulas-conmuta
  (implies (and (es-clausula C1)
                (es-clausula C2))
           (igual-conjunto (car (resolucion-cond-clausulas C1 C2))
                           (car (resolucion-cond-clausulas C2 C1))))
  :hints (("Goal" :in-theory (disable resolvente igual-conjunto)
           :use (:instance resolvente-conmuta
                           (l1 (literal-resolucion-clausulas c1 c2))
                           (l2 (literal-resolucion-clausulas c2 c1))))))

(defthm resolucion-cond-clausulas-consp-iff-conmuta
  (implies (and (es-clausula C1)
                (es-clausula C2))
           (iff (consp (resolucion-cond-clausulas C1 C2))
                (consp (resolucion-cond-clausulas C2 C1))))
  :hints (("Goal" :in-theory (disable resolvente igual-conjunto)
           :use (:instance resolvente-conmuta
                           (l1 (literal-resolucion-clausulas c1 c2))
                           (l2 (literal-resolucion-clausulas c2 c1))))))

(defthm resolucion-cond-clausulas-consp-2
  (implies (and (es-clausula C1)
                (es-clausula C3)
                (igual-conjunto C1 C2))
           (iff (consp (resolucion-cond-clausulas C3 C1))
                (consp (resolucion-cond-clausulas C3 C2))))
  :hints (("Goal" 
           :in-theory (disable igual-conjunto resolucion-cond-clausulas))))

(defthm resolucion-cond-clausulas-congruencia-2
  (implies (and (es-clausula C1)
                (es-clausula C3)
                (igual-conjunto C1 C2))
           (igual-conjunto (car (resolucion-cond-clausulas C3 C1))
                           (car (resolucion-cond-clausulas C3 C2))))
  :hints (("Goal"
           :in-theory (disable igual-conjunto resolucion-cond-clausulas))))

;;;============================================================================
;;;
;;; 2) Conjunto saturado por resolucin condicionada
;;;
;;;============================================================================

(defun saturado-cond-clausula-conjunto (C S-1 S-2)
  (declare (xargs :guard (and (true-list-listp S-1)
                              (true-list-listp S-2)
                              (true-listp C))))

  (cond ((endp S-1) t)
        (t (let ((res (resolucion-cond-clausulas C (car S-1))))
             (if (consp res)
                 (and (conjunto-miembro (car res) S-2)
                      (saturado-cond-clausula-conjunto C (cdr S-1) S-2))
                 (saturado-cond-clausula-conjunto C (cdr S-1) S-2))))))

(defun saturado-cond-conjunto-aux (S-1 S-2)
  (declare (xargs :guard (and (true-list-listp S-1)
                              (true-list-listp S-2))))

  (cond ((endp S-1) t)
        (t (and (saturado-cond-clausula-conjunto (car S-1) S-1 S-2)
                (saturado-cond-conjunto-aux (cdr S-1) S-2)))))

(defun saturado-cond-conjunto (S)
  (declare (xargs :guard (true-list-listp S)))

  (saturado-cond-conjunto-aux S S))

;;;----------------------------------------------------------------------------

(encapsulate
 (((saturado-cond-S) => *)
  ((saturado-cond-H) => *))

 (local 
  (defun saturado-cond-S () 
    '(nil)))

 (local
  (defun saturado-cond-H ()
    t))

 (defthm resolvente-cond-miembro-saturado
   (implies (and (saturado-cond-H)
                 (conjunto-miembro C1 (saturado-cond-S))
                 (conjunto-miembro C2 (saturado-cond-S))
                 (consp (resolucion-cond-clausulas C1 C2)))
            (conjunto-miembro (car (resolucion-cond-clausulas C1 C2))
                              (saturado-cond-S))))

 )

(encapsulate
 ()

 (local
  (defthm saturado-cond-clausula-conjunto-extension
    (implies (and (saturado-cond-H)
                  (subsetp-equal S (saturado-cond-S))
                  (conjunto-miembro C (saturado-cond-S)))
             (saturado-cond-clausula-conjunto C S (saturado-cond-S)))
    :hints (("Goal" :in-theory (disable resolucion-cond-clausulas)))))

 (local
  (defthm saturado-cond-conjunto-aux-extension
    (implies (and (saturado-cond-H)
                  (subsetp-equal S (saturado-cond-S)))
             (saturado-cond-conjunto-aux S (saturado-cond-S)))))
  
 (defthm saturado-cond-conjunto-extension
   (implies (saturado-cond-H)
            (saturado-cond-conjunto (saturado-cond-S))))

 )

;;;----------------------------------------------------------------------------

(encapsulate
 ()

 (local
  (in-theory (disable resolucion-cond-clausulas igual-conjunto)))

 (local
  (defthm es-forma-clausal-conjunto-miembro
    (implies (and (conjunto-miembro C S)
                  (es-forma-clausal S))
             (es-clausula C))))

 (local
  (defthm saturado-cond-clausula-conjunto-congruencia
    (implies (and (igual-conjunto C1 C2)
                  (es-clausula C1)
                  (es-forma-clausal s-1)
                  (es-forma-clausal S-2))
             (iff (saturado-cond-clausula-conjunto C1 S-1 S-2)
                  (saturado-cond-clausula-conjunto C2 S-1 S-2)))))

 (local
  (defthm saturado-cond-clausula-conjunto-contiene-resolventes
    (implies (and (saturado-cond-clausula-conjunto C1 S-1 S-2)
                  (conjunto-miembro C2 S-1)
                  (consp (resolucion-cond-clausulas C1 C2))
                  (es-forma-clausal s-1)
                  (es-forma-clausal S-2)
                  (es-clausula c1))
             (conjunto-miembro (car (resolucion-cond-clausulas C1 C2)) S-2))))

 (local
  (defthm saturado-cond-conjunto-aux-contiene-resolventes-aux
    (implies (and (saturado-cond-conjunto-aux S-1 S-2)
                  (consp S-1)
                  (conjunto-miembro C1 S-1)
                  (igual-conjunto C2 (car S-1))
                  (consp (resolucion-cond-clausulas C1 C2))
                  (es-forma-clausal s-1)
                  (es-forma-clausal S-2))
             (conjunto-miembro (car (resolucion-cond-clausulas C1 C2)) S-2))
    :hints (("Goal" 
             :use (:instance 
                   saturado-cond-clausula-conjunto-contiene-resolventes
                   (C2 C1)
                   (C1 C2))))))

 (local
  (defthm saturado-cond-conjunto-aux-contiene-resolventes
    (implies (and (saturado-cond-conjunto-aux S-1 S-2)
                  (conjunto-miembro C1 S-1)
                  (conjunto-miembro C2 S-1)
                  (consp (resolucion-cond-clausulas C1 C2))
                  (es-forma-clausal S-1)
                  (es-forma-clausal S-2))
             (conjunto-miembro (car (resolucion-cond-clausulas C1 C2)) S-2))
    :hints (("Goal" 
             :induct (conjunto-miembro c2 s-1)))))

 (defthm saturado-cond-conjunto-contiene-resolventes
   (implies (and (saturado-cond-conjunto S)
                 (conjunto-miembro C1 S)
                 (conjunto-miembro C2 S)
                 (consp (resolucion-cond-clausulas C1 C2))
                 (es-forma-clausal S))
            (conjunto-miembro (car (resolucion-cond-clausulas C1 C2)) S)))

 )

;;;============================================================================
;;;
;;; 3) Decisin de insatisfacibilidad por resolucin condicionada
;;;
;;;============================================================================

(defun resolucion-cond-clausula-conjunto (C S-2 S-N S-1-2)
  (declare (xargs :guard (and (true-listp C)
                              (true-list-listp S-2)
                              (true-list-listp S-N)
                              (true-list-listp S-1-2))))

  (if (endp S-2) 
      S-N
      (let ((res (resolucion-cond-clausulas C (car S-2))))
        (if (or (not (consp res))
                (conjunto-miembro (car res) S-1-2)
                (conjunto-miembro (car res) S-N))
            (resolucion-cond-clausula-conjunto C (cdr S-2) S-N S-1-2)
            (resolucion-cond-clausula-conjunto
             C (cdr S-2) (cons (car res) S-N) S-1-2)))))

(defthm true-list-listp-resolucion-cond-clausula-conjunto
  (implies (true-list-listp S-N)
           (true-list-listp 
            (resolucion-cond-clausula-conjunto C S-2 S-N S-1-2))))

(defthm true-listp-resolucion-cond-clausula-conjunto
  (implies (true-list-listp S-N)
           (true-listp (resolucion-cond-clausula-conjunto C S-2 S-N S-1-2))))

(defun resolucion-cond-saturacion (S-1 S-2)
  (declare (xargs :mode :program
                  :guard (and (true-list-listp S-1)
                              (true-list-listp S-2))))

  (if (endp S-1)
      S-2
      (let ((resolventes 
             (resolucion-cond-clausula-conjunto
              (car S-1) S-2 nil (append S-1 S-2))))
        (if (contiene-clausula-vacia resolventes)
            t
            (resolucion-cond-saturacion
             (append resolventes (cdr S-1)) (cons (car S-1) S-2))))))

(defun saturacion-resolucion (S)
  (declare (xargs :mode :program
                  :guard (true-list-listp S)))
  (resolucion-cond-saturacion S nil))

(defun INSAT-resolucion (S)
  (declare (xargs :mode :program
                  :guard (true-list-listp S)))
  (equal (saturacion-resolucion S) t))

;;;============================================================================
;;;
;;; 4) Terminacin del proceso de saturacin por resolucin condicionada
;;;
;;;============================================================================

;;;----------------------------------------------------------------------------
;;;
;;; 4.1) Cota superior del nmero de clusulas
;;;
;;;----------------------------------------------------------------------------

(defun numero-clausulas-cota (Xs)
  (declare (xargs :guard (true-listp Xs)))

  (cond ((endp Xs) 1)
        (t (* 4 (numero-clausulas-cota 
                 (elimina-elemento (car Xs) Xs))))))

(defthm true-listp-variables-clausula
  (implies (true-listp C)
           (true-listp (variables-clausula C))))

(defthm true-listp-variables-forma-clausal
  (implies (true-list-listp S)
           (true-listp (variables-forma-clausal S))))

(defun medida-resolucion-cota (S)
  (declare (xargs :guard (true-list-listp S)))

  (numero-clausulas-cota (variables-forma-clausal S)))

(in-theory (enable elimina-elemento))

;;; 1)
;;;   Si un conjunto est contenido en otro, la propiedad se conserva si se
;;; elimina el mismo elemento de ambos conjuntos.

(defthm subsetp-equal-elimina-elemento
  (implies (subsetp-equal S-1 S-2)
           (subsetp-equal (elimina-elemento E S-1)
                          (elimina-elemento E S-2))))

;;; 2)
;;;   La eliminacin de elementos de un conjunto es una operacin
;;; "conmutativa". 

(defthm elimina-elemento-conmuta
  (equal (elimina-elemento E1 (elimina-elemento E2 S))
         (elimina-elemento E2 (elimina-elemento E1 S))))

;;; -)

;;;   Con estas propiedades se puede demostrar que la definicin de la funcin
;;; numero-clausulas-cota puede hacerse con respecto a cualquier elemento de
;;; la lista que se pasa como argumento:

(defthm numero-clausulas-cota-elimina-elemento
  (implies (member-equal E S)
           (equal (numero-clausulas-cota S)
                  (* 4 (numero-clausulas-cota (elimina-elemento E S))))))

;;;   Finalmente podemos demostrar que la funcin numero-clausulas-cota es
;;; congruente con respecto a igualdad de conjuntos. Para realizar esta prueba
;;; se ha definido un esquema de induccin para dos conjuntos en el que la
;;; llamada recursiva se hace sobre el resultado de eliminar de ambos el primer
;;; elemento de uno de ellos.

(encapsulate
 ()

 (local
  (defun numero-clausulas-cota-subsetp-induccion (s-1 S-2)
    (declare (xargs :verify-guards nil))
    
    (cond ((endp s-1) S-2)
          (t (numero-clausulas-cota-subsetp-induccion
              (elimina-elemento (car s-1) s-1)
              (elimina-elemento (car s-1) S-2))))))
  
 (local
  (defthm numero-clausulas-cota-subsetp
    (implies (subsetp-equal s-1 S-2)
             (<= (numero-clausulas-cota s-1)
                 (numero-clausulas-cota S-2)))
    :rule-classes :linear
    :hints (("Goal" 
             :induct (numero-clausulas-cota-subsetp-induccion s-1 S-2)))))

 (defcong igual-conjunto equal (numero-clausulas-cota s) 1)

 )

(in-theory (enable union-equal))

(in-theory (disable variable-literal))

;;;   En el siguiente encapsulate demostramos que el conjunto de variables de
;;; un conjunto de clusulas es congruente con respecto a la igualdad de
;;; conjuntos. Un par de lemas previos son necesarios para demostrar que si un
;;; conjunto de clusulas est contenido en otro entonces sus conjuntos de
;;; variables mantienen la misma relacin. Este resultado no es local al
;;; encapsulate puesto que ser de utilidad ms adelante. Finalmente se
;;; demuestra la congruencia que nos interesa.

(encapsulate
 ()
 
 (local
  (defthm subsetp-equal-union-equal
    (implies (and (subsetp-equal s-1 s3)
                  (subsetp-equal S-2 s3))
             (subsetp-equal (union-equal s-1 S-2) s3))))

 (local
  (defthm subsetp-equal-variables-clausula-forma-clausal
    (implies (member-equal C S)
             (subsetp-equal (variables-clausula C)
                            (variables-forma-clausal S)))))

 (defthm variables-forma-clausal-subsetp
   (implies (subsetp-equal s-1 S-2)
            (subsetp-equal (variables-forma-clausal s-1)
                           (variables-forma-clausal S-2)))
   :rule-classes :forward-chaining)

 (defcong igual-conjunto igual-conjunto (variables-forma-clausal S) 1)

 )

;;;   Ahora ya podemos demostrar que la funcin medida-resolucion-cota es
;;; congruente con respecto a la igualdad de conjuntos de clusulas. Esto es
;;; cierto ya que la funcin numero-clausulas-cota es congruente con respecto
;;; a la igualdad de conjuntos de variables y variables-forma-clausal es
;;; congruente con respecto a la igualdad de conjuntos de clusulas.

(defcong igual-conjunto equal (medida-resolucion-cota S) 1
  :hints (("Goal" :in-theory (disable igual-conjunto))))

;;;   Para demostrar que el valor de medida-resolucion-cota permanece
;;; constante a lo largo del proceso de saturacin bastar con demostrar que el
;;; conjunto de variables se conserva (gracias a la congruencia de
;;; numero-clausulas-cota con respecto a la igualdad de conjuntos de
;;; variables). Demostramos primero un conjunto de resultados que relacionan
;;; los conjuntos de variables de una clusula y de una forma clausal con
;;; distintas operaciones como la concatenacin, unin o eliminacin de un
;;; elemento. 

(encapsulate
 ()
 
 (defthm member-equal-literal-clausula
   (implies (member-equal L C)
            (member-equal (variable-literal L)
                          (variables-clausula C))))

 (local
  (defthm variables-clausula-union-equal
    (equal (variables-clausula (union-equal c1 c2))
           (union-equal (variables-clausula c1)
                        (variables-clausula c2)))))

 (local
  (defthm variable-en-elimina-elemento
    (implies (member-equal v (variables-clausula (elimina-elemento e c)))
             (member-equal v (variables-clausula c)))))
 
 (local
  (defthm variables-forma-clausal-append-member
    (iff (member-equal v (variables-forma-clausal (append s-1 S-2)))
         (or (member-equal v (variables-forma-clausal s-1))
             (member-equal v (variables-forma-clausal S-2))))))
  
 (local
  (defthm variable-en-resolucion-cond-clausula-conjunto
    (implies (and (not (member-equal v (variables-forma-clausal S-2)))
                  (not (member-equal v (variables-forma-clausal S-N)))
                  (not (member-equal v (variables-clausula C))))
             (not (member-equal 
                   v (variables-forma-clausal
                      (resolucion-cond-clausula-conjunto C S-2 S-N S-1-2)))))))
                        
 (in-theory (disable resolucion-cond-clausula-conjunto))

 (defthm medida-resolucion-cota-invariante-variables-forma-clausal
   (igual-conjunto
    (variables-forma-clausal
     (cons C (append 
              (append (resolucion-cond-clausula-conjunto
                       C S-2 nil (cons C (append S-1 S-2)))
                      S-1)
              S-2)))
    (variables-forma-clausal (cons C (append S-1 S-2))))
   :hints (defstrategy-igual-conjunto-hint))

 (defthm medida-resolucion-cota-invariante
   (equal 
    (medida-resolucion-cota 
     (cons C (append 
              (append (resolucion-cond-clausula-conjunto
                       C S-2 nil (cons C (append S-1 S-2)))
                      S-1)
              S-2)))
    (medida-resolucion-cota (cons C (append S-1 S-2)))))
 
 )

(defthm variables-forma-clausal-cons-1
  (subsetp-equal (variables-clausula c)
                 (variables-forma-clausal (cons c s)))
  :hints (defstrategy-subsetp-hint))

(defthm variables-forma-clausal-cons-2
  (subsetp-equal (variables-forma-clausal s)
                 (variables-forma-clausal (cons c s)))
  :hints (defstrategy-subsetp-hint))

;;;----------------------------------------------------------------------------
;;;
;;; 4.2) Nmero de clusulas de una forma clausal
;;;
;;;----------------------------------------------------------------------------

;;;   Calculamos ahora el nmero de clusulas distintas que hay en una forma
;;; clausal. La propiedad que nos interesa demostrar es que este valor es menor
;;; que el nmero de clusulas mximo para el conjunto de variables que
;;; aparezcan en la forma clausal, de ah que este concepto se defina en
;;; funcin de dicho conjunto de variables.

;;;   Diremos que dos clusulas son iguales con respecto a un conjunto de
;;; variables si dichas variables aparecen de la misma forma en las dos
;;; clusulas. Es decir, si cada una de estas variables aparece en un literal
;;; positivo, en un literal negativo o no aparecen en ambas clusulas. Al
;;; contar el nmero de clusulas distintas que hay en una forma clausal, lo
;;; haremos con respecto a un conjunto de variables; estableciendo as una
;;; relacin entre este concepto y el de la cota del nmero de clusulas que se
;;; pueden obtener con un conjunto de variables.

;;;   Para realizar dicho computo vamos subdividiendo el conjunto de clusulas
;;; en funcin de si contienen o no a una variable concreta (literal positivo)
;;; o a su negacin (literal negativo). Al igual que con otras definiciones,
;;; como no se hace ninguna comprobacin sobre el tipo de los argumentos, el
;;; significado de las funciones que realizan la divisin es un poco ms
;;; general. Estas funciones separan los conjuntos que contienen un elemento
;;; concreto de los que no lo contienen, independientemente de si dicho
;;; elemento es una variable o los conjuntos son clusulas. Diremos que estas
;;; operaciones "filtran" los conjuntos de clusulas.

(defun filtra-forma-clausal-t (L S)
  (declare (xargs :guard (true-list-listp S)))

  (cond ((endp S) nil)
        ((member-equal L (car S))
         (cons (car S) (filtra-forma-clausal-t L (cdr S))))
        (t (filtra-forma-clausal-t L (cdr S)))))

(defun filtra-forma-clausal-f (L S)
  (declare (xargs :guard (true-list-listp S)))

  (cond ((endp S) nil)
        ((not (member-equal L (car S)))
         (cons (car S) (filtra-forma-clausal-f L (cdr S))))
        (t (filtra-forma-clausal-f L (cdr S)))))

;;;   A continuacin presentamos propiedades de las funciones anteriores que
;;; sern necesarias en distintos puntos de la prueba de terminacin.

;;; 1)
;;;   Para la verificacin de las guardas:

(defthm true-list-listp-filtra-forma-clausal-f
  (implies (true-list-listp S)
           (true-list-listp (filtra-forma-clausal-f L S))))

(defthm true-list-listp-filtra-forma-clausal-t
  (implies (true-list-listp S)
           (true-list-listp (filtra-forma-clausal-t L S))))

;;; 2)
;;;   Casi-congruencias:

(defthm filtra-forma-clausal-t-member
  (iff (member-equal C (filtra-forma-clausal-t L S))
       (and (member-equal C S)
            (member-equal L C)))
  :hints (("Goal" :induct (filtra-forma-clausal-t L S))))

(defthm filtra-forma-clausal-t-subsetp-equal
  (implies (subsetp-equal s-1 S-2)
           (subsetp-equal (filtra-forma-clausal-t v S-1)
                          (filtra-forma-clausal-t v S-2)))
  :hints (defstrategy-subsetp-hint)
  :rule-classes :type-prescription)

(defthm filtra-forma-clausal-f-member
  (iff (member-equal C (filtra-forma-clausal-f L S))
       (and (member-equal C S)
            (not (member-equal L C))))
  :hints (("Goal" :induct (filtra-forma-clausal-f L S))))

(defthm filtra-forma-clausal-f-subsetp-equal
  (implies (subsetp-equal s-1 S-2)
           (subsetp-equal (filtra-forma-clausal-f v S-1)
                          (filtra-forma-clausal-f v S-2)))
  :hints (defstrategy-subsetp-hint)
  :rule-classes :type-prescription)

;;; 3)
;;;   Comportamiento con respecto a la concatenacin:

(defthm filtra-forma-clausal-t-append
  (equal (filtra-forma-clausal-t v (append s-1 S-2))
         (append (filtra-forma-clausal-t v s-1)
                 (filtra-forma-clausal-t v S-2))))

(defthm filtra-forma-clausal-f-append
  (equal (filtra-forma-clausal-f v (append s-1 S-2))
         (append (filtra-forma-clausal-f v s-1)
                 (filtra-forma-clausal-f v S-2))))

;;; 4)
;;;   Conmutatividad de las operaciones de filtrado:

(defthm filtra-forma-clausal-t-conmuta
  (equal (filtra-forma-clausal-t v1 (filtra-forma-clausal-t v2 s))
         (filtra-forma-clausal-t v2 (filtra-forma-clausal-t v1 s))))

(defthm filtra-forma-clausal-f-conmuta
  (equal (filtra-forma-clausal-f v1 (filtra-forma-clausal-f v2 s))
         (filtra-forma-clausal-f v2 (filtra-forma-clausal-f v1 s))))

(defthm filtra-forma-clausal-t-conmuta-f
  (equal (filtra-forma-clausal-t v1 (filtra-forma-clausal-f v2 s))
         (filtra-forma-clausal-f v2 (filtra-forma-clausal-t v1 s))))

;;; -)

;;;   A la hora de contar el nmero de clusulas distintas de una forma
;;; clausal, la iremos filtrando con respecto a todas las variables que
;;; aparecen en ella. Cuando se terminan las variables, comprobamos si queda
;;; algn elemento susceptible de ser una clusula. La funcin que realiza esta
;;; comprobacin es la siguiente:

(defun existe-consp (S)
  (declare (xargs :guard (true-list-listp S)))

  (cond ((endp S) nil)
        ((consp (car S)) t)
        (t (existe-consp (cdr S)))))

;;;   Algunas propiedades de esta funcin que son necesarias en la prueba de
;;; terminacin son las siguientes:

;;; 1)
;;;   Comportamiento con respecto a las operaciones de filtrado:

(defthm existe-consp-filtra-forma-clausal-t-f
  (implies (and (not (existe-consp (filtra-forma-clausal-t
                                    v1 (filtra-forma-clausal-t v2 s))))
                (not (existe-consp (filtra-forma-clausal-t
                                    v1 (filtra-forma-clausal-f v2 s))))
                (not (existe-consp (filtra-forma-clausal-f
                                    v1 (filtra-forma-clausal-t v2 s))))
                (not (existe-consp (filtra-forma-clausal-f
                                    v1 (filtra-forma-clausal-f v2 s)))))
           (not (existe-consp s))))

;;; 2)
;;;   Casi-congruencia:

(defthm subsetp-equal-existe-consp
  (implies (and (subsetp-equal s-1 S-2)
                (existe-consp s-1))
           (existe-consp S-2)))

;;; 3)
;;;   Comportamiento con respecto a la concatenacin:

(defthm existe-consp-append
  (iff (existe-consp (append S-1 S-2))
       (or (existe-consp S-1)
           (existe-consp S-2))))

;;; -)

;;;   El nmero de clusulas distintas con respecto a un conjunto de variables
;;; se determina filtrando las clusulas con respecto a cada una de las
;;; variables y de sus negaciones hasta que el conjunto de variables se agota.
;;; Cuando esto ocurre, sumamos 1 si todava queda algn elemento susceptible
;;; de ser una clusula (es decir un consp) y 0 en caso contrario.

(defun numero-clausulas-variables (variables S)
  (declare (xargs :guard (and (true-list-listp S)
                              (true-listp variables))))

  (cond ((endp variables)
         (if (existe-consp S)
             1
             0))
        (t (let ((V (car variables)))
             (+ (numero-clausulas-variables
                 (elimina-elemento V variables)
                 (filtra-forma-clausal-t 
                  V (filtra-forma-clausal-t (negacion V) S)))
                (numero-clausulas-variables
                 (elimina-elemento V variables)        
                 (filtra-forma-clausal-t
                  V (filtra-forma-clausal-f (negacion V) S)))
                (numero-clausulas-variables
                 (elimina-elemento V variables)        
                 (filtra-forma-clausal-f
                  V (filtra-forma-clausal-t (negacion V) S)))
                (numero-clausulas-variables
                 (elimina-elemento V variables)        
                 (filtra-forma-clausal-f
                  V (filtra-forma-clausal-f (negacion V) S))))))))

;;;   La medida de una forma clausal viene dada por el nmero de clusulas
;;; distintas con respecto al conjunto de variables que inicialmente aparecen
;;; en las mismas:

(defun medida-resolucion (S)
  (declare (xargs :guard (true-list-listp S)))

  (numero-clausulas-variables (variables-forma-clausal S) S))

;;;   Gracias a esta ingeniosa definicin, se puede demostrar fcilmente que el
;;; nmero de clsulas distintas con respecto a un conjunto de variables es
;;; menor o igual que la cota de dicho valor, calculada con la funcin
;;; numero-clausula-cota. 

(defthm numero-clausulas-cota-es-maximo
  (<= (numero-clausulas-variables variables saturado)
      (numero-clausulas-cota variables))
  :rule-classes :linear)

;;;   A continuacin vamos a demostrar que la medida de una forma clausal
;;; calculada con la funcin medida-resolucion es congruente con respecto a
;;; la igualdad de conjuntos. Para ello tendremos que demostrar dos propiedades
;;; de congruencia de la funcin numero-clausulas-variables con respecto a la
;;; igualdad de conjuntos en cada uno de sus argumentos.

;;; 1)
;;;   La funcin numero-clausulas-variables es congruente con respecto a la
;;; igualdad de conjuntos de variables (primer argumento).

(encapsulate
 ()

;;   El nmero de clausulas distintas con respecto a un conjunto de variables
;; se puede definir recursivamente con respecto a cualquier elemento de dicho
;; conjunto, no necesariamente el primero.
 
 (local
  (defthm numero-clausulas-variables-otra-definicion
    (implies (member-equal V var)
             (equal (numero-clausulas-variables var S)
                    (+ (numero-clausulas-variables 
                        (elimina-elemento V var) 
                        (filtra-forma-clausal-t 
                         V (filtra-forma-clausal-t (negacion V) S)))
                       (numero-clausulas-variables 
                        (elimina-elemento V var) 
                        (filtra-forma-clausal-t 
                         V (filtra-forma-clausal-f (negacion V) S)))
                       (numero-clausulas-variables 
                        (elimina-elemento V var) 
                        (filtra-forma-clausal-f 
                         V (filtra-forma-clausal-t (negacion V) S)))
                       (numero-clausulas-variables
                        (elimina-elemento V var) 
                        (filtra-forma-clausal-f 
                         V (filtra-forma-clausal-f (negacion V) S))))))
    :hints (("Goal" 
             :in-theory (disable 
                         (:type-prescription numero-clausulas-variables))))))

;;   El siguiente resultado prueba que en un conjunto que contenga algn
;; consp, siempre existe una clusula distinta con respecto a cualquier
;; conjunto de variables.

 (local
  (defthm existe-consp-numero-clausulas-variables->=1
    (implies (existe-consp s)
             (<= 1 (numero-clausulas-variables var s)))))

;;   Para probar la congruencia necesitamos un esquema de induccin basado en
;; la eliminacin de un mismo elemento de los dos conjuntos de variables:

 (local
  (defun numero-clausulas-variables-induccion-1 (var1 var2 S)
    (declare (xargs :verify-guards nil))
    
    (cond ((endp var1) (or t var2 S))
          (t (let ((V (car var1)))
               (and (numero-clausulas-variables-induccion-1
                     (elimina-elemento V var1)
                     (elimina-elemento V var2)
                     (filtra-forma-clausal-t 
                      V (filtra-forma-clausal-t (negacion V) S)))
                    (numero-clausulas-variables-induccion-1
                     (elimina-elemento V var1)
                     (elimina-elemento V var2)
                     (filtra-forma-clausal-t
                      V (filtra-forma-clausal-f (negacion V) S)))
                    (numero-clausulas-variables-induccion-1
                     (elimina-elemento V var1)
                     (elimina-elemento V var2)
                     (filtra-forma-clausal-f
                      V (filtra-forma-clausal-t (negacion V) S)))
                    (numero-clausulas-variables-induccion-1
                     (elimina-elemento V var1)
                     (elimina-elemento V var2)
                     (filtra-forma-clausal-f
                      V (filtra-forma-clausal-f (negacion V) S)))))))))
  
;;   El siguiente resultado es una "casi-congruencia", necesaria para demostrar
;; la congruencia real. Este resultado no est en modo local pues una de las
;; propiedades que hemos de demostrar sobre la funcin medida-resolucion, es
;; que su valor aumenta estrictamente en determinadas condiciones. Las
;; congruencias no son tiles para demostrar dicho resultado.

 (defthm numero-clausulas-variables-subsetp-equal-1
   (implies (subsetp-equal var1 var2)
            (<= (numero-clausulas-variables var1 S)
                (numero-clausulas-variables var2 S)))
   :hints (("Goal"
            :induct (numero-clausulas-variables-induccion-1 var1 var2 S)))
   :rule-classes (:linear :rewrite))

;;   Finalmente la congruencia:

 (defcong igual-conjunto equal (numero-clausulas-variables var S) 1)

 )

;;; 2)
;;;   La funcin numero-clausulas-variables es congruente con respecto a la
;;; igualdad de conjuntos de clusulas (segundo argumento).

;;;   Para probar la congruencia necesitamos un esquema de induccin basado en
;;; el filtrado de los dos conjuntos de clusulas. Este esquema es utilizado
;;; ms tarde y por tanto no lo he incluido localmente en un encapsulate.

(defun numero-clausulas-variables-induccion-2 (var S-1 S-2)
  (declare (xargs :guard (and (true-listp var)
                              (true-list-listp S-1)
                              (true-list-listp S-2))))

  (cond ((endp var) (or t S-1 S-2))
        (t (let ((V (car var)))
             (and (numero-clausulas-variables-induccion-2
                   (elimina-elemento V var)
                   (filtra-forma-clausal-t 
                    V (filtra-forma-clausal-t (negacion V) S-1))
                   (filtra-forma-clausal-t 
                    V (filtra-forma-clausal-t (negacion V) S-2)))
                  (numero-clausulas-variables-induccion-2
                   (elimina-elemento V var)            
                   (filtra-forma-clausal-t
                    V (filtra-forma-clausal-f (negacion V) S-1))
                   (filtra-forma-clausal-t
                    V (filtra-forma-clausal-f (negacion V) S-2)))
                  (numero-clausulas-variables-induccion-2
                   (elimina-elemento V var)
                   (filtra-forma-clausal-f
                    V (filtra-forma-clausal-t (negacion V) S-1))
                   (filtra-forma-clausal-f
                    V (filtra-forma-clausal-t (negacion V) S-2)))
                  (numero-clausulas-variables-induccion-2
                   (elimina-elemento V var)
                   (filtra-forma-clausal-f
                    V (filtra-forma-clausal-f (negacion V) S-1))
                   (filtra-forma-clausal-f
                    V (filtra-forma-clausal-f (negacion V) S-2))))))))

;;;   El siguiente resultado es una "casi-congruencia", necesaria para
;;; demostrar la congruencia real. Como con la congruencia anterior, este
;;; resultado es necesario para demostrar que el valor de medida-resolucion
;;; aumenta estrictamente en determinadas condiciones.

(defthm numero-clausulas-variables-subsetp-equal-2
  (implies (subsetp-equal S-1 S-2)
           (<= (numero-clausulas-variables var S-1)
               (numero-clausulas-variables var S-2)))
  :hints (("Goal"
           :induct (numero-clausulas-variables-induccion-2 var S-1 S-2)))
  :rule-classes :linear)

;;;   Finalmente la congruencia:

(defcong igual-conjunto equal (numero-clausulas-variables var S) 2)

;;; -)

;;;   Estas dos congruencias permiten demostrar un resultado similar para la
;;; funcin medida-resolucion. Dados dos conjuntos iguales de clusulas, los
;;; conjuntos de variables de las mismas son iguales por la congruencia de
;;; variables-forma-clausal. As pues, las llamadas a la funcin
;;; medida-resolucion se transforman en llamadas a la funcin
;;; numero-clausulas-variables en las que los respectivos argumentos son
;;; conjuntos iguales. Las dos congruencias anteriores justifican la igualdad
;;; de las mismas.

(defcong igual-conjunto equal (medida-resolucion S) 1)

;;;   La medida-resolucion de un conjunto de clusulas S-2 disminuye al
;;; considerar subconjuntos del mismo.

(defthm medida-resolucion-subsetp-equal
  (implies (subsetp-equal S-1 S-2)
           (<= (medida-resolucion S-1)
               (medida-resolucion S-2)))
  :rule-classes (:linear :rewrite))

;;;----------------------------------------------------------------------------
;;;
;;; 4.3) La medida-resolucion aumenta
;;;
;;;----------------------------------------------------------------------------

;;;   En la llamada recursiva de la funcin resolucion-cond-saturacion se
;;; pueden dar dos situaciones distintas, segn el resultado devuelto por la
;;; funcin resolucion-cond-clausula-conjunto sea el conjunto vaco o
;;; no. Si el resultado es distinto de vaco entonces sus elementos no aparecen
;;; entre los argumentos originales. En esta situacin tenemos que demostrar
;;; que el nmero de clusulas distintas aumenta.

;;;   Diremos que una clusula C es nueva en un conjunto de clusulas S con
;;; respecto a un conjunto de variables var, si en S no existe ningn
;;; elemento igual a C con respecto al conjunto de variables var. Este
;;; concepto se define por recursin sobre el conjunto de variables haciendo
;;; uso de las operaciones de filtrado presentadas en la subseccin anterior.

(defun clausula-nueva-variables (C S var)
  (declare (xargs :guard (and (true-listp C)
                              (true-list-listp S)
                              (true-listp var))))

  (cond ((endp var) (not (existe-consp S)))
        ((member-equal (car var) C)
         (if (member-equal (negacion (car var)) C)
             (clausula-nueva-variables
              C (filtra-forma-clausal-t 
                 (car var) (filtra-forma-clausal-t (negacion (car var)) S))
              (elimina-elemento (car var) var))
             (clausula-nueva-variables
              C (filtra-forma-clausal-t
                 (car var) (filtra-forma-clausal-f (negacion (car var)) S))
              (elimina-elemento (car var) var))))
        (t (if (member-equal (negacion (car var)) C)
               (clausula-nueva-variables
                C (filtra-forma-clausal-f 
                   (car var) (filtra-forma-clausal-t (negacion (car var)) S))
                (elimina-elemento (car var) var))
               (clausula-nueva-variables
                C (filtra-forma-clausal-f
                   (car var) (filtra-forma-clausal-f (negacion (car var)) S))
                (elimina-elemento (car var) var))))))

;;;   Este concepto se ha definido recursivamente con respecto al primer
;;; elemento de la lista de variables, pero se podra haber hecho con respecto
;;; a cualquier elemento de dicha lista:

(defthm clausula-nueva-variables-otra-definicion
  (implies (member-equal v var)
           (equal (clausula-nueva-variables C S var)
                  (if (member-equal v C)
                      (if (member-equal (negacion v) C)
                          (clausula-nueva-variables
                           C (filtra-forma-clausal-t 
                              v (filtra-forma-clausal-t (negacion v) S))
                           (elimina-elemento v var))
                          (clausula-nueva-variables
                           C (filtra-forma-clausal-t
                              v (filtra-forma-clausal-f (negacion v) S))
                           (elimina-elemento v var)))
                      (if (member-equal (negacion v) C)
                          (clausula-nueva-variables
                           C (filtra-forma-clausal-f 
                              v (filtra-forma-clausal-t (negacion v) S))
                           (elimina-elemento v var))
                          (clausula-nueva-variables
                           C (filtra-forma-clausal-f
                              v (filtra-forma-clausal-f (negacion v) S))
                           (elimina-elemento v var)))))))

;;;   Es fcil de probar que si una clusula C es nueva en un conjunto de
;;; clusulas S con respecto a un conjunto de variables var, entonces el
;;; nmero de clusulas de (cons C S) distintas con respecto a var es uno
;;; ms que el de S.

(defthm clausula-nueva-variables-es-nueva
  (implies (and (clausula-nueva-variables C S var)
                (consp C))
           (equal (numero-clausulas-variables var (cons C S))
                  (+ 1 (numero-clausulas-variables var S)))))

;;;   El resultado anterior establece una situacin en la que el nmero de
;;; clusulas aumenta estrictamente. Si demostramos que los elementos devueltos
;;; por la funcin resolucion-cond-clausula-conjunto son nuevos en el
;;; sentido de la definicin anterior, entonces podremos afirmar que el nmero
;;; de clusulas distintas aumenta. Para conseguir esto utilizamos el hecho de
;;; que los elementos de resolucion-cond-clausula-conjunto, si es que
;;; los hay, no pertenecen al conjunto de clusulas original.

;;;   Demostramos a continuacin que si una clusula C no pertenece a S
;;; (con respecto a la relacin conjunto-miembro), entonces C es nueva en
;;; S con respecto a cualquier conjunto de variables que contenga a las de
;;; C y a las de S. Por tanto, el nmero de clusulas distintas de 
;;; (cons C S) es mayor (en realidad uno ms) que el nmero de clusulas
;;; distintas de S.

(encapsulate
 ()

;;   Primero vamos a demostrar que si dos clusulas no son iguales, entonces
;; una de ellas es distinta en el conjunto formado por la otra con respecto a
;; cualquier conjunto de variables que contenga a las variables de ambas. Para
;; ello utilizaremos un literal cualquiera que establezca la diferencia entre
;; ambas clusulas.

;;   Primero demostramos la propiedad cuando en la primera clusula hay un
;; literal que no aparece en la segunda. Bastar con que la variable de dicho
;; literal este entre las variables con respecto a las que se comprueba que las
;; clusulas son distintas.

 (local
  (defthm clausula-nueva-variables-nil
    (implies (consp C)
             (clausula-nueva-variables C nil var))))
 
 (defthm member-equal-consp
   (implies (member-equal v c)
            (consp c)))
 
 (local
  (defthm clausula-nueva-variables-1
    (implies (and (member-equal (variable-literal L) var)
                  (member-equal L C1)
                  (not (member-equal L C2)))
             (clausula-nueva-variables C1 (list C2) var))
    :hints (("Goal" :in-theory (enable variable-literal)))))
 
;;   A continuacin demostramos la propiedad cuando en la segunda clusula hay
;; un literal que no aparece en la primera. De nueva bastar con que la
;; variable de dicho literal este entre las variables con respecto a las que
;; se comprueba que las clusulas son distintas.

 (local
  (defthm literal-negativo-variable-literal
    (implies (es-literal-negativo l)
             (equal (negacion (arg1 l)) l))))
 
 (local
  (defthm clausula-nueva-variables-2
    (implies (and (member-equal (variable-literal L) var)
                  (not (member-equal L C1))
                  (consp C1)
                  (member-equal L C2))
             (clausula-nueva-variables C1 (list C2) var))
    :hints (("Goal" :in-theory (enable variable-literal)))))

;;   Los resultados anteriores han de ser instanciados para un literal concreto
;; en el que ambas clusulas difieran. La siguiente funcin devuelve un literal
;; con esa caracterstica:

 (local
  (defun elemento-diferencia (C1 C2)
    (declare (xargs :guard (and (true-listp c1)
                                (true-listp c2))))
    
    (cond ((endp C1) nil)
          ((not (member-equal (car C1) C2))
           (car C1))
          (t (elemento-diferencia (cdr C1) C2)))))
 
 (local
  (defthm elemento-diferencia-propiedades
    (implies (not (subsetp-equal C1 C2))
             (and (member-equal (elemento-diferencia C1 C2) C1)
                  (not (member-equal (elemento-diferencia C1 C2) C2))))))
 
;;   El elemento que diferencia dos clusulas es el literal que hace falta en
;; los resultados clausula-nueva-variables-1 y clausula-nueva-variables-2
;; para probar que una de ellas es nueva en el conjunto formado por la otra con
;; respecto a cualquier conjunto de variables que contenga a las variables de
;; ambas.

 (local
  (defthm clausula-nueva-variables-subsetp-equal-1
    (implies (and (not (subsetp-equal C1 C2))
                  (subsetp-equal (variables-clausula C1) var))
             (clausula-nueva-variables C1 (list C2) var))
    :hints (("Goal" 
             :use (:instance clausula-nueva-variables-1
                             (L (elemento-diferencia C1 C2)))))))
 
 (local
  (defthm clausula-nueva-variables-subsetp-equal-2
    (implies (and (not (subsetp-equal C2 C1))
                  (consp C1)
                  (subsetp-equal (variables-clausula C2) var))
             (clausula-nueva-variables C1 (list C2) var))
    :hints (("Goal" 
             :use (:instance clausula-nueva-variables-2
                             (L (elemento-diferencia C2 C1)))))))
 
 (local
  (defthm clausula-nueva-variables-igual-conjunto
    (implies (and (not (igual-conjunto C1 C2))
                  (consp C1)
                  (subsetp-equal (variables-clausula C1) var)
                  (subsetp-equal (variables-clausula C2) var))
             (clausula-nueva-variables C1 (list C2) var))))
 
;;   Un detalle ms, el concepto de clusula nueva en un conjunto de clusulas
;; con respecto a un conjunto de variables se comporta adecuadamente con
;; respecto a la concatenacin de conjuntos de clusulas, es decir, si una
;; clusula es nueva en dos conjuntos de clusulas con respecto a un conjunto
;; de variables entonces tambin lo es en la concatenacin de dichos conjuntos
;; de clusulas:

 (local
  (defthm clausula-nueva-variables-append
    (implies (and (consp C)
                  (clausula-nueva-variables C S-1 var)
                  (clausula-nueva-variables C S-2 var))
             (clausula-nueva-variables C (append S-1 S-2) var))
    :hints (("Goal" 
             :induct (numero-clausulas-variables-induccion-2 var S-1 S-2)))))

;;   Por fin podemos demostrar que si una clusula C no pertenece a S
;; (con respecto a la relacin conjunto-miembro), entonces C es nueva en
;; S con respecto a cualquier conjunto de variables que contenga a las de
;; C y a las de S.

 (local
  (defthm clausula-nueva-variables-conjunto-miembro
    (implies (and (not (conjunto-miembro C S))
                  (consp C)
                  (subsetp-equal (variables-forma-clausal S) var)
                  (subsetp-equal (variables-clausula C) var))
             (clausula-nueva-variables C S var))
    :hints (("Goal" :induct (conjunto-miembro C S))
            ("Subgoal *1/3" 
             :use (:instance clausula-nueva-variables-append
                             (S-1 (list (car S)))
                             (S-2 (cdr S)))))))

;;   Y finalmente, si la clusula C no pertenece al conjunto S entonces el
;; nmero de clusulas distintas del conjunto (cons C S) es mayor que el del
;; conjunto S.

 (local
  (in-theory (disable variables-forma-clausal)))

 (local
  (defthm numero-clausulas-variables-cons
    (<= (numero-clausulas-variables (variables-forma-clausal S) S)
        (numero-clausulas-variables (variables-forma-clausal (cons C S)) S))
    :rule-classes :linear))

 (defthm medida-resolucion-not-conjunto-miembro
   (implies (and (not (conjunto-miembro C S))
                 (consp C))
            (< (medida-resolucion S)
               (medida-resolucion (cons C S))))
   :rule-classes :linear)

 )

;;;   Sin enbargo, la propiedad que queremos demostrar sobre el conjunto
;;; devuelto por resolucion-cond-clausula-conjunto es que aumenta la medida
;;; del conjunto original. No se trata slo de un elemento, por tanto ser
;;; necesario ampliar el resultado anterior a conjuntos. Si consideramos un
;;; conjunto de clusulas S-1 en el que al menos exista un elemento C no
;;; perteneciente a otro conjunto de clusulas S-2, entonces el nmero de
;;; clusulas distintas de (append S-1 S-2) es mayor que el de S-2. La
;;; prueba de este resultado es la siguiente:

;      (medida-resolucion S-2) < (medida-resolucion (cons C S-2)) <=
;                             <= (medida-resolucion (append S-1 S-2))

;;; donde la primera desigualdad se tiene por el teorema
;;; medida-resolucion-not-conjunto-miembro pues C no pertence a S-2 y la
;;; segunda por medida-resolucion-subsetp-equal pues (cons C S-2) est
;;; contenido en (append S-1 S-2). Por supuesto, para esto habra que
;;; desactivar la definicin de medida-resolucion.

(encapsulate
 ()
 
 (local
  (defthm subsetp-equal-cons-append
    (implies (member-equal C S-1)
             (subsetp-equal (cons C S-2) (append S-1 S-2)))))

 (local
  (in-theory (disable medida-resolucion)))

 (local
  (defthm medida-resolucion-cons-append
    (implies (member-equal C S-1)
             (<= (medida-resolucion (cons C S-2))
                 (medida-resolucion (append S-1 S-2))))
    :rule-classes :linear))

 (defthm medida-resolucion-not-conjunto-miembro-append
   (implies (and (not (conjunto-miembro C S-2))
                 (member-equal C S-1)
                 (consp C))
            (< (medida-resolucion S-2)
               (medida-resolucion (append S-1 S-2))))
   :rule-classes :rewrite)

 )

;;;   Para poder utilizar el teorema anterior tendremos que demostrar en el
;;; conjunto devuelto por resolucion-cond-clausula-conjunto al menos hay un
;;; elemento que es un consp y que no pertenece al conjunto de clusulas
;;; original. De hecho esto es cierto para cualquier elemento de dicho
;;; conjunto: 

;;; 1)
;;;   Si nil no est en el conjunto de clusulas devuelto por
;;; resolucion-cond-clausula-conjunto, entonces cualquier elemento de dicho
;;; conjunto es un consp. En particular esto es cierto para el primer
;;; elemento de dicho conjunto.

(encapsulate
 ()

 (local
  (defthm member-resolucion-cond-clausula-conjunto-consp
    (implies 
     (and (not (contiene-clausula-vacia
                (resolucion-cond-clausula-conjunto C-2 S-2 S-N S-1-2)))
          (member-equal 
           C-1 (resolucion-cond-clausula-conjunto C-2 S-2 S-N S-1-2))
          (not (member-equal C-1 S-N)))
     (consp C-1))
    :hints (("Goal" 
             :in-theory (enable resolucion-cond-clausula-conjunto)))))

 (defthm member-resolucion-cond-clausula-conjunto-consp-car
   (implies (and (resolucion-cond-clausula-conjunto C S-2 nil S-1-2)
                 (not (contiene-clausula-vacia
                       (resolucion-cond-clausula-conjunto C S-2 nil S-1-2))))
            (consp (car (resolucion-cond-clausula-conjunto C S-2 nil S-1-2))))
   :rule-classes :forward-chaining)

 )

;;; 2)
;;;   Cualquier elemento del conjunto de clusulas devuelto por
;;; resolucion-cond-clausula-conjunto que no pertenezca inicialmente al
;;; acumulador de dicha funcin (el tercer argumento) tampoco pertenece al
;;; al conjunto original (el cuarto argumento). En particular esto es cierto
;;; para el primer elemento de dicho conjunto de clusulas.

(encapsulate
 ()

 (local
  (defthm member-resolucion-cond-clausula-conjunto-not-conjunto-miembro
    (implies 
     (and (member-equal 
           C-1 (resolucion-cond-clausula-conjunto C-2 S-2 S-N S-1-2))
          (not (member-equal C-1 S-N)))
     (not (conjunto-miembro C-1 S-1-2)))
    :rule-classes nil
    :hints (("Goal"
             :in-theory (enable resolucion-cond-clausula-conjunto)))))

 (defthm member-resolucion-cond-clausula-conjunto-not-conjunto-miembro-car
   (implies (consp (resolucion-cond-clausula-conjunto C-2 S-2 nil S-1-2))
            (not (conjunto-miembro
                  (car (resolucion-cond-clausula-conjunto C-2 S-2 nil S-1-2)) 
                  S-1-2)))
   :rule-classes :forward-chaining
   :hints (("Goal"
            :use
            (:instance
             member-resolucion-cond-clausula-conjunto-not-conjunto-miembro
             (S-N nil)
             (C-1 
              (car (resolucion-cond-clausula-conjunto C-2 S-2 nil S-1-2)))))))

 )
 
;;; -)

;;;   Las dos propiedades anteriores del conjunto de clusulas devuelto por
;;; resolucion-cond-clausula-conjunto, junto con el resultado
;;; medida-resolucion-not-conjunto-miembro-append, aseguran que en la unin
;;; de dicho conjunto con el original hay ms clusulas de las que haba
;;; inicialmente: 

(defthm medida-resolucion-aumenta-aux
  (implies (and (not (contiene-clausula-vacia
                      (resolucion-cond-clausula-conjunto C S-2 nil S-1-2)))
                (consp (resolucion-cond-clausula-conjunto C S-2 nil S-1-2)))
           (< (medida-resolucion S-1-2)
              (medida-resolucion
               (append (resolucion-cond-clausula-conjunto C S-2 nil S-1-2) 
                       S-1-2))))
  :hints (("Goal" :in-theory (disable medida-resolucion)))
  :rule-classes :forward-chaining)

;;;   El resultado anterior no refleja la situacin exacta que se tiene en la
;;; prueba de terminacin de resolucion-cond-saturacion. El resultado
;;; interesante es una instancia del anterior en la que S-1 toma el valor
;;; (cons C saturado) y S-2 el valor (cons c (append s saturado)).

(defthm igual-conjunto-append-cons-append
  (igual-conjunto (append S-1 (cons C (append S-2 S-3)))
                  (cons C (append (append S-1 S-2) S-3))))

(defthm medida-resolucion-aumenta
  (implies 
   (and (not (contiene-clausula-vacia
              (resolucion-cond-clausula-conjunto 
               C S-2 nil (cons C (append S-1 S-2)))))
        (consp (resolucion-cond-clausula-conjunto 
                C S-2 nil (cons C (append S-1 S-2)))))
   (< (medida-resolucion (cons C (append S-1 S-2)))
      (medida-resolucion
       (cons C (append 
                (append (resolucion-cond-clausula-conjunto 
                         C S-2 nil (cons C (append S-1 S-2)))
                        S-1)
                S-2)))))
  :hints (("Goal" :use (:instance medida-resolucion-aumenta-aux
                                  (S-1-2 (cons C (append S-1 S-2))))))
  :rule-classes :linear)

;;;----------------------------------------------------------------------------
;;;
;;; 4.4) La prueba de terminacin.
;;;
;;;----------------------------------------------------------------------------

;;;   Finalmente probamos la terminacin de resolucion-cond-saturacion.
;;; Para ello utilizamos como medida el siguiente par:
;     (cons (+ (medida-resolucion-cota S-1) 1 (- (medida-resolucion S-1)))
;           (len S-2))
;;; Obsrvese que se suma 1 al primer elemento del par para que sea un
;;; ordinal segn la representancin que de ellos hay en ACL2.

(defthm e0-ordinalp-medida-res
  (e0-ordinalp (cons (+ (- (medida-resolucion S-1)) 1
                        (medida-resolucion-cota S-1))
                     (len S-2))))

;;;   Adems de la propiedad anterior, se necesita el siguiente resultado para
;;; demostrar las guardas:

(defthm true-list-listp-append
  (implies (and (true-list-listp lista-1)
		(true-list-listp lista-2))
	   (true-list-listp (append lista-1 lista-2))))

;;;   Y este otro para tratar el caso degenerado en el que el conjunto devuelto
;;; por resolucion-cond-clausula-conjunto sea vaco:

(defthm igual-conjunto-append-cons
  (igual-conjunto (append s-1 (cons c S-2))
                  (cons c (append s-1 S-2))))

;;;   El resultado de terminacin es el siguiente:

(defun medida-saturacion (S-1 S-2)
  (declare (xargs :guard (and (true-list-listp S-1)
                              (true-list-listp S-2))))
  (cons (- (+ (medida-resolucion-cota (append S-1 S-2)) 1)
           (medida-resolucion (append S-1 S-2)))
        (len S-1)))

(verify-termination resolucion-cond-saturacion
  (declare (xargs :measure (medida-saturacion S-1 S-2)
                  :hints (("Goal" 
                           :in-theory
                           (disable resolucion-cond-clausula-conjunto
                                    medida-resolucion-cota
                                    medida-resolucion
                                    e0-ordinalp)
                           :expand
                           (medida-saturacion S-1 S-2))
                          ("Subgoal 1"
                           :cases ((consp (resolucion-cond-clausula-conjunto 
                                           (car S-1) S-2
                                           nil (append S-1 S-2)))))))))

(verify-termination saturacion-resolucion)

(verify-termination INSAT-resolucion)

;;;   Veamos en detalle la prueba del resultado anterior. Para facilitar la
;;; explicacin llamaremos (medida-res S) al valor
;     (- (+ (medida-resolucion-cota S) 1) (medida-resolucion S))

;;;   El evento anterior genera dos subobjetivos. El primero consiste en
;;; demostrar que la medida propuesta es un ordinal y el segundo en probar que
;;; la medida de los argumentos en la llamada recursiva es menor que la medida
;;; de los argumentos originales. El teorema e0-ordinalp-medida-res
;;; proporciona la prueba del primer subobjetivo. Para demostrar el segundo
;;; subobjetivo tendremos en cuenta dos situaciones ditintas, segn el conjunto
;;; devuelto por la funcin resolucion-cond-clausula-conjunto sea o no vaco.
;;; Veamos a continuacin estas dos situaciones:

;;; 1)
;;;   Si el conjunto es vaco entonces los argumentos de la llamada recursiva
;;; son: (cdr S-1) y (cons (car S-1) S-2). En este caso hay que demostrar
;;; que el ordinal 
;     ((medida-res (append (cdr S-1) (cons (car S-1) S-2))) . (len (cdr S-1)))
;;; es menor que 
;     ((medida-res (append S-1 S-2)) . (len S-1))

;;;   Por el teorema igual-conjunto-append-cons se tiene que (append (cdr
;;; S-1) (cons (car S-1) S-2)) y (append S-1 S-2) son el mismo conjunto. Las
;;; cogruencias demostradas para las funciones medida-resolucion y
;;; medida-resolucion-cota nos aseguran que sus valores coinciden para ambas
;;; listas, de aqu que la funcin medida-res tambin coincida. Por otro lado
;;; el valor (len (cdr S-1)) es menor que (len S-1).  Esto demuestra que la
;;; medida de los argumentos de la llamada recursiva es menor que la de los
;;; argumentos originales.

;;; 2)
;;;   Llamaremos resolventes al valor:
;     (RESOLUCION-COND-CLAUSULA-CONJUNTO 
;        (CAR S-1) (CONS (CAR S-1) S-2) NIL (APPEND S-1 S-2))
;;; En este caso hay que demostrar que el ordinal
;     ((medida-res (append (append resolventes (cdr S-1)) 
;                          (cons (car S-1) S-2)))    . (len (cdr S-1)))
;;; es menor que
;     ((medida-res (append S-1 S-2)) . (len S-1))

;;;   Como ahora resolventes es no vaco, bastar con probar que el valor de
;;; (medida-res (append (append resolventes (cdr S-1)) (cons (car S-1)
;;; S-2))) es menor que (medida-res (append S-1 S-2)). Como la
;;; medida-resolucion-cota se mantiene constante (demostrado en el teorema
;;; medida-resolucion-cota-invariante), habr que probar que el valor de
;;; (medida-resolucion (append (append resolventes (cdr S-1)) (cons (car S-1)
;;; S-2))) es mayor que (medida-resolucion (append S-1 S-2)). Esto
;;; ltimo se tiene por el teorema medida-resolucion-aumenta, con lo que la
;;; prueba termina.

;;; -)
 
;;;============================================================================
